/* $Id: BezierNode.cc,v 1.13 2005/05/07 17:39:32 pklaus Exp $
 *
 * BezierNode.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "BezierNode.hh"
#include "TransformationMatrix.hh"
#include "DocWindow.hh"
#include "BezierPath.hh"
#include "BoundingBox.hh"
#include "GDDocument.hh"
#include <libart_lgpl/art_misc.h>
#include <libart_lgpl/art_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_rgb_svp.h>
#include <libart_lgpl/art_svp_vpath_stroke.h>
#include <math.h>

/* The Bezier base matrix 
 *
 * calculate from the following formulas:
 * xt = a * t^3 + b * t^2 + c * t + d
 * where is:
 * a = -1 * x0 + 3 * x1 - 3 * x2 + 1 * x3
 * b =  3 * x0 - 6 * x1 + 3 * x2
 * c = -3 * x0 + 3 * x1
 * d =  1 * x0
 * and this is how this looks like in a matrix
 */

GdkPoint	gdkpts[SUBDIVISIONS+2];

static BezierMatrix bezierBase = {	{	-1,  3, -3,  1 },
										{	 3, -6,  3,  0 },
										{	-3,  3,  0,  0 },
										{	 1,  0,  0,  0 }
									};

#define CURVE_LINE_SQUARE_TOLERANCE 1.0e-6
#define BETWEEN(a,b,c)	((b < c) ? (a >= b) && (a <= c) : (a >= c) && (a <= b))
#define INSIDE(a,b,c)	((b < c) ? (a >= b) && (a < c) : (a > c) && (a <= b) || a == b && b == c)

inline gint checkLineIntersect(const position p1, const position p2, const position q1, const position q2, position& cross, gdouble& tp, gdouble& tq, gdouble needratio = true) {
	/* check if and where the two lines p1-p2 and q1-q2 intersect and return the
	 * intersection point and the split ratio in both lines */
	// first check the first points for equality
	if (p1.x == q1.x && p1.y == q1.y) {
		cross = p1;
		tp = 0;
		tq = 0;
		return 1;
	}
	/*
	 * the two lines are intersected using the following formulas (general format):
	 * p-line: a * x + b * y = c
	 * where a = p2.y - p1.y, b = p1.x - p2.x and 
	 * c = p1.x * (p2.y - p1.y) + p1.y * (p1.x - p2.x)
	 * 
	 * q-line: d * x + e * y = f
	 * where d = q2.y - q1.y, b = q1.x - q2.x and
	 * f = q1.x * (q2.y - q1.y) + q1.y * (q1.x - q2.x)
	 *
	 *
	 *  for y:                            for x:
	 * adx + bdy = cd                    aex + bey = ce
	 * adx + aey = af                    bdx + bey = bf
	 * 
	 *                  combined:
	 * bdy - aey = cd - af               bdx - aex = bf -ce
	 * y = (cd - af) / (bd - ae)         x = (bf - ce) / (bd - ae)
	 * 
	 * if denominator (bd-ae) is 0 lines are parallel, if numerators are then 0 then the lines coincide
	*/
	gdouble a = p2.y - p1.y, b = p1.x - p2.x;
	gdouble c = p1.x * a + p1.y * b;
	gdouble d = q2.y - q1.y, e = q1.x - q2.x;
	gdouble f = q1.x * d + q1.y * e;
	gdouble denom = b * d - a * e;
	/* if (needratio) {
		g_printerr("checkLineIntersect: a=%f b=%f c=%f d=%f e=%f f=%f den=%f\n", a, b, c, d, e, f, denom);
	} */
	if (denom == 0) return 0;
	cross.x = (b * f - c * e) / denom;
	cross.y = (c * d - a * f) / denom;
	if (needratio) {
		/*
		g_printerr("checkLineIntersect: cross=%f/%f\n", cross.x, cross.y);
		g_printerr("p1=%f/%f p2=%f/%f\n", p1.x, p1.y, p2.x, p2.y);
		g_printerr("q1=%f/%f q2=%f/%f\n", q1.x, q1.y, q2.x, q2.y);
		g_printerr("p1 < c < p1 = %d/%d q1 < c < q2 = %d/%d\n", INSIDE(cross.x, p1.x, p2.x), INSIDE(cross.y, p1.y, p2.y), INSIDE(cross.x, q1.x, q2.x), INSIDE(cross.y, q1.y, q2.y));
		*/
		if ((INSIDE(cross.x, p1.x, p2.x) && INSIDE(cross.y, p1.y, p2.y) || cross.x == p1.x && cross.y == p1.y) &&
			(INSIDE(cross.x, q1.x, q2.x) && INSIDE(cross.y, q1.y, q2.y) || cross.x == q1.x && cross.y == q2.y)) {
			gdouble px = fabs(p2.x - p1.x);
			gdouble py = fabs(p2.y - p1.y);
			if (px == 0 && py == 0) {
				tp = 0;
			} else {
				tp = (px > py) ? fabs(cross.x - p1.x) / px : fabs(cross.y - p1.y) / py;
			}
			gdouble qx = fabs(q2.x - q1.x);
			gdouble qy = fabs(q2.y - q1.y);
			if (qx == 0 && qy == 0) {
				tq = 0;
			} else {
				tq = (qx > qy) ? fabs(cross.x - q1.x) / qx : fabs(cross.y - q1.y) / qy;
			}
			//g_printerr("GOT INTERSECTION!!!\n");
			return 1;
		}
	} else if (BETWEEN(cross.x, p1.x, p2.x) && BETWEEN(cross.x, q1.x, q2.x) &&
			BETWEEN(cross.y, p1.y, p2.y) && BETWEEN(cross.y, q1.y, q2.y)) {
		return 1;
	}
	return 0;
}

inline gdouble pointLineSqrDist(position p1, position p2, position o) {
	gdouble dx = p2.x - p1.x;
	gdouble dy = p2.y - p1.y;
	gdouble m = G_MAXDOUBLE;
	if (dx != 0.0) {
		m = dy / dx;
	}
	gdouble xs, ys; // coordinates of crossing point
	// allow extrema
	if (fabs(m) < 0.000001) {
		// almost horizontal
		xs = o.x;
		ys = p1.y;
	} else if (fabs(m) > 1000000.0) {
		// almost vertical
		xs = p1.x;
		ys = o.y;
	} else {
		// normal case
		// n1: cross point of the segpoint line with the x-axis
		// n2: cross point of the right-angled line with the x-axis
		gdouble n1, n2;
		n1 = p1.x - p1.y / m;
		n2 = o.x + o.y * m;
		ys = (m * (n2 - n1)) / (1 + m * m);
		xs = n1 + (ys / m);
	}
	gdouble deltax = o.x - xs;
	gdouble deltay = o.y - ys;
	return deltax * deltax + deltay * deltay;
}

void divideSegment(const CurveSegment& source, CurveSegment& first, CurveSegment& second) {
	first.origNode = source.origNode;
	first.segmentSize = source.segmentSize / 2.0;
	first.splitbase = source.splitbase;
	first.oc1 = source.oc1;
	second.origNode = source.origNode;
	second.segmentSize = source.segmentSize - first.segmentSize;
	second.splitbase = source.splitbase + first.segmentSize;
	second.oc2 = source.oc2;
	first.ha1.x = source.oc1.x * 0.5 + source.ha1.x * 0.5;
	first.ha1.y = source.oc1.y * 0.5 + source.ha1.y * 0.5;
	first.ha2.x = source.oc1.x * 0.25 + source.ha1.x * 0.5 + source.ha2.x * 0.25;
	first.ha2.y = source.oc1.y * 0.25 + source.ha1.y * 0.5 + source.ha2.y * 0.25;
	first.oc2.x = source.oc1.x * 0.125 + source.ha1.x * 0.375 + source.ha2.x * 0.375 + source.oc2.x * 0.125;
	first.oc2.y = source.oc1.y * 0.125 + source.ha1.y * 0.375 + source.ha2.y * 0.375 + source.oc2.y * 0.125;
	second.oc1 = first.oc2;
	second.ha1.x = source.ha1.x * 0.25 + source.ha2.x * 0.5 + source.oc2.x * 0.25;
	second.ha1.y = source.ha1.y * 0.25 + source.ha2.y * 0.5 + source.oc2.y * 0.25;
	second.ha2.x = source.ha2.x * 0.5 + source.oc2.x * 0.5;
	second.ha2.y = source.ha2.y * 0.5 + source.oc2.y * 0.5;
}

IntersectionVector *BezierNode::getIntersections(CurveSegment& a, CurveSegment& b) {
	/* check if curve segment a and curve segment b intersect and return all
	 * point where they do (nine are theoretically possible).
	 * This is done using the algorithm described in "Applied Geometry for
	 * Computer Graphics and CAD" by Duncan Marsh in section 6.9 using a
	 * recursive subdivision (by Casteljau algorithm)
	 */
	IntersectionVector *result = NULL;
	position ints;
	gdouble ta, tb;
	// check if both segments are almost linear
	if (pointLineSqrDist(a.oc1, a.oc2, a.ha1) < CURVE_LINE_SQUARE_TOLERANCE &&
		pointLineSqrDist(a.oc1, a.oc2, a.ha2) < CURVE_LINE_SQUARE_TOLERANCE) {
		a.isLine = true;
	}
	if (pointLineSqrDist(b.oc1, b.oc2, b.ha1) < CURVE_LINE_SQUARE_TOLERANCE &&
		pointLineSqrDist(b.oc1, b.oc2, b.ha2) < CURVE_LINE_SQUARE_TOLERANCE) {
		b.isLine = true;
	}
	//g_printerr("BezierNode::getIntersections a.isLine=%d b.isLine=%d a.sb=%f b.sb=%f\n", a.isLine, b.isLine, a.splitbase, b.splitbase);
	if (a.isLine && b.isLine) {
		if (checkLineIntersect(a.oc1, a.oc2, b.oc1, b.oc2, ints, ta, tb, true)) {
			result = new IntersectionVector();
			Intersection is;
			is.nodeA = a.origNode;
			is.nodeB = b.origNode;
			is.intersectPos = ints;
			is.splitRatioA = a.splitbase + ta * a.segmentSize;
			is.splitRatioB = b.splitbase + tb * b.segmentSize;
			result->push_back(is);
			//g_printerr("BezierNode::getIntersections GOT INTERSECTION result=%lX\n", (long) result);
		}
		return result;
	}
	if (checkLineIntersect(a.oc1, a.ha1, b.oc1, b.ha1, ints, ta, tb, false) ||
		checkLineIntersect(a.oc1, a.ha1, b.ha1, b.ha2, ints, ta, tb, false) ||
		checkLineIntersect(a.oc1, a.ha1, b.ha2, b.oc2, ints, ta, tb, false) ||
		checkLineIntersect(a.oc1, a.ha1, b.oc2, b.oc1, ints, ta, tb, false) ||
		checkLineIntersect(a.ha1, a.ha2, b.oc1, b.ha1, ints, ta, tb, false) ||
		checkLineIntersect(a.ha1, a.ha2, b.ha1, b.ha2, ints, ta, tb, false) ||
		checkLineIntersect(a.ha1, a.ha2, b.ha2, b.oc2, ints, ta, tb, false) ||
		checkLineIntersect(a.ha1, a.ha2, b.oc2, b.oc1, ints, ta, tb, false) ||
		checkLineIntersect(a.ha2, a.oc2, b.oc1, b.ha1, ints, ta, tb, false) ||
		checkLineIntersect(a.ha2, a.oc2, b.ha1, b.ha2, ints, ta, tb, false) ||
		checkLineIntersect(a.ha2, a.oc2, b.ha2, b.oc2, ints, ta, tb, false) ||
		checkLineIntersect(a.ha2, a.oc2, b.oc2, b.oc1, ints, ta, tb, false) ||
		checkLineIntersect(a.oc2, a.oc1, b.oc1, b.ha1, ints, ta, tb, false) ||
		checkLineIntersect(a.oc2, a.oc1, b.ha1, b.ha2, ints, ta, tb, false) ||
		checkLineIntersect(a.oc2, a.oc1, b.ha2, b.oc2, ints, ta, tb, false) ||
		checkLineIntersect(a.oc2, a.oc1, b.oc2, b.oc1, ints, ta, tb, false)) {
		// convex hulls of the two segments intersect, curve intersection is possible
		gint asubs = 1;
		gint bsubs = 1;
		CurveSegment asegs[2];
		CurveSegment bsegs[2];
		if (! a.isLine) {
			asubs = 2;
			divideSegment(a, asegs[0], asegs[1]);
		} else {
			asegs[0] = a;
		}
		if (! b.isLine) {
			bsubs = 2;
			divideSegment(b, bsegs[0], bsegs[1]);
		} else {
			bsegs[0] = b;
		}
		for (int i = 0; i < asubs; i++) {
			for (int j = 0; j < bsubs; j++) {
				IntersectionVector *intersects = getIntersections(asegs[i], bsegs[j]);
				if (intersects) {
					if (result) {
						for (IntersectionVector::iterator it = intersects->begin(); it != intersects->end(); it++) {
							result->push_back(*it);
						}
						delete intersects;
					} else {
						result = intersects;
					}
				}
			}
		}
	}
	return result;
}
									
BezierNode::BezierNode(BezierPath *parent) {
	path = parent;
	straightHandles = true;
	selection = 0;
}

void BezierNode::matrixMultiply(BezierMatrix a, BezierMatrix b, BezierMatrix ab) {
	int i,j;
	
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 4; j++) {
			ab[i][j] = a[i][0] * b[0][j] + a[i][1] * b[1][j] + a[i][2] * b[2][j] + a[i][3] * b[3][j];
		}
	}
}

gint32 BezierNode::calcSegment(BezierNode *b, int subdivisions, 
									GdkPoint *gdkpts, gdouble zoom, 
									gdouble xoffset, gdouble yoffset) {

	BezierNode *a = this;
	BezierMatrix		coords, ma, temp, incr;
	gdouble		x, y, dx, dy, dx2, dy2, dx3, dy3;
	gdouble		dt, dts, dtc;
	gint32		lx, ly, nx, ny;
	gint32		ind, i;
	gdouble X0 = a->onCurve.x;
	gdouble Y0 = a->onCurve.y;
	gdouble X1 = a->nextHandle.x;
	gdouble Y1 = a->nextHandle.y;
	gdouble X2 = b->prevHandle.x;
	gdouble Y2 = b->prevHandle.y;
	gdouble X3 = b->onCurve.x;
	gdouble Y3 = b->onCurve.y;

	/* fill the coordinates in the temporary matrix */
	coords[0][0] = X0; coords[0][1] = Y0; coords[0][2] = 0; coords[0][3] = 0;
	coords[1][0] = X1; coords[1][1] = Y1; coords[1][2] = 0; coords[1][3] = 0;
	coords[2][0] = X2; coords[2][1] = Y2; coords[2][2] = 0; coords[2][3] = 0;
	coords[3][0] = X3; coords[3][1] = Y3; coords[3][2] = 0; coords[3][3] = 0;
	/* calculate the delta t's and dt^2 as well as dt^3 */
	dt = 1.0 / subdivisions;
	dts = dt * dt;
	dtc = dts * dt;
	/* fill the matrix for calculating the loop increments */
	temp[0][0] = 0; temp[0][1] = 0; temp[0][2] = 0; temp[0][3] = 1;
	temp[1][0] = dtc; temp[1][1] = dts; temp[1][2] = dt; temp[1][3] = 0;
	temp[2][0] = 6*dtc; temp[2][1] = 2*dts; temp[2][2] = 0; temp[2][3] = 0;
	temp[3][0] = 6*dtc; temp[3][1] = 0; temp[3][2] = 0; temp[3][3] = 0;
	/* calc the main matrix */
	matrixMultiply(bezierBase, coords, ma);
	/* and using this and the temp we calc the increments */
	matrixMultiply(temp, ma, incr);
	/* extract the increments */
	x = incr[0][0]; dx = incr[1][0]; dx2 = incr[2][0]; dx3 = incr[3][0];
	y = incr[0][1]; dy = incr[1][1]; dy2 = incr[2][1]; dy3 = incr[3][1];
	
	/* now loop for subdivisions times */
	lx = ROUND((x - xoffset) * zoom);
	ly = ROUND((y - yoffset) * zoom);
	gdkpts[0].x = (lx);
	gdkpts[0].y = (ly);
	ind = 1;
	for (i = 0; i < subdivisions; i++) {
		x += dx; dx += dx2; dx2 += dx3;
		y += dy; dy += dy2; dy2 += dy3;
		nx = ROUND((x - xoffset) * zoom);
		ny = ROUND((y - yoffset) * zoom);
		if ((lx != nx) || (ly != ny)) {
			/* only add new points */
			gdkpts[ind].x = nx;
			gdkpts[ind].y = ny;
			ind++;
		}
		lx = nx; ly = ny;
	}
	return ind;
}

void BezierNode::calcSegment(BezierNode *b, int subdivisions, position *pos) {

	BezierNode *a = this;
	BezierMatrix		coords, ma, temp, incr;
	gdouble		x, y, dx, dy, dx2, dy2, dx3, dy3;
	gdouble		dt, dts, dtc;
	gint32		i;
	gdouble X0 = a->onCurve.x;
	gdouble Y0 = a->onCurve.y;
	gdouble X1 = a->nextHandle.x;
	gdouble Y1 = a->nextHandle.y;
	gdouble X2 = b->prevHandle.x;
	gdouble Y2 = b->prevHandle.y;
	gdouble X3 = b->onCurve.x;
	gdouble Y3 = b->onCurve.y;

	/* fill the coordinates in the temporary matrix */
	coords[0][0] = X0; coords[0][1] = Y0; coords[0][2] = 0; coords[0][3] = 0;
	coords[1][0] = X1; coords[1][1] = Y1; coords[1][2] = 0; coords[1][3] = 0;
	coords[2][0] = X2; coords[2][1] = Y2; coords[2][2] = 0; coords[2][3] = 0;
	coords[3][0] = X3; coords[3][1] = Y3; coords[3][2] = 0; coords[3][3] = 0;
	/* calculate the delta t's and dt^2 as well as dt^3 */
	dt = 1.0 / subdivisions;
	dts = dt * dt;
	dtc = dts * dt;
	/* fill the matrix for calculating the loop increments */
	temp[0][0] = 0; temp[0][1] = 0; temp[0][2] = 0; temp[0][3] = 1;
	temp[1][0] = dtc; temp[1][1] = dts; temp[1][2] = dt; temp[1][3] = 0;
	temp[2][0] = 6*dtc; temp[2][1] = 2*dts; temp[2][2] = 0; temp[2][3] = 0;
	temp[3][0] = 6*dtc; temp[3][1] = 0; temp[3][2] = 0; temp[3][3] = 0;
	/* calc the main matrix */
	matrixMultiply(bezierBase, coords, ma);
	/* and using this and the temp we calc the increments */
	matrixMultiply(temp, ma, incr);
	/* extract the increments */
	x = incr[0][0]; dx = incr[1][0]; dx2 = incr[2][0]; dx3 = incr[3][0];
	y = incr[0][1]; dy = incr[1][1]; dy2 = incr[2][1]; dy3 = incr[3][1];
	
	/* now loop for subdivisions times */
	pos[0].x = x;
	pos[0].y = y;
	for (i = 1; i <= subdivisions; i++) {
		x += dx; dx += dx2; dx2 += dx3;
		y += dy; dy += dy2; dy2 += dy3;
		pos[i].x = x;
		pos[i].y = y;
	}
}

void BezierNode::drawCurve(DocWindow *docwin, Gdk_GC& context) {
	gint32		numpts;
	
	if (next) {
		numpts = calcSegment(next, SUBDIVISIONS, gdkpts, docwin->osBuffer.scaleFactor,
							docwin->osBuffer.origin.x, docwin->osBuffer.origin.y);
		if (numpts)
			docwin->DocWinDrawingArea->get_window().draw_lines(context, gdkpts, numpts);
	}
}
	
void BezierNode::drawSegment(DocWindow *docwin, Gdk_GC& context) {
	gint32		numpts;
	
	if (prev && ! prev->selection) {
		numpts = prev->calcSegment(this, SUBDIVISIONS, gdkpts, 
					docwin->osBuffer.scaleFactor, docwin->osBuffer.origin.x, docwin->osBuffer.origin.y);
		if (numpts)
			docwin->DocWinDrawingArea->get_window().draw_lines(context, gdkpts, numpts);
	}
	if (next) {
		numpts = calcSegment(next, SUBDIVISIONS, gdkpts, docwin->osBuffer.scaleFactor,
								docwin->osBuffer.origin.x, docwin->osBuffer.origin.y);
		if (numpts)
			docwin->DocWinDrawingArea->get_window().draw_lines(context, gdkpts, numpts);
	}
}
	
void BezierNode::drawSegments(DocWindow *docwin, Gdk_GC& context) {
	gint32		numpts;
	
	if (prev) {
		numpts = prev->calcSegment(this, SUBDIVISIONS, gdkpts, 
						docwin->osBuffer.scaleFactor, docwin->osBuffer.origin.x, docwin->osBuffer.origin.y);
		if (numpts)
			docwin->DocWinDrawingArea->get_window().draw_lines(context, gdkpts, numpts);
	}
	if (next) {
		numpts = calcSegment(next, SUBDIVISIONS, gdkpts, docwin->osBuffer.scaleFactor,
						docwin->osBuffer.origin.x, docwin->osBuffer.origin.y);
		if (numpts)
			docwin->DocWinDrawingArea->get_window().draw_lines(context, gdkpts, numpts);
	}
}
	
void BezierNode::setOnCurvePoint(gdouble x, gdouble y) {
	onCurve.x = x;
	onCurve.y = y;
}

void BezierNode::setPreviousHandle(gdouble x, gdouble y) {
	prevHandle.x = x;
	prevHandle.y = y;
}

void BezierNode::setNextHandle(gdouble x, gdouble y) {
	nextHandle.x = x;
	nextHandle.y = y;
}

void BezierNode::connectNodes(BezierNode *to) {
	//g_printerr("BN:connectNodes connecting %lX and %lX\n", (long) this, (long) to);
	BezierNode *node1 = this;
	BezierNode *node2 = to;
	gboolean reverseBack = false;
	if (node1->next) {
		//g_printerr("BN::connectNodes reversing path 1 (%lX) (prev=%lX)\n", (long) node1->path, (long) node1->prev);
		node1->path->reverseOrder();
		reverseBack = true; // we change the direction of the resulting path, remind to change back
	}
	if (node1->path == node2->path) {
		// we are just closing the path
		//g_printerr("BN:connectNodes closing path %lX\n", (long) node1->path);
		node1->path->nodes = node1;
	} else {
		if (node2->prev) {
			//g_printerr("BN::connectNodes reversing path 2 (%lX) (next=%lX)\n", (long) node2->path, (long) node2->next);
			node2->path->reverseOrder();
		}
		// attach nodes of path of node2 to path of node1
		BezierPath *oldpath = node2->path;
		BezierNode *node = node2;
		while (node) {
			//g_printerr("BN::connectNodes setting node %lX path to %lX\n", (long) node, (long) node1->path);
			node->path = node1->path;
			node = node->next;
		}
		//g_printerr("BN::connectNodes removing old path %lX\n", (long) oldpath);
		oldpath->nodes = NULL;
		oldpath->doc->removeObject(oldpath);
		delete oldpath;
	}
	// remove node2 and connect the two paths
	node1->next = node2->next;
	node1->next->prev = node1;
	node1->nextHandle.x = node2->nextHandle.x;
	node1->nextHandle.y = node2->nextHandle.y;
	node1->straightHandles = false;
	delete node2;
	if (reverseBack) {
		node1->path->reverseOrder();
	}
}


void BezierNode::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox) || (next && drawArea.intersect(next->getBoundingBox()))) {
		drawSegment(docwin, context);
		drawPoint(docwin, context);
		if ((! prev || prev->selection == 0) && (! next || next->selection == 0)) {
			drawPreviousHandle(docwin, context);
			drawNextHandle(docwin, context);
		}
	}
}

SelectableItem *BezierNode::clone() {
	BezierNode *clone = new BezierNode(path);
	clone->onCurve = onCurve;
	clone->prevHandle = prevHandle;
	clone->nextHandle = nextHandle;
	clone->selection = selection;
	clone->selectedItem = selectedItem;
	clone->next = next;
	clone->prev = prev;
	clone->straightHandles = straightHandles;
	return clone;
}

void BezierNode::setCoords(SelectableItem *from) {
	BezierNode *f = (BezierNode *) from;
	onCurve = f->onCurve;
	prevHandle = f->prevHandle;
	nextHandle = f->nextHandle;
}

void BezierNode::bufferDrawPoint(OffscreenBuffer& osBuffer, guint32 color) {
	ArtVpath apath[6];
	ArtSVP *asvp;
	if (selection == 1) {
		apath[0].code = ART_MOVETO;
		apath[0].x = ROUND((onCurve.x - osBuffer.origin.x) * osBuffer.scaleFactor) - BEZIER_HALFSIZE;
		apath[0].y = ROUND((onCurve.y - osBuffer.origin.y) * osBuffer.scaleFactor) - BEZIER_HALFSIZE;
		apath[1].code = ART_LINETO;
		apath[1].x = apath[0].x + BEZIER_SIZE;
		apath[1].y = apath[0].y;
		apath[2].code = ART_LINETO;
		apath[2].x = apath[1].x;
		apath[2].y = apath[1].y + BEZIER_SIZE;
		apath[3].code = ART_LINETO;
		apath[3].x = apath[0].x;
		apath[3].y = apath[2].y;
		apath[4].code = ART_LINETO;
		apath[4].x = apath[0].x;
		apath[4].y = apath[0].y;
		apath[5].code = ART_END;
		apath[5].x = 0;
		apath[5].y = 0;
		asvp = art_svp_from_vpath(apath);
		art_rgb_svp_alpha(asvp, 0, 0,
					osBuffer.pxbWidth,
					osBuffer.pxbHeight,
					color,
					osBuffer.pixbuffer,
					osBuffer.pxbRowstride,
					NULL);
	} else {
		apath[0].code = ART_MOVETO;
		apath[0].x = ROUND((onCurve.x - osBuffer.origin.x) * osBuffer.scaleFactor) - BEZIER_HALFSIZE;
		apath[0].y = ROUND((onCurve.y - osBuffer.origin.y) * osBuffer.scaleFactor) - BEZIER_HALFSIZE;
		apath[1].code = ART_LINETO;
		apath[1].x = apath[0].x + BEZIER_SIZE;
		apath[1].y = apath[0].y;
		apath[2].code = ART_LINETO;
		apath[2].x = apath[1].x;
		apath[2].y = apath[1].y + BEZIER_SIZE;
		apath[3].code = ART_LINETO;
		apath[3].x = apath[0].x;
		apath[3].y = apath[2].y;
		apath[4].code = ART_LINETO;
		apath[4].x = apath[0].x;
		apath[4].y = apath[0].y;
		apath[5].code = ART_END;
		apath[5].x = 0;
		apath[5].y = 0;
		asvp = art_svp_from_vpath(apath);
		art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
		art_svp_free(asvp);
		apath[0].x++;
		apath[0].y++;
		apath[1].x--;
		apath[1].y++;
		apath[2].x--;
		apath[2].y--;
		apath[3].x++;
		apath[3].y--;
		apath[4].x++;
		apath[4].y++;
		asvp = art_svp_from_vpath(apath);
		art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				0xFFFFFFFF,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
		
	}
	art_svp_free(asvp);
}

void BezierNode::bufferDrawPreviousHandle(OffscreenBuffer& osBuffer, guint32 color) {
	ArtVpath apath[3];
	apath[0].code = ART_MOVETO;
	apath[0].x = ROUND((onCurve.x - osBuffer.origin.x) * osBuffer.scaleFactor);
	apath[0].y = ROUND((onCurve.y - osBuffer.origin.y) * osBuffer.scaleFactor);
	apath[1].code = ART_LINETO;
	apath[1].x = ROUND((prevHandle.x - osBuffer.origin.x) * osBuffer.scaleFactor);
	apath[1].y = ROUND((prevHandle.y - osBuffer.origin.y) * osBuffer.scaleFactor);
	apath[2].code = ART_END;
	apath[2].x = 0;
	apath[2].y = 0;
	ArtSVP *asvp = art_svp_vpath_stroke(apath,
								ART_PATH_STROKE_JOIN_MITER,
								ART_PATH_STROKE_CAP_BUTT,
								1, 4, 0.5);
	art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
	art_svp_free(asvp);
	ArtVpath *avp = art_vpath_new_circle(
				ROUND((prevHandle.x - osBuffer.origin.x) * osBuffer.scaleFactor),
				ROUND((prevHandle.y - osBuffer.origin.y) * osBuffer.scaleFactor),
				BEZIER_RADIUS);
	asvp = art_svp_from_vpath(avp);
	art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
	art_svp_free(asvp);
	
}

void BezierNode::bufferDrawNextHandle(OffscreenBuffer& osBuffer, guint32 color) {
	ArtVpath apath[3];
	apath[0].code = ART_MOVETO;
	apath[0].x = ROUND((onCurve.x - osBuffer.origin.x) * osBuffer.scaleFactor);
	apath[0].y = ROUND((onCurve.y - osBuffer.origin.y) * osBuffer.scaleFactor);
	apath[1].code = ART_LINETO;
	apath[1].x = ROUND((nextHandle.x - osBuffer.origin.x) * osBuffer.scaleFactor);
	apath[1].y = ROUND((nextHandle.y - osBuffer.origin.y) * osBuffer.scaleFactor);
	apath[2].code = ART_END;
	apath[2].x = 0;
	apath[2].y = 0;
	ArtSVP *asvp = art_svp_vpath_stroke(apath,
								ART_PATH_STROKE_JOIN_MITER,
								ART_PATH_STROKE_CAP_BUTT,
								1, 4, 0.5);
	art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
	art_svp_free(asvp);
	ArtVpath *avp = art_vpath_new_circle(
				ROUND((nextHandle.x - osBuffer.origin.x) * osBuffer.scaleFactor),
				ROUND((nextHandle.y - osBuffer.origin.y) * osBuffer.scaleFactor),
				BEZIER_RADIUS);
	asvp = art_svp_from_vpath(avp);
	art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
	art_svp_free(asvp);
}

void BezierNode::drawSingleNode(DocWindow *docwin, Gdk_GC& context) {
	Gdk_Window window = docwin->DocWinDrawingArea->get_window();
	gint x = (gint) ((onCurve.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor);
	gint y = (gint) ((onCurve.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor);
	window.draw_line(context, x - BEZIER_HALFSIZE, y - BEZIER_HALFSIZE, 
						x + BEZIER_HALFSIZE, y + BEZIER_HALFSIZE);
	window.draw_line(context, x - BEZIER_HALFSIZE, y + BEZIER_HALFSIZE, 
						x + BEZIER_HALFSIZE, y - BEZIER_HALFSIZE);
}

void BezierNode::drawPoint(DocWindow *docwin, Gdk_GC& context) {
	docwin->DocWinDrawingArea->get_window().draw_rectangle(context, selection == 1, 
		ROUND((onCurve.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor) - BEZIER_HALFSIZE,
		ROUND((onCurve.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor) - BEZIER_HALFSIZE,
						BEZIER_SIZE, BEZIER_SIZE);
}

void BezierNode::drawPreviousHandle(DocWindow *docwin, Gdk_GC& context) {
	/* draw the handle point */
	docwin->DocWinDrawingArea->get_window().draw_arc(context, 1, 
		ROUND((prevHandle.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor) - BEZIER_RADIUS,
		ROUND((prevHandle.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor) - BEZIER_RADIUS, BEZIER_WIDTH, BEZIER_WIDTH,
				0, 23040);
	/* draw the line between the points */
	docwin->DocWinDrawingArea->get_window().draw_line(context, 
		ROUND((onCurve.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor), 
		ROUND((onCurve.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor),
		ROUND((prevHandle.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor), 
		ROUND((prevHandle.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor));
}

void BezierNode::drawNextHandle(DocWindow *docwin, Gdk_GC& context) {
	/* draw the handle point */
	docwin->DocWinDrawingArea->get_window().draw_arc(context, 1, 
		ROUND((nextHandle.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor) - BEZIER_RADIUS,
		ROUND((nextHandle.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor) - BEZIER_RADIUS, 
		BEZIER_WIDTH, BEZIER_WIDTH, 0, 23040);
	/* draw the line between the points */
	docwin->DocWinDrawingArea->get_window().draw_line(context, 
			ROUND((onCurve.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor), 
			ROUND((onCurve.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor),
			ROUND((nextHandle.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor), 
			ROUND((nextHandle.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor));
}

gboolean BezierNode::testPreviousHandleVisible() {
	if (selection == 1 && (! prev || prev->selection != 1)
						&& (! next || next->selection != 1)) {
		return true;
	} else if (selection != 1 && prev && prev->selection) {
		return true;
	}
	return false;
}

gboolean BezierNode::testNextHandleVisible() {
	if (selection == 1 && (! prev || prev->selection != 1)
						&& (! next || next->selection != 1)) {
		return true;
	} else if (selection != 1 && next && next->selection) {
		return true;
	}
	return false;
}

gboolean BezierNode::checkRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom) {
	return (onCurve.x >= left && onCurve.x <= right && onCurve.y >= top && onCurve.y <= bottom);
}

gint BezierNode::checkPoint(gdouble x, gdouble y, gdouble tolerance) {
	gdouble l, r, t, b;
	
	l = onCurve.x - tolerance;
	r = onCurve.x + tolerance;
	t = onCurve.y - tolerance;
	b = onCurve.y + tolerance;

	if ((x >= l) && (x <= r) && (y >= t) && (y <= b)) {
		return 1;
	}
	return 0;
}

gint BezierNode::checkHandles(gdouble x, gdouble y, gdouble tolerance) {
	gdouble l1, r1, t1, b1, l2, r2, t2, b2;
	
	l1 = prevHandle.x - tolerance;
	r1 = prevHandle.x + tolerance;
	t1 = prevHandle.y - tolerance;
	b1 = prevHandle.y + tolerance;
	l2 = nextHandle.x - tolerance;
	r2 = nextHandle.x + tolerance;
	t2 = nextHandle.y - tolerance;
	b2 = nextHandle.y + tolerance;
		
	if ((x >= l1) && (x <= r1) && (y >= t1) && (y <= b1)) {
		return -1;
	}
	if ((x >= l2) && (x <= r2) && (y >= t2) && (y <= b2)) {
		return 1;
	}
	
	return 0;
}

void BezierNode::boundingBox(gdouble& l, gdouble& t, gdouble& r, gdouble& b) {
	if (onCurve.x < l)
		l = onCurve.x;
	if (onCurve.x > r)
		r = onCurve.x;
	if (onCurve.y < t)
		t = onCurve.y;
	if (onCurve.y > b)
		b = onCurve.y;
	if (prevHandle.x < l)
		l = prevHandle.x;
	if (prevHandle.x > r)
		r = prevHandle.x;
	if (prevHandle.y < t)
		t = prevHandle.y;
	if (prevHandle.y > b)
		b = prevHandle.y;
	if (nextHandle.x < l)
		l = nextHandle.x;
	if (nextHandle.x > r)
		r = nextHandle.x;
	if (nextHandle.y < t)
		t = nextHandle.y;
	if (nextHandle.y > b)
		b = nextHandle.y;
	//g_printerr("BezierNode::boundingBox: %f,%f %f,%f\n", l, t, r, b);
	//g_printerr("BezierNode::boundingBox: vals: %f,%f  %f,%f  %f,%f\n", prevHandle.x, prevHandle.y, onCurve.x, onCurve.y, nextHandle.x, nextHandle.y);
}

void BezierNode::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	//g_printerr("BezierNode::transformByMatrix before onCurve: %f,%f\n", onCurve.x, onCurve.y);
	m.transform(onCurve);
	m.transform(prevHandle);
	m.transform(nextHandle);
	//g_printerr("BezierNode::transformByMatrix after onCurve: %f,%f\n", onCurve.x, onCurve.y);
}

void BezierNode::select() {
	//g_printerr("BezierNode::select()  node: %lX, path: %lX\n", this, path);
	// remember BezierPath selectAll
	selection = 1;
	path->doc->selection->insertObject(path);
}

void BezierNode::selectCurve() {
	//g_printerr("BezierNode::selectCurve\n");
	selection = 2;
	path->doc->selection->insertObject(path);
}

void BezierNode::deSelect() {
	// remember BezierPath deSelectAll
	//g_printerr("BezierNode::deSelect()  node: %lX, path: %lX\n", this, path);
	selection = 0;
	if (! path->partiallySelected()) {
		path->doc->selection->removeObject(path);
	}
}

gboolean BezierNode::selected() {
	return selection == 1;
}

void BezierNode::select(gboolean shiftSelect) {
	if (shiftSelect && selected()) {
		deSelect();
	} else {
		select();
	}
}

void BezierNode::selectPath(gboolean shiftSelect) {
	path->select(shiftSelect);
}

void BezierNode::deSelectPath() {
	path->deSelect();
}

void BezierNode::selectObject(gboolean shiftSelect) {
	path->selectObject(shiftSelect);
}

void BezierNode::deSelectObject() {
	path->deSelectObject();
}

int BezierNode::getCoordinates(gdouble& x, gdouble& y) {
	x = onCurve.x;
	y = onCurve.y;
	return 1;
}

void BezierNode::reverseOrder() {
	position tmp = nextHandle;
	nextHandle = prevHandle;
	prevHandle = tmp;
}

void BezierNode::clear() {
	if (selection == 2) {
		path->deleteCurve(this);
	} else {
		path->deleteNode(this);
		delete this;
	}
}

BezierNode *BezierNode::subDivide(gdouble t) {
	// as described in "Computer Graphics, Principles and Practice"
	// Foley, van Dam, Feiner, Hughes et al. page 508 but converted
	// from factor 1/2 to dynamic factor 't'
	if (! next) {
		BezierNode *newNode = (BezierNode *) clone();
		newNode->prevHandle = newNode->onCurve;
		newNode->nextHandle = newNode->onCurve;
		next = newNode;
		newNode->next = NULL;
		newNode->prev = this;
	}
	//g_printerr("BezierNode::subDivide t=%f\n", t);
	BezierNode *newNode = new BezierNode(path);
	position p1, p2, p3, p4;
	p1 = onCurve;
	p2 = nextHandle;
	p3 = next->prevHandle;
	p4 = next->onCurve;
	// handle special case of straight lines
	if (p1.x == p2.x && p1.y == p2.y && p3.x == p4.x && p3.y == p4.y) {
		gdouble tr = 1 - t;
		newNode->onCurve.x = tr * p1.x + t * p4.x;
		newNode->onCurve.y = tr * p1.y + t * p4.y;
		newNode->prevHandle = newNode->onCurve;
		newNode->nextHandle = newNode->onCurve;
	} else {
		//g_printerr("p1: %f, %f\n", p1.x, p1.y);
		//g_printerr("p2: %f, %f\n", p2.x, p2.y);
		//g_printerr("p3: %f, %f\n", p3.x, p3.y);
		//g_printerr("p4: %f, %f\n", p4.x, p4.y);
		t = 1 - t;
		gdouble ts = t * t;
		gdouble tc = ts * t;
		gdouble tr = 1 - t;
		gdouble trs = tr * tr;
		gdouble trc = trs * tr;
		//g_printerr("ts: %f, tc: %f\n", ts, tc);
		// left side
		/* tr = (1-t)
						1		0		0		0
				L		t		tr		0		0
			  D		=	t^2		2ttr	tr^2	0
				B		t^3		3t^2tr	3ttr^2	tr^3
		*/
		nextHandle.x = p1.x * t + p2.x * tr;
		nextHandle.y = p1.y * t + p2.y * tr;
		newNode->prevHandle.x = p1.x * ts + (p2.x + p2.x) * t * tr + p3.x * trs;
		newNode->prevHandle.y = p1.y * ts + (p2.y + p2.y) * t * tr + p3.y * trs;
		newNode->onCurve.x = p1.x * tc + (p2.x + p2.x + p2.x) * ts * tr + 
										 (p3.x + p3.x + p3.x) * t * trs + p4.x * trc;
		newNode->onCurve.y = p1.y * tc + (p2.y + p2.y + p2.y) * ts * tr + 
										 (p3.y + p3.y + p3.y) * t * trs + p4.y * trc;
		// right side
		/*
						t^3		3t^2tr	3ttr^2	tr^3
				R		0		t^2		2ttr	tr^2
			  D		=	0		0		t		tr
				B		0		0		0		1
		*/
		newNode->nextHandle.x = p2.x * ts + (p3.x + p3.x) * t * tr + p4.x * trs;
		newNode->nextHandle.y = p2.y * ts + (p3.y + p3.y) * t * tr + p4.y * trs;
		next->prevHandle.x = p3.x * t + p4.x * tr;
		next->prevHandle.y = p3.y * t + p4.y * tr;
	}
	next->prev = newNode;
	newNode->next = next;
	next = newNode;
	newNode->prev = this;
	if (selected()) {
		newNode->select();
	}
	//g_printerr("NewNode: onCurve %f,%f\n", newNode->onCurve.x, newNode->onCurve.y);
	return newNode;
}

BoundingBox BezierNode::getBoundingBox() {
	BoundingBox bBox;
	boundingBox(bBox.left, bBox.top, bBox.right, bBox.bottom);
	return bBox;
}


gchar *BezierNode::identify() {
	return "BezierNode";
}

BezierNode::~BezierNode() {
	/* deSelect(); removed for clean-up */
	//g_printerr("Removing Bezier Node\n");
}
