/* $Id: ClipBoard.cc,v 1.12 2004/07/21 06:31:20 pklaus Exp $
 *
 * ClipBoard.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "ClipBoard.hh"
#include "SelectionItemList.hh"
#include "GDDocument.hh"
#include "DocWindow.hh"
#include "BoundingBox.hh"
#include "gestalter.h"
#include "GDApplication.hh"
#include "TextTool.hh"
#include "TextObject.hh"
#include "AddAction.hh"
#include "TextAction.hh"
#include "DocumentMenu.hh"
#include "Gradient.hh"
#include "Pattern.hh"

ClipBoard::ClipBoard() {
	content = NULL;
}

void ClipBoard::copy(GDDocument *doc) {
	if (content) {
		delete content;
		content = NULL;
	}
	textContent.clear();
	if (! strncmp(doc->application->currentTool->toolName, "Text", 4) && ((TextTool *) (doc->application->currentTool))->editing()) {
		TextTool *tTool = (TextTool *) doc->application->currentTool;
		textContent = tTool->getCurrentObject()->getTextSelection();
		//g_printerr("Copied %d characters to text clipboard\n", textContent.size());
	} else {
		content = new SelectionItemList();
		doc->getSelection(*content);
		doc->application->docMenu->updateItemSensitivity(doc);
		//g_printerr("Copied selection into clipboard.content=%lX\n", content);
	}
	//g_printerr("New content copied to clipboard\n");
}

void ClipBoard::paste(GDDocument *doc, DocWindow *docwin) {
	//g_printerr("Pasting current content (%lX) to doc %lX\n", content, doc);
	if (! strncmp(doc->application->currentTool->toolName, "Text", 4) && ((TextTool *) (doc->application->currentTool))->editing()) {
		//g_printerr("Pasting text content of size %d\n", textContent.size());
		TextTool *tTool = (TextTool *) doc->application->currentTool;
		TextAction *newAction = new TextAction(tTool->getCurrentObject(), textContent);
		doc->setAction(newAction);
		tTool->getCurrentObject()->insertTextSelection(textContent);
	} else {
		if (! content) {
				return;
		}
		BoundingBox Bbox = content->getBoundingBox();
		//g_printerr("Bounding box %f, %f    %f, %f\n", Bbox.left, Bbox.top, Bbox.right, Bbox.bottom);
		if (docwin) {
			gdouble w = docwin->DocWinDrawingArea->width() / docwin->osBuffer.scaleFactor;
			gdouble h = docwin->DocWinDrawingArea->height() / docwin->osBuffer.scaleFactor;
			gdouble newX = docwin->osBuffer.origin.x + (w - Bbox.right + Bbox.left) / 2;
			gdouble newY = docwin->osBuffer.origin.y + (h - Bbox.bottom + Bbox.top) / 2;
			TransformationMatrix tm;
			tm.setTranslation(newX - Bbox.left, newY - Bbox.top);
			content->transformByMatrixRelative(tm);
			//g_printerr("Moving target by %f, %f\n", newX - Bbox.left, newY - Bbox.top);
		}
		AddAction *newAction = new AddAction(content);
		doc->setAction(newAction);
		/* insert the gradients if there are some into this document */
		GDResourceList resList;
		content->gatherResource(resList, GDRESOURCE_TYPE_GRADIENT | GDRESOURCE_TYPE_PATTERN);
		for (GDResourceListIterator it = resList.begin(); it != resList.end(); it++) {
			if ((*it)->getType() == GDRESOURCE_TYPE_GRADIENT) {
				Gradient *gradient = (Gradient *) (*it);
				GradientMap::iterator it;
				if ((it = doc->gradients.find(gradient->getName())) != doc->gradients.end()) {
					if (!(*((*it).second) == *gradient)) {
						// there is a gradient with the same name but different content in this document
						// rename the old one (to avoid naming conflicts during paste
						uint count = 1;
						gchar number[12];
						snprintf(number, 11, "-%d", count);
						string newname = gradient->getName() + string(number);
						while (doc->gradients.find(newname) != doc->gradients.end()) {
							count++;
							snprintf(number, 11, "-%d", count);
							newname = gradient->getName() + string(number);
						}
						doc->gradients[newname] = (*it).second;
						doc->gradients.erase(it);
						Gradient *gradCopy = new Gradient(*gradient);
						doc->gradients[gradCopy->getName()] = gradCopy;
					}
				} else {
					// no gradient with that name already exists, insert a copy into this document
					Gradient *gradCopy = new Gradient(*gradient);
					doc->gradients[gradCopy->getName()] = gradCopy;
				}
			} else if ((*it)->getType() == GDRESOURCE_TYPE_PATTERN) {
				Pattern *pattern = (Pattern *) (*it);
				PatternMap::iterator it;
				if ((it = doc->patterns.find(pattern->getName())) != doc->patterns.end()) {
					if (!(*((*it).second) == *pattern)) {
						// there is a pattern with the same name but different content in this document
						// rename the old one (to avoid naming conflicts during paste)
						uint count = 1;
						gchar number[12];
						snprintf(number, 11, "-%d", count);
						string newname = pattern->getName() + string(number);
						while (doc->patterns.find(newname) != doc->patterns.end()) {
							count++;
							snprintf(number, 11, "-%d", count);
							newname = pattern->getName() + string(number);
						}
						doc->patterns[newname] = (*it).second;
						doc->patterns.erase(it);
						Pattern *pattCopy = new Pattern(*pattern);
						pattCopy->setDocument(doc);
						doc->patterns[pattCopy->getName()] = pattCopy;
					}
				} else {
					// no gradient with that name already exists, insert a copy into this document
					Pattern *pattCopy = new Pattern(*pattern);
					pattCopy->setDocument(doc);
					doc->patterns[pattCopy->getName()] = pattCopy;
				}
			}
		}
		/* insert the current clipboard content into the selected document */
		doc->deSelectAll();
		content->insertMoving(doc);
		/* recopy the previously pasted content to the clipboard again */
		copy(doc);
	}
	/* redraw the document's windows */
	doc->redraw();
}

gboolean ClipBoard::empty() {
	//g_printerr("ClipBoard::empty: content=%lX\n", content);
	return content == NULL;
}

ClipBoard::~ClipBoard() {
	delete content;
}
