/* $Id: FontCache.cc,v 1.3 2005/05/07 17:39:32 pklaus Exp $
 *
 * FontCache.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "FontCache.hh"
#include "Preferences.hh"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>

class FontCacheXMLBoolean : public XMLOBJECT {
	
public:
	FontCacheXMLBoolean(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, gboolean *val) :
			XMLOBJECT(p, tag, parser, attrs) {
		string value = getAttributeValue(attrs, "value");
		if (value == "true") {
			*val = true;
		} else {
			*val = false;
		}
	}
};

class FontCacheXMLString : public XMLOBJECT {

	string *value;
	string str;
	
public:
	FontCacheXMLString(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, string *val) :
			XMLOBJECT(p, tag, parser, attrs) {
		value = val;
	}
	void handleText(const char *text, int len) {
		str.append(text, len);
	}
	void handleEndTag(const char *name) {
		if (str.size()) {
			*value = str;
		}
		XMLOBJECT::handleEndTag(name);
	}
};

class FontCacheXMLEntry : public XMLOBJECT {
	FontCacheXMLString *str;
	FontCacheXMLBoolean *fcbool;
	FontCache *fcache;
	FontCacheEntry *fentry;
	string path;
	
public:
	FontCacheXMLEntry(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, FontCache *fc) :
			XMLOBJECT(p, tag, parser, attrs) {
		fcache = fc;
		fentry = new FontCacheEntry();
		path = getAttributeValue(attrs, "path");
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "family")) {
			str = new FontCacheXMLString(this, name, parser, attrs, &(fentry->familyName));
			str->init();
		} else if (! strcmp(name, "print-name")) {
			str = new FontCacheXMLString(this, name, parser, attrs, &(fentry->printName));
			str->init();
		} else if (! strcmp(name, "full-name")) {
			str = new FontCacheXMLString(this, name, parser, attrs, &(fentry->fullName));
			str->init();
		} else if (! strcmp(name, "weight")) {
			str = new FontCacheXMLString(this, name, parser, attrs, &(fentry->weight));
			str->init();
		} else if (! strcmp(name, "slant")) {
			str = new FontCacheXMLString(this, name, parser, attrs, &(fentry->slant));
			str->init();
		} else if (! strcmp(name, "fixed-pitch")) {
			fcbool = new FontCacheXMLBoolean(this, name, parser, attrs, &(fentry->isFixedPitch));
			fcbool->init();
		}
	}
	void handleEndTag(const char *name) {
		fcache->fontCacheMap[path] = fentry;
		XMLOBJECT::handleEndTag(name);
	}
};

class FontCacheXMLDocument : public XMLOBJECT {
	FontCacheXMLEntry *entry;
	FontCache *fcache;
public:
	FontCacheXMLDocument(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, FontCache *fc) :
			XMLOBJECT(p, tag, parser, attrs) {
		fcache = fc;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "entry")) {
			entry = new FontCacheXMLEntry(this, name, parser, attrs, fcache);
			entry->init();
		}
	}
};


class FontCacheXMLRoot : public XMLOBJECT {
	FontCacheXMLDocument *document;
	FontCache *fcache;
	
public:
	FontCacheXMLRoot(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, FontCache *fc) :
			XMLOBJECT(p, tag, parser, attrs) {
		fcache = fc;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "font-cache")) {
			char *savedversion = getAttributeValue(attrs, "version");
			// check for a current version of the fontcache, if version changed regenerate cache
			if (savedversion != NULL) {
				fcache->savedversion = string(savedversion);
			}
			document = new FontCacheXMLDocument(this, name, parser, attrs, fcache);
			document->init();
		}
	}
};

FontCache::FontCache() {
}

void FontCache::save(Preferences *prefs) {
	string cacheFile = prefs->getPrefDirectory();
	cacheFile.append("/fontcache");
	ofstream cacheOut(cacheFile.c_str());
	cacheOut.setf(ios::fixed, ios::floatfield);
	cacheOut.precision(12);
	cacheOut << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
	cacheOut << "<font-cache version=\"" << FONTCACHE_VERSION << "\">\n";
	for (FCache::iterator it = fontCacheMap.begin(); it != fontCacheMap.end(); it++) {
		cacheOut << "<entry path=\"" << (*it).first << "\">\n";
		cacheOut << "<family>" << (*it).second->familyName << "</family>\n";
		cacheOut << "<full-name>" << (*it).second->fullName << "</full-name>\n";
		cacheOut << "<print-name>" << (*it).second->printName << "</print-name>\n";
		cacheOut << "<weight>" << (*it).second->weight << "</weight>\n";
		cacheOut << "<slant>" << (*it).second->slant << "</slant>\n";
		cacheOut << "<fixed-pitch value=\"" << ((*it).second->isFixedPitch ? "true" : "false") << "\"/>\n";
		cacheOut << "</entry>\n";
	}
	cacheOut << "</font-cache>\n";
}

void FontCache::load(Preferences *prefs) {
	string cacheFile = prefs->getPrefDirectory();
	cacheFile.append("/fontcache");
	struct stat statbuf;
	int ret = stat(cacheFile.c_str(), &statbuf);
	if (ret) {
		//g_printerr("No fontcache file, generating one\n");
		save(prefs);
	}
	fontCacheMap.clear();
	XMLPARSER *xmlParser = new XMLPARSER();
	FontCacheXMLRoot *pxmlr = new FontCacheXMLRoot(NULL, "font-cache", NULL, NULL, this);
	xmlParser->parse(pxmlr, cacheFile);
	if (savedversion.compare(FONTCACHE_VERSION)) {
		fontCacheMap.clear();
	}
}

FontCache::~FontCache() {
	// free all entries in the map
	for (FCache::iterator it = fontCacheMap.begin(); it != fontCacheMap.end(); it++) {
		delete (*it).second;
	}
}
