/* $Id: GDDocument.cc,v 1.46 2005/05/07 17:39:32 pklaus Exp $
 *
 * GDDocument.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "GDDocument.hh"
#include "DocWindow.hh"
#include "GDApplication.hh"
#include "BezierNode.hh"
#include "BezierPath.hh"
#include "PixelImage.hh"
#include "DocumentMenu.hh"
#include "DocWindow.hh"
#include "SelectionItemList.hh"
#include "JoinedPointDialog.hh"
#include "PrintingDialog.hh"
#include "Layers.hh"
#include "GradientListOrgDialog.hh"
#include "PatternListOrgDialog.hh"
#include "Properties.hh"
#include "TextProperties.hh"
#include "FontPool.hh"
#include "Layer.hh"
#include "Undo.hh"
#include "DeleteAction.hh"
#include "JoinAction.hh"
#include "MoveDialog.hh"
#include "Gradient.hh"
#include "Pattern.hh"
#include "CustomColor.hh"
#include "AlignDistribute.hh"
#include <gdk--.h>
#include "gestalter.h"
#include <gnome--.h>
#include <libgnome/libgnome.h>
#include <fstream>
#include <sstream>
#ifdef GEST_CXX_EXT_STDIO_FILEBUF
#include <ext/stdio_filebuf.h>
#endif
#include <math.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <libart_lgpl/art_misc.h>
#include <libart_lgpl/art_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_svp_vpath_stroke.h>
#include <libart_lgpl/art_rgb_svp.h>
#include <libart_lgpl/art_vpath_dash.h>
#include "pixmaps/stipple_dither.xbm"

GDDocument::GDDocument(GDApplication *gda) {
	layers = NULL;
	layerNumber = 0;
	deleteAfterSave = false;
	application = gda;
	Gdk_Window win = application->toolWindow->get_window();
	docWindowList = new G_List<DocWindow *>();
	undo = new Undo(this);
	selection = new SelectionList(this);
	layers = new Layer(this);
	activeLayer = layers;
	stippleDither = Gdk_Pixmap(win, (gchar *)stipple_dither_bits, stipple_dither_width, stipple_dither_height, 1, Gdk_Color(string("black")), Gdk_Color(string("white")));
	pageWidth = application->preferences.pageWidth;
	pageHeight = application->preferences.pageHeight;
	printWidth = application->preferences.pageWidth - 10;
	printHeight = application->preferences.pageHeight - 10;
	pageFormat = application->preferences.pageFormat;
	printOffset.x = 5;
	printOffset.y = 5;
	guidesLocked = true;
	//g_printerr("GDDocument::GDDocument pw: %f ph: %f\n", pageWidth, pageHeight);
	nullPoint.x = (DOCUMENT_SIZE_X - pageWidth) / 2;
	nullPoint.y = (DOCUMENT_SIZE_Y - pageHeight) / 2;
	printFactor.x = 1;
	printFactor.y = 1;
	context = Gdk_GC(win);
	currNode = NULL;
	currPath = NULL;
	fileselection = NULL;
	placeFileSelection = NULL;
	joinedPointDialog = NULL;
	printDialog = NULL;
	newDocWindow();
	currentTextStyle.setFont(application->fontPool->getFont("Helvetica"));
	if (! currentTextStyle.getFont()) {
		g_printerr("Couldn't find any known font for default text style. Exiting.");
		exit(1);
	}
	//g_printerr("GDDocument::GDDocument %lX\n", this);
}

void GDDocument::setFilePath(string filePath) {
	this->filePath = filePath;
	updateWindowTitle();
}

void GDDocument::updateWindowTitle() {
	G_List_Iterator<DocWindow *> winit = docWindowList->begin();
	DocWindow *win;
	while ((win = *winit++)) {
		ostringstream wintit;
		wintit << (undo->getSaved() ? "" : "* ") << (filePath.size() ? filePath : _("untitled"));
		win->setWindowTitle(wintit.str());
	}
}

void GDDocument::deleteSelectionItems() {
	//g_printerr("GDDocument::deleteSelectionItems\n");
	DeleteAction *newAction = new DeleteAction();
	setAction(newAction);
	{
		// sub context to delete the SelectionItemList before redrawing the document
		SelectionItemList selItems;
		getSelection(selItems);
		selItems.clearOriginal();
	}
	// reset currPath and currNode
	currPath = NULL;
	currNode = NULL;
	// recalc the selection
	recalcSelection();
	//g_printerr("GDDocument::deleteSelectionItems calling draw\n");
	redraw();
}

void GDDocument::addLayer(Layer *layer) {
	if (layers) {
		layers->insert(NULL, layer);
	} else {
		layers = layer;
	}
	layersChanged(this);
}

void GDDocument::removeLayer(Layer *layer) {
	if (layers) {
		layers = layers->remove(layer);
	}
	if (! layers) {
		// ensure at least one layer always exists
		layers = new Layer(this);
		activeLayer = layers;
	}
	layersChanged(this);
}

void GDDocument::addPath(BezierPath *path) {
	currPath = path;
	activeLayer->addObject(path);
}

void GDDocument::removeObject(GObject *object) {
	Layer *layer = layers;
	while (layer && ! layer->removeObject(object)) {
		layer = layer->next;
	}
	selection->removeObject(object);
}

Layer *GDDocument::findObjectLayer(GObject *object) {
	Layer *layer = layers;
	while (layer) {
		if (layer->findObjectLayer(object)) {
			return layer;
		}
		layer = layer->next;
	}
	return NULL;
}

guint GDDocument::getObjectNumber(GObject *object) {
	guint number = 0;
	Layer *layer = layers;
	while (layer) {
		if (layer->getObjectNumber(object, number)) {
			return number;
		}
		layer = layer->next;
	}
	return 0;
}

DocWindow *GDDocument::newDocWindow() {
	DocWindow *docWin = new DocWindow(this);
	docWindowList->insert(docWindowList->end(), docWin);
	RectangleArea rect = docWin->getContentRectangle();
	draw(docWin, rect);
	return docWin;
}

gint GDDocument::closeWindow(DocWindow *docWin) {
	if (docWindowList->size() == 1 && ! undo->getSaved()) {
		Gnome::MessageBox *mBox = new Gnome::MessageBox(_("The document has changed since it was saved. Do you want to save the document before closing?"), GNOME_MESSAGE_BOX_QUESTION);
		mBox->append_button(GNOME_STOCK_BUTTON_YES);
		mBox->append_button(GNOME_STOCK_BUTTON_NO);
		if (! application->doQuit) {
			mBox->append_button(GNOME_STOCK_BUTTON_CANCEL);
		}
		gint ret = mBox->run();
		delete mBox;
		//g_printerr("GDDocument::closeWindow MessageBox returned %d\n", ret);
		if (ret == 2) {
			return -2;
		} else if (ret == 0) {
			save();
		}
	}
	docWindowList->remove(docWin);
	application->docMenu->setWindow(NULL);
	application->properties->setDocument(NULL);
	application->textProperties->setDocument(NULL);
	application->layers->setDocument(NULL);
	application->gradients->setDocument(NULL);
	application->patterns->setDocument(NULL);
	//application->customColors->setDocument(NULL);
	application->alignDistribute->setDocument(NULL);
	if (docWindowList->size() < 1) {
		GDApplication *app = application;
		app->removeDocument(this);
		delete this;
	}
	return 0;
}

gint GDDocument::closeDocument() {
	G_List_Iterator<DocWindow *> winit = docWindowList->begin();
	DocWindow *win;
	gint ret = 0;
	while ((win = *winit++)) {
		ret = closeWindow(win);
		if (! ret) {
			delete win;
		}
	}
	return ret;
}

void GDDocument::redraw(RectangleArea& drawArea) {
	redraw();
}

void GDDocument::redraw() {
	G_List_Iterator<DocWindow *> winit = docWindowList->begin();
	DocWindow *win;
	while ((win = *winit++)) {
		redraw(win);
	}
}

void GDDocument::redraw(DocWindow *docwin) {
	docwin->needRedraw = true;
	docwin->DocWinDrawingArea->queue_draw();
}

void GDDocument::redraw(DocWindow *docwin, RectangleArea& drawArea) {
	redrawRect.combine(drawArea);
	redraw(docwin);
}

void GDDocument::draw(RectangleArea& drawArea) {
	//g_printerr("GDDocument::draw()\n");
	G_List_Iterator<DocWindow *> winit = docWindowList->begin();
	DocWindow *win;
	while ((win = *winit++)) {
		//win->DocWinDrawingArea->get_window().clear();
		draw(win, drawArea);
	}
}

void GDDocument::draw(DocWindow *docwin, RectangleArea& drawArea) {
	RectangleArea winRect = docwin->getContentRectangle();
	winRect.do_intersect(drawArea);
	//g_printerr("GDDocument::draw(docwin)\n");
	if (docwin->displayWireFrame) {
		Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
		Gdk_GC blackgc = docwin->DocWinDrawingArea->get_style()->get_black_gc();
		Gdk_Color white = docwin->DocWinDrawingArea->get_style()->gtkobj()->white;
		Gdk_Colormap cmap = Gdk_Colormap::get_system();
		drawable.set_background(white);
		drawable.clear();
		Layer *layerList = layers;
		while (layerList) {
			//g_printerr("GDDocument::draw(docwin) draw layer wired %lX\n", layerList);
			layerList->drawWireFrame(docwin, blackgc, winRect);
			layerList = layerList->next;
		}
		selection->drawWireFrame(docwin, winRect);
		drawWireFramePageRectangle(docwin, winRect);
	} else {
		docwin->allocPixBuffer();
		docwin->whiteBackground();
		//drawPageRectangle(docwin->osBuffer, winRect);
		Layer *layerList = layers;
		while (layerList) {
			//g_printerr("GDDocument::draw(docwin) draw layer %lX\n", layerList);
			layerList->draw(docwin->osBuffer, winRect);
			layerList = layerList->next;
		}
		selection->draw(docwin, winRect);
		docwin->drawPixmap();
		drawWireFramePageRectangle(docwin, winRect);
	}
}

void GDDocument::drawPageRectangle(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	// draw first the page boundings
	ArtVpath apath[6];
	apath[0].code = ART_MOVETO;
	apath[0].x = (nullPoint.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	apath[0].y = (nullPoint.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	apath[1].code = ART_LINETO;
	apath[1].x = (nullPoint.x + pageWidth - osBuffer.origin.x) * osBuffer.scaleFactor;
	apath[1].y = (nullPoint.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	apath[2].code = ART_LINETO;
	apath[2].x = (nullPoint.x + pageWidth - osBuffer.origin.x) * osBuffer.scaleFactor;
	apath[2].y = (nullPoint.y + pageHeight - osBuffer.origin.y) * osBuffer.scaleFactor;
	apath[3].code = ART_LINETO;
	apath[3].x = (nullPoint.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	apath[3].y = (nullPoint.y + pageHeight - osBuffer.origin.y) * osBuffer.scaleFactor;
	apath[4].code = ART_LINETO;
	apath[4].x = (nullPoint.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	apath[4].y = (nullPoint.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	apath[5].code = ART_END;
	apath[5].x = 0;
	apath[5].y = 0;
	ArtSVP *asvp = art_svp_vpath_stroke(apath, 
								ART_PATH_STROKE_JOIN_MITER,
								ART_PATH_STROKE_CAP_BUTT,
								1, 4, 0.5);
	art_rgb_svp_alpha(asvp, 0, 0,
					osBuffer.pxbWidth,
					osBuffer.pxbHeight,
					0x444444FF,
					osBuffer.pixbuffer,
					osBuffer.pxbRowstride,
					NULL);
	art_svp_free(asvp);
	// then the printing rectangle
	ArtVpath papath[6];
	papath[0].code = ART_MOVETO;
	papath[0].x = (nullPoint.x + printOffset.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	papath[0].y = (nullPoint.y + printOffset.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	papath[1].code = ART_LINETO;
	papath[1].x = (nullPoint.x + printOffset.x + printWidth - osBuffer.origin.x) * osBuffer.scaleFactor;
	papath[1].y = (nullPoint.y + printOffset.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	papath[2].code = ART_LINETO;
	papath[2].x = (nullPoint.x + printOffset.x + printWidth - osBuffer.origin.x) * osBuffer.scaleFactor;
	papath[2].y = (nullPoint.y + printOffset.y + printHeight - osBuffer.origin.y) * osBuffer.scaleFactor;
	papath[3].code = ART_LINETO;
	papath[3].x = (nullPoint.x + printOffset.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	papath[3].y = (nullPoint.y + printOffset.y + printHeight - osBuffer.origin.y) * osBuffer.scaleFactor;
	papath[4].code = ART_LINETO;
	papath[4].x = (nullPoint.x + printOffset.x - osBuffer.origin.x) * osBuffer.scaleFactor;
	papath[4].y = (nullPoint.y + printOffset.y - osBuffer.origin.y) * osBuffer.scaleFactor;
	papath[5].code = ART_END;
	papath[5].x = 0;
	papath[5].y = 0;
	ArtVpathDash dash;
	gdouble dashes[2];
	dash.dash = dashes;
	dash.n_dash = 2;
	dash.offset = 0;
	dashes[0] = 4;
	dashes[1] = 2;
	ArtVpath *avpath = art_vpath_dash(papath, &dash);
	ArtSVP *pasvp = art_svp_vpath_stroke(avpath, 
								ART_PATH_STROKE_JOIN_MITER,
								ART_PATH_STROKE_CAP_BUTT,
								1, 4, 0.5);
	art_free(avpath);
	art_rgb_svp_alpha(pasvp, 0, 0,
					osBuffer.pxbWidth,
					osBuffer.pxbHeight,
					0xCCCCCCFF,
					osBuffer.pixbuffer,
					osBuffer.pxbRowstride,
					NULL);
	art_svp_free(pasvp);
}	

void GDDocument::drawWireFramePageRectangle(DocWindow *docwin, RectangleArea& drawArea) {
	Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
	Gdk_Window win = application->toolWindow->get_window();
	Gdk_GC drawContext(win);
	guint x = (guint) ((nullPoint.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor);
	guint y = (guint) ((nullPoint.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor);
	guint w = (guint) (pageWidth * docwin->osBuffer.scaleFactor);
	guint h = (guint) (pageHeight * docwin->osBuffer.scaleFactor);
	drawable.draw_rectangle(drawContext, 0, x, y, w, h);
	x = (guint) ((nullPoint.x + printOffset.x - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor);
	y = (guint) ((nullPoint.y + printOffset.y - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor);
	w = (guint) (printWidth * docwin->osBuffer.scaleFactor);
	h = (guint) (printHeight * docwin->osBuffer.scaleFactor);
	drawContext.set_dashes(2);
	drawContext.set_line_style(GDK_LINE_ON_OFF_DASH);
	drawable.draw_rectangle(drawContext, 0, x, y, w, h);
	drawContext.set_line_style(GDK_LINE_SOLID);
}

BoundingBox GDDocument::getBoundingBox() {
	BoundingBox bbox;
	Layer *layer = layers;
	while (layer) {
		BoundingBox bb = layer->getBoundingBox();
		bbox.combine(bb);
		layer = layer->next;
	}
	return bbox;
}

gdouble GDDocument::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = G_MAXDOUBLE;
	gdouble temp_ax = ax, temp_ay = ay;
	Layer *layer = layers;
	while (layer) {
		gdouble res;
		if ((res = layer->attractPoint(x, y, temp_ax, temp_ay, tolerance, attractType)) < nearest) {
			nearest = res;
			ax = temp_ax;
			ay = temp_ay;
		}
		layer = layer->next;
	}
	return nearest;
}


SelectableItem *GDDocument::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	SelectableItem *found = NULL;
	Layer *layer = layers;
	while (layer) {
		SelectableItem *candidate;
		if ((candidate = layer->checkPoints(x, y, tolerance, onlyCurvePoints)) != NULL) {
			found = candidate;
		}
		layer = layer->next;
	}
	return found;
}

SelectableItem *GDDocument::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	SelectableItem *found = NULL;
	Layer *layer = layers;
	while (layer) {
		SelectableItem *candidate;
		if ((candidate = layer->checkCurves(x, y, tolerance)) != NULL) {
			found = candidate;
		}
		layer = layer->next;
	}
	return found;
}

GObject *GDDocument::checkArea(gdouble x, gdouble y) {
	GObject *found = NULL;
	Layer *layer = layers;
	while (layer) {
		GObject *candidate;
		if ((candidate = layer->checkArea(x, y)) != NULL) {
			found = candidate;
		}
		layer = layer->next;
	}
	return found;
}

void GDDocument::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	Layer *layer = layers;
	while (layer) {
		layer->selectRectangle(left, top, right, bottom, what);
		layer = layer->next;
	}
}

void GDDocument::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	Layer *layer = layers;
	while (layer) {
		layer->deSelectRectangle(left, top, right, bottom, what);
		layer = layer->next;
	}
}

void GDDocument::selectAll() {
	Layer *layer = layers;
	while (layer) {
		layer->selectAll();
		layer = layer->next;
	}
}

void GDDocument::deSelectAll() {
	Layer *layer = layers;
	while (layer) {
		layer->deSelectAll();
		layer = layer->next;
	}
}

void GDDocument::getSelection(SelectionItemList& selectionList, gint mode) {
	Layer *layer = layers;
	while (layer) {
		layer->getSelection(selectionList, mode);
		layer = layer->next;
	}
}

void GDDocument::recalcSelection() {
	selection->reset();
	SelectionItemList currSelection;
	getSelection(currSelection, SELECTIONITEM_MODE_ORIG_ONLY);
	currSelection.reSelect();
	// set currPath and currNode to NULL because the objects they pointed to possibly don't exist anymore
	currPath = NULL;
	currNode = NULL;
}

void GDDocument::gatherResource(GDResourceList& collection, gint type) {
	Layer *layer = layers;
	while (layer) {
		layer->gatherResource(collection, type);
		layer = layer->next;
	}
}

void GDDocument::onPlaceImage() {
	gchar* title = _("Place pixel image:");
	if (placeFileSelection == NULL) {
		placeFileSelection = new Gtk::FileSelection(Gtk::nstring(title));
		placeFileSelection->get_ok_button()->clicked.connect(SigC::slot(this, &GDDocument::placeImage));
		placeFileSelection->get_cancel_button()->clicked.connect(SigC::slot(this, &GDDocument::cancelPlaceImageDialog));
		placeFileSelection->delete_event.connect(SigC::slot(this, &GDDocument::deletePlaceImageDialog));
		placeFileSelection->set_modal(true);
	}
	placeFileSelection->show();
}

gint GDDocument::deletePlaceImageDialog(GdkEventAny *ev) {
	cancelPlaceImageDialog();
}

void GDDocument::cancelPlaceImageDialog() {
	delete placeFileSelection;
	placeFileSelection = NULL;
}

void GDDocument::placeImage() {
	deSelectAll();
	if (placeFileSelection) {
		string filep = placeFileSelection->get_filename();
		//g_printerr("GDDocument::placeImage file:%s\n", filep.c_str());
		PixelImage *pimg = new PixelImage(filep, this);
		cancelPlaceImageDialog();
		activeLayer->addObject(pimg);
		DocWindow *dw = application->docMenu->menuDocWindow;
		if (dw) {
			gdouble x = dw->osBuffer.origin.x + 
						dw->DocWinDrawingArea->width() * dw->osBuffer.scaleFactor / 2 -
						pimg->width() / 2;
			gdouble y = dw->osBuffer.origin.y + 
						dw->DocWinDrawingArea->height() * dw->osBuffer.scaleFactor / 2 -
						pimg->height() / 2;
			TransformationMatrix tm;
			tm.setTranslation(x, y);
			pimg->transformByMatrix(tm);
		}
		redraw();
	}
}

void GDDocument::showPrintDialog() {
	if (! printDialog) {
		printDialog = new PrintingDialog();
		printDialog->getPrintButton()->clicked.connect(SigC::slot(this, &GDDocument::doPrint));
		printDialog->getCancelButton()->clicked.connect(SigC::slot(this, &GDDocument::cancelPrintDialog));
		printDialog->delete_event.connect(SigC::slot(this, &GDDocument::deletePrintDialog));
		printDialog->set_modal(true);
	}
	printDialog->show();
}

gint GDDocument::deletePrintDialog(GdkEventAny *ev) {
	cancelPrintDialog();
}

void GDDocument::cancelPrintDialog() {
	//g_printerr("GDDocument::cancelPrintDialog\n");
	delete printDialog;
	printDialog = NULL;
}

void GDDocument::doPrint() {
	if (printDialog->isFilePath()) {
		// write postscript code to a file
		string psFile = printDialog->getFilePath();
		//g_printerr("GDDocument::doPrint: printing to a file: %s\n", psFile.c_str());
		cancelPrintDialog();
		if (psFile.size() > 0) {
			ofstream psout(psFile.c_str());
			writePS(psout);
		}
	} else {
		// write postscript code into a pipe
		string cmd = printDialog->getPrintCommand();
		//g_printerr("GDDocument::doPrint: printing to a command: %s\n", cmd.c_str());
		cancelPrintDialog();
		if (cmd.size() < 1) {
			g_printerr("GDDocument::doPrint: no printing command specified\n");
			return;
		}
		int fds[2];
		if (pipe(fds) < 0) {
			g_printerr("GDDocument::doPrint: pipe() failed\n");
			return;
		}
		pid_t pid = fork();
		if (pid < 0) {
			g_printerr("GDDocument::doPrint: fork() failed\n");
			return;
		}
		if (pid == 0) {
			close(0);
			dup(fds[0]);
			close(fds[0]);
			close(fds[1]);
			const char *shell = getenv("SHELL");
			if (! shell) {
				shell = "/bin/sh";
			}
			const char *p, *q;
			for (p = q = shell; *p; p++) {
				if (*p == '/') q = p + 1;
			}
			execlp(shell, q, "-c", cmd.c_str(), (const char *) 0);
			g_printerr("GDDocument::doPrint: execution of %s failed\n", cmd.c_str());
			exit(5);
		}
		// pid > 0
		close(fds[0]);
#ifdef GEST_CXX_EXT_STDIO_FILEBUF
		__gnu_cxx::stdio_filebuf<char> fbuf(fds[1], ios_base::out, true, 4096);
		ostream psout(&fbuf);
#else
		ofstream psout(fds[1]);
#endif
		writePS(psout);
	}
}

void GDDocument::onSaveDocumentAs() {
	gchar* title = _("Save document as:");
	Gtk::nstring ntitle(title);
	
	if (fileselection == NULL) {
		fileselection = new Gtk::FileSelection(ntitle);
		fileselection->get_ok_button()->clicked.connect(SigC::slot(this, &GDDocument::saveAs));
		fileselection->get_cancel_button()->clicked.connect(SigC::slot(this, &GDDocument::cancelSaveAsDialog));
		fileselection->delete_event.connect(SigC::slot(this, &GDDocument::deleteSaveAsDialog));
		fileselection->set_modal(true);
	}
	fileselection->show();
	//g_printerr("GDDocument::onSaveDocumentAs calling GtkMain::run\n");
	Gtk::Main::run();
	//g_printerr("GDDocument::onSaveDocumentAs run returned\n");
}

gint GDDocument::deleteSaveAsDialog(GdkEventAny *ev) {
	cancelSaveAsDialog();
}

void GDDocument::cancelSaveAsDialog() {
	fileselection->hide();
	delete fileselection;
	fileselection = NULL;
	Gtk::Main::quit();
}

void GDDocument::saveAs() {
	if (fileselection) {
		setFilePath(fileselection->get_filename());
		fileselection->hide();
		delete fileselection;
		fileselection = NULL;
		if (filePath.size()) {
			save();
		}
	}
	Gtk::Main::quit();
}

gint GDDocument::save() {
	if (filePath.size() < 1) {
		onSaveDocumentAs();
	} else {
		ofstream xmlout(filePath.c_str());
		writeXML(xmlout);
		undo->setSaved();
		updateWindowTitle();
	}
	return 0;
}

gint GDDocument::writeXML(ostream& xmlout) {
	xmlout.setf(ios::fixed, ios::floatfield);
	xmlout.precision(12);
	xmlout << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
	xmlout << "<svg xmlns=\"http://www.w3.org/2000/svg\" version=\"1.0\" ";
	xmlout << "xmlns:gestalter=\"http://www.linotux.ch/gestalter/svgns\" ";
	xmlout << "gestalter:pageX=\"" << nullPoint.x <<"\" ";
	xmlout << "gestalter:pageY=\"" << nullPoint.y <<"\" ";
	xmlout << ">\n";
	GDResourceList resList;
	gatherResource(resList, GDRESOURCE_TYPE_FONT | GDRESOURCE_TYPE_GRADIENT | GDRESOURCE_TYPE_PATTERN);
	if (resList.size()) {
		xmlout << "<defs>\n";
	}
	for (GDResourceListIterator it = resList.begin(); it != resList.end(); it++) {
		//g_printerr("GDDocument::writeXML: resType=%d\n", (*it)->getType());
		if ((*it)->getType() == GDRESOURCE_TYPE_GRADIENT || (*it)->getType() == GDRESOURCE_TYPE_PATTERN) {
			(*it)->writeXML(xmlout);
			//g_printerr("wrote out\n");
		}
	}
	if (resList.size()) {
		xmlout << "</defs>\n";
	}
	Layer *layer = layers;
	while (layer) {
		layer->writeXML(xmlout);
		layer = layer->next;
	}
	xmlout << "</svg>\n";
	return 0;
}

gint GDDocument::writePS(ostream& psout) {
	// define the abreviations
	time_t timeval = time(NULL);
	char timestr[30];
	ctime_r(&timeval, timestr);
	char *nl;
	if ((nl = strchr(timestr, '\n'))) {
		*nl = '\0';
	}
	psout << "%!PS-Adobe-3.0\n";
	psout << "%%Title: (" << filePath << ")\n";
	psout << "%%Version: 1.0\n";
	psout << "%%Creator: (Gestalter)\n";
	psout << "%%CreationDate: (" << timestr << ")\n";
	psout << "%%DocumentData: Clean7Bit\n";
	psout << "%%LanguageLevel: 2\n";
	psout << "%%BoundingBox: 0 0 595 842\n"; // Papersize
	psout << "%%Pages: 1 0\n";
	psout << "%%DocumentProcessColors: (atend)\n";
	psout << "%%DocumentNeededResources: (atend)\n";
	psout << "%%DocumentSuppliedResources: (atend)\n";
	psout << "%%EndComments\n";
	psout << "%%BeginDefaults\n";
	psout << "%%EndDefaults\n";
	psout << "%%BeginProlog\n";
	psout << "% printing document " << filePath << endl;
	psout << "%%BeginResource: procset iso-reencoding-gestalter-procs" << endl;
	psout << "/reencdict 12 dict def /ReEncode { reencdict begin" << endl;
	psout << "/newcodesandnames exch def /newfontname exch def /basefontname exch def" << endl;
	psout << "/basefontdict basefontname findfont def /newfont basefontdict maxlength dict def" << endl;
	psout << "basefontdict { exch dup /FID ne { dup /Encoding eq" << endl;
	psout << "{ exch dup length array copy newfont 3 1 roll put }" << endl;
	psout << "{ exch newfont 3 1 roll put } ifelse } { pop pop } ifelse } forall" << endl;
	psout << "newfont /FontName newfontname put newcodesandnames aload pop" << endl;
	psout << "128 1 255 { newfont /Encoding get exch /.notdef put } for" << endl;
	psout << "newcodesandnames length 2 idiv { newfont /Encoding get 3 1 roll put } repeat" << endl;
	psout << "newfontname newfont definefont pop end } def" << endl;
	psout << "/isovec [" << endl;
	psout << "8#055 /minus 8#200 /grave 8#201 /acute 8#202 /circumflex 8#203 /tilde" << endl;
	psout << "8#204 /macron 8#205 /breve 8#206 /dotaccent 8#207 /dieresis" << endl;
	psout << "8#210 /ring 8#211 /cedilla 8#212 /hungarumlaut 8#213 /ogonek 8#214 /caron" << endl;
	psout << "8#220 /dotlessi 8#230 /oe 8#231 /OE" << endl;
	psout << "8#240 /space 8#241 /exclamdown 8#242 /cent 8#243 /sterling" << endl;
	psout << "8#244 /currency 8#245 /yen 8#246 /brokenbar 8#247 /section 8#250 /dieresis" << endl;
	psout << "8#251 /copyright 8#252 /ordfeminine 8#253 /guillemotleft 8#254 /logicalnot" << endl;
	psout << "8#255 /hyphen 8#256 /registered 8#257 /macron 8#260 /degree 8#261 /plusminus" << endl;
	psout << "8#262 /twosuperior 8#263 /threesuperior 8#264 /acute 8#265 /mu 8#266 /paragraph" << endl;
	psout << "8#267 /periodcentered 8#270 /cedilla 8#271 /onesuperior 8#272 /ordmasculine" << endl;
	psout << "8#273 /guillemotright 8#274 /onequarter 8#275 /onehalf" << endl;
	psout << "8#276 /threequarters 8#277 /questiondown 8#300 /Agrave 8#301 /Aacute" << endl;
	psout << "8#302 /Acircumflex 8#303 /Atilde 8#304 /Adieresis 8#305 /Aring" << endl;
	psout << "8#306 /AE 8#307 /Ccedilla 8#310 /Egrave 8#311 /Eacute" << endl;
	psout << "8#312 /Ecircumflex 8#313 /Edieresis 8#314 /Igrave 8#315 /Iacute" << endl;
	psout << "8#316 /Icircumflex 8#317 /Idieresis 8#320 /Eth 8#321 /Ntilde 8#322 /Ograve" << endl;
	psout << "8#323 /Oacute 8#324 /Ocircumflex 8#325 /Otilde 8#326 /Odieresis 8#327 /multiply" << endl;
	psout << "8#330 /Oslash 8#331 /Ugrave 8#332 /Uacute 8#333 /Ucircumflex" << endl;
	psout << "8#334 /Udieresis 8#335 /Yacute 8#336 /Thorn 8#337 /germandbls 8#340 /agrave" << endl;
	psout << "8#341 /aacute 8#342 /acircumflex 8#343 /atilde 8#344 /adieresis 8#345 /aring" << endl;
	psout << "8#346 /ae 8#347 /ccedilla 8#350 /egrave 8#351 /eacute" << endl;
	psout << "8#352 /ecircumflex 8#353 /edieresis 8#354 /igrave 8#355 /iacute" << endl;
	psout << "8#356 /icircumflex 8#357 /idieresis 8#360 /eth 8#361 /ntilde 8#362 /ograve" << endl;
	psout << "8#363 /oacute 8#364 /ocircumflex 8#365 /otilde 8#366 /odieresis 8#367 /divide" << endl;
	psout << "8#370 /oslash 8#371 /ugrave 8#372 /uacute 8#373 /ucircumflex" << endl;
	psout << "8#374 /udieresis 8#375 /yacute 8#376 /thorn 8#377 /ydieresis] def" << endl;
	psout << "%%EndResource\n";
	psout << "%%BeginResource: procset gestalter-print-procs-0.2\n";
	psout << "/c {curveto} bind def\n";
	psout << "/m {moveto} bind def\n";
	psout << "/f {fill} bind def\n";
	psout << "/s {stroke} bind def\n";
	psout << "/cp {closepath} bind def\n";
	psout << "/np {newpath} bind def\n";
	psout << "/sc {setrgbcolor} bind def\n";
	psout << "/gs {gsave} bind def\n";
	psout << "/gr {grestore} bind def\n";
	psout << "/slw {setlinewidth} bind def\n";
	psout << "/sh {show} bind def\n";
	psout << "/rm {rmoveto} bind def\n";
	psout << "/sl {scale} bind def\n";
	psout << "/ffsf {exch findfont exch scalefont setfont} bind def\n";
	psout << "/ts {show} bind def\n";
	psout << "/fcp {false charpath} bind def\n";
	psout << "/cup {currentpoint} bind def\n";
	psout << "/sd {setdash} bind def\n";
	psout << "/slc {setlinecap} bind def\n";
	psout << "/slj {setlinejoin} bind def\n";
	psout << "/sml {setmiterlimit} bind def\n";
	psout << "/mpt {makepattern} bind def\n";
	psout << "/spt {setpattern} bind def\n";
	psout << "/r {roll} bind def\n";
	psout << "/fdraw {setrgbcolor show} bind def\n";
	psout << "/fgdraw {makepattern setpattern show} bind def\n";
	psout << "/fsdraw {setrgbcolor currentpoint 12 -1 roll dup show currentpoint 5 -2 roll newpath moveto 3 -1 roll false charpath 11 2 roll setrgbcolor setlinewidth setmiterlimit setlinejoin setlinecap setdash stroke moveto} bind def\n";
	psout << "/fgsdraw {makepattern setpattern currentpoint 12 -1 roll dup show currentpoint 5 -2 roll newpath moveto 3 -1 roll false charpath 11 2 roll setrgbcolor setlinewidth setmiterlimit setlinejoin setlinecap setdash stroke moveto} bind def\n";
	psout << "/sdraw {currentpoint 12 2 roll 10 -1 roll dup currentpoint newpath moveto false charpath 12 1 roll setrgbcolor setlinewidth setmiterlimit setlinejoin setlinecap setdash stroke moveto stringwidth rmoveto} bind def\n";
	psout << "/mcjust {stringwidth pop add 2 div neg 0 rmoveto} def\n";
	psout << "/mrjust {stringwidth pop add neg 0 rmoveto} def\n";
	psout << "%%EndResource\n";
	psout << "%%EndProlog\n";
	psout << "%%BeginSetup\n";
	GDResourceList resList;
	gatherResource(resList, GDRESOURCE_TYPE_FONT | GDRESOURCE_TYPE_GRADIENT | GDRESOURCE_TYPE_PATTERN);
	psout << "% fonts\n";
	for (GDResourceListIterator it = resList.begin(); it != resList.end(); it++) {
		if ((*it)->getType() == GDRESOURCE_TYPE_FONT) {
			(*it)->writePS(psout);
		}
	}
	psout << "% gradients\n";
	for (GDResourceListIterator it = resList.begin(); it != resList.end(); it++) {
		if ((*it)->getType() == GDRESOURCE_TYPE_GRADIENT) {
			(*it)->writePS(psout);
		}
	}
	psout << "% patterns\n";
	for (GDResourceListIterator it = resList.begin(); it != resList.end(); it++) {
		if ((*it)->getType() == GDRESOURCE_TYPE_PATTERN) {
			(*it)->writePS(psout);
		}
	}
	
	psout << "%%EndSetup\n";
	psout << "%%Page: 1 1\n";
	psout << "%%BeginPageSetup\n";
	psout << "%%EndPageSetup\n";
	Layer *layer = layers;
	while (layer) {
		layer->writePS(psout);
		layer = layer->next;
	}
	psout << "showpage\n";
	psout << "%%PageTrailer\n";
	psout << "%%EndPage\n";
	psout << "%%Trailer\n";
	psout << "%%DocumentProcessColors: Cyan Magenta Yellow Black\n";
	psout << "%%DocumentNeededResources:\n";
	psout << "%%+ font Helvetica\n";
	psout << "%%DocumentSuppliedResources:\n";
	psout << "%%+ font Univers\n";
	psout << "%%EOF\n";
	return 0;
}

gint GDDocument::deleteJoinObjectDialog(GdkEventAny *ev) {
	cancelJoinObjectDialog();
}

void GDDocument::cancelJoinObjectDialog() {
	delete joinedPointDialog;
	joinedPointDialog = NULL;
}

void GDDocument::onJoinObjects() {
	joinObjectSelection = new SelectionItemList();
	getSelection(*joinObjectSelection);
	if (joinObjectSelection->size() != 2) {
		// exactly 2 items must be selected
		application->showError(_("Joining objects works only on two node points.\nPlease check that the selection contains exactly two node points."));
		//g_printerr("GDDocument::onJoinObjects Not exactly 2 items selected\n");
		return;
	}
	G_List_Iterator<SelectionItem *> it = joinObjectSelection->begin();
	SelectableItem *item1 = (*it++)->orig;
	SelectableItem *item2 = (*it)->orig;
	if (strcmp(item1->identify(), "BezierNode") ||
			strcmp(item2->identify(), "BezierNode")) {
		// items must be BezierNodes
		application->showError(_("Joining objects works only on two node points.\nPlease check that the selection contains exactly two node points."));
		//g_printerr("GDDocument::onJoinObjects 2 items not both BezierNodes\n");
		return;
	}
	BezierNode *node1 = (BezierNode *) item1;
	BezierNode *node2 = (BezierNode *) item2;
	if (!((node1->prev == NULL || node1->next == NULL) && 
						(node2->prev == NULL || node2->next == NULL))) {
		// not an open path
		application->showError(_("Joining objects works only on two open node points.\nYou must select the end points of a path."));
		//g_printerr("GDDocument::onJoinObjects at least one path is not open\n");
		return;
	}
	if (node1->onCurve.x == node2->onCurve.x && node1->onCurve.y == node2->onCurve.y) {
		//g_printerr("GDDocument::onJoinObjects items identical loc: show dialog\n");
		if (! joinedPointDialog) {
			joinedPointDialog = new JoinedPointDialog();
			joinedPointDialog->getOKButton()->clicked.connect(SigC::slot(this, &GDDocument::joinObjects));
			joinedPointDialog->getCancelButton()->clicked.connect(SigC::slot(this, &GDDocument::cancelJoinObjectDialog));
			joinedPointDialog->delete_event.connect(SigC::slot(this, &GDDocument::deleteJoinObjectDialog));
			joinedPointDialog->set_modal(true);
		}
		joinedPointDialog->show();
	} else {
		// draw a straight line from node1 to node2
		JoinAction *newAction = new JoinAction(node1, node2);
		setAction(newAction);
		if (node1->path == node2->path) {
			//g_printerr("GDDocument::onJoinObjects same path, just close it\n");
			// both are on the same path, so they must be beginning and end point
			if (node1->next) {
				BezierNode *tmp = node2;
				node2 = node1;
				node1 = tmp;
			}
			node1->next = node2;
			node1->nextHandle.x = node1->onCurve.x;
			node1->nextHandle.y = node1->onCurve.y;
			node2->prev = node1;
			node2->prevHandle.x = node2->onCurve.x;
			node2->prevHandle.y = node2->onCurve.y;
		} else {
			BezierPath *oldpath;
			if (node1->next == NULL) {
				//g_printerr("GDDocument::onJoinObjects node1 is endpoint\n");
				if (node2->prev) {
					//g_printerr("GDDocument::onJoinObjects node2 is endpoint->reverse\n");
					node2->path->reverseOrder();
				}
				// connect the end of node1 to the beginning of node2
				oldpath = node2->path;
				BezierNode *node = node2->path->nodes;
				while (node) {
					node->path = node1->path;
					node = node->next;
				}
				//g_printerr("GDDocument::onJoinObjects add node2 to path1 next:%lX, prev:%lX\n", node2->next, node2->prev);
				node1->path->addNode(node2);
			} else {
				//g_printerr("GDDocument::onJoinObjects node1 is startpoint\n");
				if (node2->next) {
					node2->path->reverseOrder();
				}
				oldpath = node1->path;
				BezierNode *node = node1;
				while (node) {
					node->path = node2->path;
					node = node->next;
				}
				node2->path->addNode(node1);
				// swap nodes
				node = node2;
				node2 = node1;
				node1 = node;
			}
			node1->nextHandle.x = node1->onCurve.x;
			node1->nextHandle.y = node1->onCurve.y;
			node2->prevHandle.x = node2->onCurve.x;
			node2->prevHandle.y = node2->onCurve.y;
			oldpath->nodes = NULL;
			removeObject(oldpath);
			delete oldpath;
		}
		redraw();
	}
}

void GDDocument::joinObjects() {
	G_List_Iterator<SelectionItem *> it = joinObjectSelection->begin();
	BezierNode *node1 = (BezierNode *) (*it++)->orig;
	BezierNode *node2 = (BezierNode *) (*it)->orig;
	joinedPointDialog->hide();
	gboolean isSmooth = joinedPointDialog->isSmooth();
	JoinAction *newAction = new JoinAction(node1, node2, isSmooth);
	setAction(newAction);
	node1->connectNodes(node2);
	if (isSmooth) {
		TransformationMatrix tm;
		int corr = 1;
		gdouble alpha = tm.getAngle(node1->nextHandle.x - node1->onCurve.x,
									node1->nextHandle.y - node1->onCurve.y) -
						tm.getAngle(node1->prevHandle.x - node1->onCurve.x,
									node1->prevHandle.y - node1->onCurve.y);
		if (alpha < 0) {
			alpha += 2 * M_PI;
		}
		if (alpha > M_PI) {
			alpha = 2 * M_PI - alpha;
			corr = -1;
		}
		//g_printerr("GDDocument::joinObjects angle between %f, corr factor: %d\n", alpha *180/M_PI, corr);
		alpha = (M_PI - alpha) / 2;
		//g_printerr("GDDocument::joinObjects correction angle %f\n", alpha *180/M_PI);
		tm.setRotation(alpha * corr);
		tm.setCenter(node1->onCurve.x, node1->onCurve.y);
		tm.transform(node1->nextHandle);
		tm.reset();
		tm.setRotation(-alpha * corr);
		tm.setCenter(node1->onCurve.x, node1->onCurve.y);
		tm.transform(node1->prevHandle);
		node1->straightHandles = true;
	}
	delete joinObjectSelection;
	joinObjectSelection = NULL;
	redraw();
}

void GDDocument::onMoveObjects() {
	if (! application->moveDialog) {
		application->moveDialog = new MoveDialog(application);
		application->moveDialog->getOKButton()->clicked.connect(SigC::slot(this, &GDDocument::moveObjects));
		application->moveDialog->getCopyButton()->clicked.connect(SigC::slot(this, &GDDocument::copyMovedObjects));
		application->moveDialog->getCancelButton()->clicked.connect(SigC::slot(this, &GDDocument::cancelMoveDialog));
		application->moveDialog->delete_event.connect(SigC::slot(this, &GDDocument::deleteMoveDialog));
		application->moveDialog->set_modal(true);
	}
	application->moveDialog->show();

}

gint GDDocument::deleteMoveDialog(GdkEventAny *ev) {
	cancelMoveDialog();
}

void GDDocument::cancelMoveDialog() {
	application->moveDialog->hide();
}

SelectionItemList *GDDocument::translateSelection() {
	SelectionItemList *selection = new SelectionItemList();
	getSelection(*selection);
	gdouble hm = application->moveDialog->getHorizontalTranslation();
	gdouble vm = application->moveDialog->getVerticalTranslation();
	TransformationMatrix tm;
	tm.setTranslation(hm, vm);
	selection->transformByMatrix(tm);
	application->moveDialog->hide();
	return selection;
}

void GDDocument::moveObjects() {
	SelectionItemList *selection = translateSelection();
	selection->applyToOriginal();
	redraw();
	delete selection;
}

void GDDocument::copyMovedObjects() {
	SelectionItemList *selection = translateSelection();
	deSelectAll();
	selection->insertMoving(this);
	redraw();
	delete selection;
}

GDDocument::~GDDocument() {
	if (fileselection) {
		delete fileselection;
		fileselection = NULL;
	}
	G_List_Iterator<DocWindow *> winit = docWindowList->begin();
	DocWindow *win;
	while ((win = *winit++)) {
		delete(win);
	}
	Layer *layer = layers;
	while (layer) {
		Layer *next = layer->next;
		delete layer;
		layer = next;
	}
	// free the resources
	for (GradientMap::iterator it = gradients.begin(); it != gradients.end(); it++) {
		delete (*it).second;
	}
	for (PatternMap::iterator it = patterns.begin(); it != patterns.end(); it++) {
		delete (*it).second;
	}
	for (CustomColorMap::iterator it = customColors.begin(); it != customColors.end(); it++) {
		delete (*it).second;
	}
	//g_printerr("GDDocument::~GDDocument %lX\n", this);
}
