/* $Id: Layer.cc,v 1.19 2005/05/07 17:39:33 pklaus Exp $
 *
 * Layer.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "Layer.hh"
#include "GObject.hh"
#include "SelectionItem.hh"
#include "GDDocument.hh"
#include <stdio.h>

guint32 selectionColorChoices[] = {
	0x0000FFFF,
	0x00FF00FF,
	0xFF0000FF,
	0x000000FF,
	0xFFFF00FF,
	0xFF00FFFF,
	0x00FFFFFF,
	0x0000BFFF,
	0x00BF00FF,
	0xBF0000FF, // 10
	0xFF00BFFF,
	0xFFBF00FF,
	0x00FFBFFF,
	0xBFFF00FF,
	0xBF00FFFF,
	0x00BFFFFF
	};

Layer::Layer(GDDocument *document) {
	doc = document;
	selectionColor = selectionColorChoices[doc->layerNumber & 0x0000000F];
	paths = NULL;
	doc->layerNumber++;
	name = "Layer ";
	if (doc->layerNumber > 999) {
		name += string(1, (char) ('0' + (doc->layerNumber / 1000)));
	}
	if (doc->layerNumber > 99) {
		name += string(1, (char) ('0' + ((doc->layerNumber % 1000) / 100)));
	}
	if (doc->layerNumber > 9) {
		name += string(1, (char) ('0' + ((doc->layerNumber % 100) / 10)));
	}
	name += string(1, (char) ('0' + (doc->layerNumber % 10)));
	viewable = true;
	editable = true;
}

Layer::Layer(Layer &fromLayer) {
	doc = fromLayer.doc;
	selectionColor = fromLayer.selectionColor;
	paths = NULL;
	name = fromLayer.name;
	viewable = fromLayer.viewable;
	editable = fromLayer.editable;
}
guint32 Layer::getSelectionColor() {
	return selectionColor;
}

void Layer::setSelectionColor(guint32 color) {
	selectionColor = color;
}

void Layer::setName(string name) {
	this->name = name;
}

string Layer::getName() {
	return name;
}

void Layer::addObject(GObject *newObject) {
	//g_printerr("Layer::addObject object=%lX\n", newObject);
	if (paths) {
		paths->insert(NULL, newObject);
	} else {
		paths = newObject;
	}
}

GObject *Layer::removeObject(GObject *object) {
	GObject *got;
	GObject *path = paths;
	if (object == paths) {
		if (paths->next) {
			paths = paths->next;
		} else {
			paths = NULL;
		}
		return object;
	} else {
		while (path) {
			if ((got = path->removeObject(object))) {
				return got;
			}
			path = path->next;
		}
	}
	return NULL;
}

gboolean Layer::findObjectLayer(GObject *object) {
	//g_printerr("Layer::findPathLayer\n");
	GObject *got;
	GObject *path = paths;
	while (path) {
		//g_printerr("Layer::findPathLayer check path: %lX, next: %lX\n", path, path->next);
		if ((got = path->findObject(object))) {
			//g_printerr("Layer::findPathLayer got it\n");
			return true;
		}
		path = path->next;
	}
	return false;
}

gboolean Layer::getObjectNumber(GObject *object, guint& number) {
	//g_printerr("Layer::findPathLayer\n");
	GObject *got;
	GObject *path = paths;
	while (path) {
		//g_printerr("Layer::findPathLayer check path: %lX, next: %lX\n", path, path->next);
		if ((got = path->findObject(object))) {
			//g_printerr("Layer::findPathLayer got it\n");
			return true;
		}
		path = path->next;
		number++;
	}
	return false;
}

void Layer::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("Layer::draw\n");
	if (! viewable) return;
	GObject *pathList = paths;
	while (pathList) {
		//g_printerr("Layer::draw GObject %lX (%s)\n", pathList, pathList->identify());
		pathList->draw(osBuffer, drawArea);
		pathList = pathList->next;
	}
	//g_printerr("Layer::draw %lX finished\n", this);
}

void Layer::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	if (! viewable) return;
	GObject *pathList = paths;
	while (pathList) {
		//g_printerr("Layer::drawWireFrame GObject %lX\n", pathList);
		pathList->drawWireFrame(docwin, context, drawArea);
		pathList = pathList->next;
	}
}

BoundingBox Layer::getBoundingBox() {
	BoundingBox bbox;
	if (! viewable) return bbox;
	GObject *pathList = paths;
	while (pathList) {
		BoundingBox bb = pathList->getBoundingBox();
		bbox.combine(bb);
		pathList = pathList->next;
	}
	return bbox;
}

gdouble Layer::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = G_MAXDOUBLE;
	gdouble temp_ax = ax, temp_ay = ay;
	GObject *path = paths;
	while (path) {
		gdouble res;
		if ((res = path->attractPoint(x, y, temp_ax, temp_ay, tolerance, attractType)) < nearest) {
			nearest = res;
			ax = temp_ax;
			ay = temp_ay;
		}
		path = path->next;
	}
	return nearest;
}


SelectableItem *Layer::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	if (! editable) return NULL;
	SelectableItem *found = NULL;
	GObject *path = paths;
	while (path) {
		SelectableItem *candidate;
		if ((candidate = path->checkPoints(x, y, tolerance, onlyCurvePoints)) != NULL) {
			found = candidate;
		}
		path = path->next;
	}
	return found;
}

SelectableItem *Layer::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	if (! editable) return NULL;
	SelectableItem *found = NULL;
	GObject *path = paths;
	while (path) {
		SelectableItem *candidate;
		if ((candidate = path->checkCurves(x, y, tolerance)) != NULL) {
			found = candidate;
		}
		path = path->next;
	}
	return found;
}

GObject *Layer::checkArea(gdouble x, gdouble y) {
	if (! editable) return NULL;
	GObject *found = NULL;
	GObject *path = paths;
	while (path) {
		GObject *candidate;
		if ((candidate = path->checkArea(x, y)) != NULL) {
			found = candidate;
		}
		path = path->next;
	}
	return found;
}


void Layer::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	if (! editable) return;
	GObject *path = paths;
	while (path) {
		path->selectRectangle(left, top, right, bottom, what);
		path = path->next;
	}
}

void Layer::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	if (! editable) return;
	GObject *path = paths;
	while (path) {
		path->deSelectRectangle(left, top, right, bottom, what);
		path = path->next;
	}
}

void Layer::selectAll() {
	//g_printerr("Layer::selectAll()\n");
	if (! editable) return;
	GObject *path = paths;
	while (path) {
		path->selectAll();
		path = path->next;
	}
}

void Layer::deSelectAll() {
	//if (! viewable) return;
	GObject *path = paths;
	while (path) {
		path->deSelectAll();
		path = path->next;
	}
}

gint Layer::getSelection(SelectionItemList& selectionList, gint mode) {
	//if (! viewable) return;
	GObject *path = paths;
	gint selectedParts = 0;
	while (path) {
		selectedParts += path->getSelection(selectionList, mode);
		path = path->next;
	}
	containsSelectedItems = (selectedParts > 0);
	return selectedParts;
}

void Layer::gatherResource(GDResourceList& collection, gint type) {
	if (! viewable) return;
	GObject *path = paths;
	while (path) {
		path->gatherResource(collection, type);
		path = path->next;
	}
}

Layer *Layer::clone(gboolean copy) {
	Layer *layer;
	if (copy) {
		layer = new Layer(doc);
	} else {
		layer = new Layer(*this);
	}
	GObject *path = paths;
	while (path) {
		layer->addObject((GObject *) path->clone());
		path = path->next;
	}
	return layer;
}

void Layer::setDocument(GDDocument *document) {
	GObject *path = paths;
	while (path) {
		path->setDocument(document);
		path = path->next;
	}
}

gint Layer::writeXML(ostream& xmlout) {
	GObject *path = paths;
	char colorval[10];
	xmlout << "<g gestalter:layer=\"";
	xmlout << name;
	xmlout << "\" gestalter:selcolor=\"";
	snprintf(colorval, 8, "#%06hhX", (selectionColor >> 8) & 0xFFFFFF);
	xmlout << colorval;
	xmlout << "\" gestalter:viewable=\"";
	xmlout << (viewable ? "yes" : "no");
	xmlout << "\" gestalter:editable=\"";
	xmlout << (editable ? "yes" : "no");
	xmlout << "\">\n";
	while (path) {
		path->writeXML(xmlout);
		path = path->next;
	}
	xmlout << "</g>\n";
	return 0;
}

gint Layer::writePS(ostream& psout) {
	if (! viewable) return 1;
	psout << "% layer " << name << "\n";
	psout << "gs\n"; // gsave
	GObject *path = paths;
	while (path) {
		path->writePS(psout);
		path = path->next;
	}
	psout << "gr\n"; // grestore
	return 0;
}

Layer::~Layer() {
	GObject *path = paths;
	while (path) {
		GObject *next = path->next;
		delete path;
		path = next;
	}
}
