/* $Id: MorphTool.cc,v 1.2 2005/05/08 08:38:17 pklaus Exp $
 *
 * MorphTool.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include "config.h"
#include <libgnome/libgnome.h> // to get gettext support
#include "../pixmaps/morph.xpm"
#include "../cursors/crosshair_source.xbm"
#include "../cursors/crosshair_source_mask.xbm"
#include "../cursors/crosshair_target.xbm"
#include "../cursors/crosshair_target_mask.xbm"
#include "../cursors/crosshair.xbm"
#include "../cursors/crosshair_mask.xbm"
#include "MorphTool.hh"
#include "GDDocument.hh"
#include "BezierPath.hh"
#include "BezierNode.hh"
#include "ObjectGroup.hh"
#include "AddAction.hh"
#include "Layer.hh"
#include "DocWindow.hh"
#include "CursorFactory.hh"
#include "GDApplication.hh"

MorphTool::MorphTool() {
	pixmapData = morph_xpm;
	toolName = "MorphTool";
	moving = false;
	mode = MORPH_TOOL_MODE_SOURCE;
	dialogDoc = NULL;
	valueDialog = NULL;
	CursorFactory::createCursorFromData(crosshairSourceCursor,
					crosshair_source_bits, crosshair_source_mask_bits,
					crosshair_source_width, crosshair_source_height,
					crosshair_source_x_hot, crosshair_source_y_hot);
	CursorFactory::createCursorFromData(crosshairTargetCursor,
					crosshair_target_bits, crosshair_target_mask_bits,
					crosshair_target_width, crosshair_target_height,
					crosshair_target_x_hot, crosshair_target_y_hot);
	CursorFactory::createCursorFromData(crosshairCursor,
					crosshair_bits, crosshair_mask_bits,
					crosshair_width, crosshair_height,
					crosshair_x_hot, crosshair_y_hot);
}

void MorphTool::activate() {
	Tool::activate();
	mode = MORPH_TOOL_MODE_SOURCE;
}

void MorphTool::deactivate() {
	Tool::deactivate();
	mode = MORPH_TOOL_MODE_SOURCE;
}

gint MorphTool::buttonDown(GdkEventButton *ev, DocWindow *docwin) {

	if (ev->button == 1) {
		dialogDoc = docwin->doc;
		gdouble x = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
		gdouble y = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
		gboolean isCurve = false;
		SelectableItem *node = dialogDoc->checkPoints(x, y,
				dialogDoc->application->preferences.catchingRange / docwin->osBuffer.scaleFactor, true);
		if (node && ! strcmp(node->identify(), "BezierNode")) {
			BezierNode *bnode = (BezierNode *) node;
			if (! (bnode->path && ! bnode->path->parent)) {
				dialogDoc->application->showError(_("Valid nodes for the morph operation must be part of ungrouped paths"));
				return -1;
			}
			if (mode == MORPH_TOOL_MODE_TARGET) {
				if (source && bnode->path == source->path) {
					dialogDoc->application->showError(_("Source and target node of morph operation must be on two different objects"));
					mode = MORPH_TOOL_MODE_SOURCE;
					return -1;
				}
				target = bnode;
				// show dialog
				mode = MORPH_TOOL_MODE_DIALOG;
				docwin->DocWinDrawingArea->get_window().set_cursor(crosshairCursor);
				if (! valueDialog) {
					valueDialog = new MorphDialog();
					valueDialog->getOKButton()->clicked.connect(SigC::slot(this, &MorphTool::doMorph));
					valueDialog->getCancelButton()->clicked.connect(SigC::slot(this, &MorphTool::cancelValueDialog));
					valueDialog->delete_event.connect(SigC::slot(this, &MorphTool::deleteValueDialog));
					valueDialog->set_modal(true);
				}
				valueDialog->show();

			} else {
				source = bnode;
				mode = MORPH_TOOL_MODE_TARGET;
			}
		}
	} else {
		Tool::buttonDown(ev, docwin);
	}
	return 0;
}

gint MorphTool::motion(GdkEventMotion *ev, DocWindow *docwin) {
	if (ev->state & GDK_BUTTON1_MASK) {
		//GDDocument *doc = docwin->doc;
	} else if (ev->state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) {
		Tool::motion(ev, docwin);
	} else {
		setCursor(ev->state, docwin);
	}
	
  return 0;
}

gint MorphTool::buttonUp(GdkEventButton *ev, DocWindow *docwin) {
	if (downButton == 1) {
		//GDDocument *doc = docwin->doc;
	} else {
		Tool::buttonUp(ev, docwin);
	}
	return 0;
}

gint MorphTool::deleteValueDialog(GdkEventAny *ev) {
	cancelValueDialog();
	return 1;
}

void MorphTool::cancelValueDialog() {
	valueDialog->hide();
	mode = MORPH_TOOL_MODE_SOURCE;
}

#define morphPoint(s,t,n)	s + (t - s) * n

void MorphTool::doMorph() {
	valueDialog->hide();
	dialogDoc->deSelectAll();
	int steps = valueDialog->getStepCount();
	steps = abs(steps);
	float firstStep = valueDialog->getFirstStep();
	guint sNum = dialogDoc->getObjectNumber(source->path);
	guint tNum = dialogDoc->getObjectNumber(target->path);
	if (sNum > tNum) {
		// target is lower than source, switch target and source
		BezierNode *tmp = target;
		target = source;
		source = tmp;
	}
	BezierPath *sourcePath = (BezierPath *) source->path->clone();
	BezierPath *targetPath = (BezierPath *) target->path->clone();
	guint sIndex = source->path->getNodeIndex(source);
	guint tIndex = target->path->getNodeIndex(target);
	int sCount = sourcePath->countNodes();
	int tCount = targetPath->countNodes();
	if (sourcePath->nodes->prev && targetPath->nodes->prev) {
		// both paths are closed
		if (sCount < tCount) {
			BezierNode *sNode = sourcePath->getIndexNode(sIndex);
			for (int i = 0; i < tCount - sCount; i++) {
				BezierNode *newNode = sNode->subDivide(0.5);
			}
		} else if (tCount < sCount) {
			BezierNode *tNode = targetPath->getIndexNode(tIndex);
			for (int i = 0; i < sCount - tCount; i++) {
				BezierNode *newNode = tNode->subDivide(0.5);
			}
		}
		sourcePath->nodes = sourcePath->getIndexNode(sIndex);
		targetPath->nodes = targetPath->getIndexNode(tIndex);
	} else {
		if (sourcePath->nodes->prev) {
			// target path is not closed
			BezierNode *sNode = sourcePath->getIndexNode(sIndex);
			int splitIndex = sIndex;
			splitIndex -= tIndex;
			while (splitIndex < 0) {
				splitIndex += sCount;
			}
			BezierNode *splitNode = sourcePath->getIndexNode(splitIndex);
			BezierNode *newNode = (BezierNode *) splitNode->clone();
			newNode->next = NULL;
			newNode->prev = NULL;
			splitNode->path->nodes->insert(splitNode, newNode);
			splitNode->path->deleteCurve(splitNode);
			sCount++; // we just added a new node
			sIndex = sourcePath->getNodeIndex(sNode);
		} else if (targetPath->nodes->prev) {
			// source path is not closed
			BezierNode *tNode = targetPath->getIndexNode(tIndex);
			int splitIndex = tIndex;
			splitIndex -= sIndex;
			while (splitIndex < 0) {
				splitIndex += tCount;
			}
			BezierNode *splitNode = targetPath->getIndexNode(splitIndex);
			BezierNode *newNode = (BezierNode *) splitNode->clone();
			newNode->next = NULL;
			newNode->prev = NULL;
			splitNode->path->nodes->insert(splitNode, newNode);
			splitNode->path->deleteCurve(splitNode);
			tCount++; // we just added a new node
			tIndex = targetPath->getNodeIndex(tNode);
		}
		// now both paths are not closed
		if (sIndex > tIndex) {
			BezierNode *tNode = targetPath->getIndexNode(tIndex);
			if (tNode->prev) {
				for (guint i = 0; i < sIndex - tIndex; i++) {
					tNode->prev->subDivide(0.5);
				}
			} else {
				for (guint i = 0; i < sIndex - tIndex; i++) {
					BezierNode *newNode = (BezierNode *) tNode->clone();
					newNode->prevHandle = newNode->onCurve;
					newNode->nextHandle = newNode->onCurve;
					newNode->prev = NULL;
					newNode->next = tNode->path->nodes;
					tNode->path->nodes->prev = newNode;
					tNode->path->nodes = newNode;
				}
			}
		} else if (tIndex > sIndex) {
			BezierNode *sNode = sourcePath->getIndexNode(sIndex);
			if (sNode->prev) {
				for (guint i = 0; i < tIndex - sIndex; i++) {
					sNode->prev->subDivide(0.5);
				}
			} else {
				for (guint i = 0; i < tIndex - sIndex; i++) {
					BezierNode *newNode = (BezierNode *) sNode->clone();
					newNode->prevHandle = newNode->onCurve;
					newNode->nextHandle = newNode->onCurve;
					newNode->prev = NULL;
					newNode->next = sNode->path->nodes;
					sNode->path->nodes->prev = newNode;
					sNode->path->nodes = newNode;
				}
			}
		}
		guint sUpCount = sCount - sIndex;
		guint tUpCount = tCount - tIndex;
		if (sUpCount > tUpCount) {
			BezierNode *tNode = targetPath->getIndexNode(tIndex);
			for (guint i = 0; i < sUpCount - tUpCount; i++) {
				tNode->subDivide(0.5);
			}
		} else if (tUpCount > sUpCount) {
			BezierNode *sNode = sourcePath->getIndexNode(sIndex);
			for (guint i = 0; i < tUpCount - sUpCount; i++) {
				sNode->subDivide(0.5);
			}
		}
	}
	int nodeCount = sourcePath->countNodes();
	ObjectGroup *resultGroup = new ObjectGroup(dialogDoc);
	double stepval = 1.0 / (steps + 1);
	for (double t = stepval; t < 1.0; t += stepval) {
		BezierPath *calcPath = (BezierPath *) sourcePath->clone();
		BezierNode *calcNode = calcPath->nodes;
		BezierNode *sNode = sourcePath->nodes;
		BezierNode *tNode = targetPath->nodes;
		for (int n = 0; n < nodeCount; n++) {
			if (sNode == NULL || tNode == NULL || calcNode == NULL) {
				break;
			}
			calcNode->onCurve.x = morphPoint(sNode->onCurve.x, tNode->onCurve.x, t);
			calcNode->onCurve.y = morphPoint(sNode->onCurve.y, tNode->onCurve.y, t);
			calcNode->prevHandle.x = morphPoint(sNode->prevHandle.x, tNode->prevHandle.x, t);
			calcNode->prevHandle.y = morphPoint(sNode->prevHandle.y, tNode->prevHandle.y, t);
			calcNode->nextHandle.x = morphPoint(sNode->nextHandle.x, tNode->nextHandle.x, t);
			calcNode->nextHandle.y = morphPoint(sNode->nextHandle.y, tNode->nextHandle.y, t);
			//g_printerr("Morph Node: %f/%f   Source: %f/%f    Target %f/%f     n:%f\n", calcNode->onCurve.x, calcNode->onCurve.y, sNode->onCurve.x, sNode->onCurve.y, tNode->onCurve.x, tNode->onCurve.y, t);
			sNode = sNode->next;
			tNode = tNode->next;
			calcNode = calcNode->next;
		}
		calcPath->calcMorphValues(sourcePath, targetPath, t);
		resultGroup->addPart(calcPath);
		//g_printerr("Adding new path %lX to group %lX (val: %f, nCount: %d)\n", (long)calcPath, (long)resultGroup, t, nodeCount);
	}
	source->path->insert(source->path, resultGroup);
	Layer *sourceLayer = dialogDoc->findObjectLayer(source->path);
	AddAction *newAction = new AddAction(resultGroup, sourceLayer);
	dialogDoc->setAction(newAction);
	resultGroup->selectAll();
	sourcePath->deSelectAll();
	targetPath->deSelectAll();
	delete sourcePath;
	delete targetPath;
	//g_printerr("Added group %lX to layer %lX\n", (long)resultGroup, (long)dialogDoc->activeLayer);
	mode = MORPH_TOOL_MODE_SOURCE;
	source = NULL;
	target = NULL;
	dialogDoc->redraw();
}

void MorphTool::setCursor(guint state, DocWindow *docwin) {
	if (mode == MORPH_TOOL_MODE_TARGET) {
		docwin->DocWinDrawingArea->get_window().set_cursor(crosshairTargetCursor);
	} else if (mode == MORPH_TOOL_MODE_SOURCE) {
		docwin->DocWinDrawingArea->get_window().set_cursor(crosshairSourceCursor);
	} else {
		docwin->DocWinDrawingArea->get_window().set_cursor(crosshairCursor);
	}
}
