/* $Id: PathMingler.cc,v 1.5 2004/06/13 16:20:12 pklaus Exp $
 *
 * PathMingler.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "PathMingler.hh"
#include "GDDocument.hh"
#include "BezierPath.hh"
#include "BezierNode.hh"
#include "Layer.hh"
#include <algorithm>
#include <map>

#define DBG(exp) 

typedef map<BezierPath *, gint, ltpath> PathObjectDepthMap;

typedef set<BezierPath *, ltpath> BezierPathSet;

ISObjectVector PathMingler::getIntersectionObjects(GDDocument *doc, GObjectDepthMap& gobjects) {
	ISObjectVector objectPaths;
	SelectionItemList *selection = new SelectionItemList();
	doc->getSelection(*selection);
	G_List_Iterator<SelectionItem *> it = selection->begin();
	SelectionItem *item;
	int objectCount = 1;
	while ((item = *it++)) {
		SelectableItem *sitem = item->orig;
		GObject *object = sitem->getGObject();
		if (gobjects.find(object) == gobjects.end()) {
			gobjects[object] = objectCount;
			BezierPathVector *paths = object->getBezierPaths();
			if (paths) {
				ISObject *isO = new ISObject();
				isO->paths = paths;
				isO->object = object;
				objectPaths.push_back(isO);
			}
		}
		objectCount++;
	}
	return objectPaths;
}

ObjectIntersectionVector PathMingler::calcIntersections(ISObjectVector& objectPaths) {
	ObjectIntersectionVector result;
	if (objectPaths.size() > 1) { // path intersection checks make only sense if more than one object is to be tested
		for (ISObjectVector::iterator it1 = objectPaths.begin(); it1 != objectPaths.end(); it1++) {
			ISObjectVector::iterator it2 = it1;
			for (it2++; it2 != objectPaths.end(); it2++) {
				// check every selected object with every other selected object but itself
				for (BezierPathVector::iterator sitA = (*it1)->paths->begin(); sitA != (*it1)->paths->end(); sitA++) {
					for (BezierPathVector::iterator sitB = (*it2)->paths->begin(); sitB != (*it2)->paths->end(); sitB++) {
						// check for intersections every path of object A with every path of object B
						IntersectionVector *ints = (*sitA)->getIntersections(*sitB);
						//g_printerr("DocumentMenu::onPathMingleDivide check IS %lX and %lX = %lX\n", *sitA, *sitB, ints);
						if (ints) {
							for (IntersectionVector::iterator it = ints->begin(); it != ints->end(); it++) {
								ObjectIntersection *objInt = new ObjectIntersection();
								objInt->objectA = (*it1)->object;
								objInt->objectB = (*it2)->object;
								objInt->fStyleA = *((*it1)->object->getFillStyle());
								objInt->sStyleA = *((*it1)->object->getStrokeStyle());
								objInt->fStyleB = *((*it2)->object->getFillStyle());
								objInt->sStyleB = *((*it2)->object->getStrokeStyle());
								objInt->intersection = *it;
								result.push_back(objInt);
							}
							delete ints;
						}
					}
				}
			}
		}
	}
	return result;
}

void PathMingler::createIntersectionNodes(ObjectIntersectionVector& result) {
	if (result.size()) {
		// create the intersection nodes
		//g_printerr("DocumentMenu::onPathMingleDivide %d Intersections found\n", result.size());
		vector<SplitNode> sortVector;
		for (ObjectIntersectionVector::iterator it = result.begin(); it != result.end(); it++) {
			//g_printerr("x=%f y=%f srA=%f srB=%f nA=%lX nB=%lX\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->intersection.splitRatioA, (*it)->intersection.splitRatioB, (long) (*it)->intersection.nodeA, (long) (*it)->intersection.nodeB);
			SplitNode sn;
			sn.baseNode = (*it)->intersection.nodeA;
			sn.splitRatio = (*it)->intersection.splitRatioA;
			sn.newNode = &((*it)->newNodeA);
			sortVector.push_back(sn);
			//sn = new SplitNode();
			sn.baseNode = (*it)->intersection.nodeB;
			sn.splitRatio = (*it)->intersection.splitRatioB;
			sn.newNode = &((*it)->newNodeB);
			sortVector.push_back(sn);
		}
		sort(sortVector.begin(), sortVector.end());
		//g_printerr("Sorted splits are:\n");
		BezierNode *base = NULL;
		BezierNode *curve = NULL;
		gdouble lastSplit = 0;
		for (vector<SplitNode>::iterator it = sortVector.begin(); it != sortVector.end(); it++) {
			//g_printerr("n=%lX sr=%f\n", (long) (*it).baseNode, (*it).splitRatio);
			
			if (base != (*it).baseNode) {
				base = (*it).baseNode;
				curve = base;
				lastSplit = 0;
				//g_printerr("New base\n");
			}
			gdouble ratio = ((*it).splitRatio - lastSplit) / (1.0 - lastSplit);
			BezierNode *bnode;
			if (ratio > 0) {
				bnode = curve->subDivide(ratio);
			} else {
				bnode = curve;
			}
			*((*it).newNode) = bnode;
			curve = bnode;
			lastSplit = (*it).splitRatio;
			//g_printerr("Split curve: split node=%lX ratio=%f\n", (long) bnode, ratio);
			
		}
	}
}

ObjectIntersectionVector PathMingler::getISVector(GDDocument *doc, GObjectDepthMap& gobjects, GObject **topObj, GObject **bottomObj) {
	ISObjectVector objectPaths = getIntersectionObjects(doc, gobjects);
	if (topObj) {
		*topObj = objectPaths.back()->object;
	}
	if (bottomObj) {
		*bottomObj = objectPaths.front()->object;
	}
	// get all intersections
	ObjectIntersectionVector result = calcIntersections(objectPaths);
	// clean up
	for (ISObjectVector::iterator it = objectPaths.begin(); it != objectPaths.end(); it++) {
		delete (*it)->paths;
		delete *it;
	}
	createIntersectionNodes(result);
	return result;
}

/*ObjectIntersectionVector PathMingler::getTopObjectISVector(GDDocument *doc, GObjectDepthMap& gobjects, GObject **topObj) {
	ISObjectVector objectPaths = getIntersectionObjects(doc, gobjects);
	// get all intersections
	ObjectIntersectionVector result;
	if (objectPaths.size() > 1) { // path intersection checks make only sense if more than one object is to be tested
		for (ISObjectVector::iterator it1 = objectPaths.begin(); it1 != objectPaths.end(); it1++) {
			ISObject *obj2 = objectPaths.back();
			*topObj = obj2->object;
			if (*it1 == obj2) continue;
			// check every selected object with every other selected object but itself
			for (BezierPathVector::iterator sitA = (*it1)->paths->begin(); sitA != (*it1)->paths->end(); sitA++) {
				for (BezierPathVector::iterator sitB = obj2->paths->begin(); sitB != obj2->paths->end(); sitB++) {
					// check for intersections every path of object A with every path of object B
					IntersectionVector *ints = (*sitA)->getIntersections(*sitB);
					//g_printerr("DocumentMenu::onPathMingleDivide check IS %lX and %lX = %lX\n", *sitA, *sitB, ints);
					if (ints) {
						for (IntersectionVector::iterator it = ints->begin(); it != ints->end(); it++) {
							ObjectIntersection *objInt = new ObjectIntersection();
							objInt->objectA = (*it1)->object;
							objInt->objectB = obj2->object;
							objInt->fStyleA = *((*it1)->object->getFillStyle());
							objInt->sStyleA = *((*it1)->object->getStrokeStyle());
							objInt->fStyleB = *(obj2->object->getFillStyle());
							objInt->sStyleB = *(obj2->object->getStrokeStyle());
							objInt->intersection = *it;
							result.push_back(objInt);
						}
						delete ints;
					}
				}
			}
		}
	}
	// clean up
	for (ISObjectVector::iterator it = objectPaths.begin(); it != objectPaths.end(); it++) {
		delete (*it)->paths;
		delete *it;
	}
	createIntersectionNodes(result);
	return result;
}*/

int PathMingler::getCoverage(ObjectIntersection *objIntersection, GObjectDepthMap& gobjects, GObject **coverObject) {
	gint coverage = 0;
	*coverObject = NULL;
	gint maxDepth = -1;
	/*gint maxDepth = gobjects[objIntersection->objectA];
	if (maxDepth < gobjects[objIntersection->objectB]) {
		maxDepth = gobjects[objIntersection->objectB];
	}*/
	for (GObjectDepthMap::iterator it = gobjects.begin(); it != gobjects.end(); it++) {
		if ((*it).first == objIntersection->objectA || (*it).first == objIntersection->objectB) {
			continue;
		}
		if ((*it).first->checkArea(objIntersection->intersection.intersectPos.x, objIntersection->intersection.intersectPos.y)) {
			coverage++;
			if ((*it).second > maxDepth) {
				*coverObject = (*it).first;
				maxDepth = (*it).second;
			}
		}
	}
	return coverage;
}

BezierNode *PathMingler::getDupNode(BezierNode *node, BezierPathMap& duplicates) {
	BezierPath *dPath = duplicates[node->path];
	DBG(g_printerr("PathMingler::pathMingleDivide path=%lX dup=%lX\n", (long) node->path, (long) dPath);)
	DBG(g_printerr("p.S=%lX p.E=%lX node=%lX\n", (long) node->path->getStartNode(), (long) node->path->getEndNode(), (long) node);)
	if (node->path->getStartNode() == node) { // newNode is start of Path
		return dPath->getStartNode();
	} else {
		return dPath->getEndNode();
	}
}

void PathMingler::getMiddlePoint(BezierNode *startNode, position& middle) {
	middle.x = startNode->onCurve.x * 0.125 + startNode->nextHandle.x * 0.375 + startNode->next->prevHandle.x * 0.375 + startNode->next->onCurve.x * 0.125;
	middle.y = startNode->onCurve.y * 0.125 + startNode->nextHandle.y * 0.375 + startNode->next->prevHandle.y * 0.375 + startNode->next->onCurve.y * 0.125;;
	
}

int PathMingler::getStatus(BezierNode *newNode, GObject *oppObject) {
	position middle;
	getMiddlePoint(newNode, middle);
	if (oppObject->checkArea(middle.x, middle.y)) {
		return PATHMINGLER_FOREIGN_COVER_NEXT;
	} else {
		return PATHMINGLER_FOREIGN_COVER_PREVIOUS;
	}
}

void PathMingler::calcCoverage(ObjectIntersectionVector& oisVector, GObjectDepthMap& godMap, gboolean doCoverObject) {
	for (ObjectIntersectionVector::iterator it = oisVector.begin(); it != oisVector.end(); it++) {
		(*it)->statusA = getStatus((*it)->newNodeA, (*it)->objectB);
		(*it)->statusB = getStatus((*it)->newNodeB, (*it)->objectA);
		if (doCoverObject) {
			(*it)->coverage = getCoverage(*it, godMap, &((*it)->coverObject));
		}
	}
}

BezierNode *PathMingler::splitNode(BezierNode *node) {
	BezierNode *newNode  = (BezierNode *) node->clone();
	newNode->next = NULL;
	newNode->prev = NULL;
	node->path->nodes->insert(node, newNode);
	node->path->deleteCurve(node);
	return newNode;
}

void PathMingler::splitPaths(ObjectIntersectionVector& oisVector) {
	for (ObjectIntersectionVector::iterator it = oisVector.begin(); it != oisVector.end(); it++) {
		// split up A part
		(*it)->addNodeA = splitNode((*it)->newNodeA);
		// split up B part
		(*it)->addNodeB = splitNode((*it)->newNodeB);
	}
}

void PathMingler::connectNodePairs(NodePairVector& nodePairs) {
	for (NodePairVector::iterator it = nodePairs.begin(); it != nodePairs.end(); it++) {
		DBG(g_printerr("PathMingler::connectNodePairs nA=%lX pNA=%lX nB=%lX pNB=%lX\n", (long) (*it).nodeA, (long) (*it).nodeA->path, (long) (*it).nodeB, (long) (*it).nodeB->path);)
		(*it).nodeA->connectNodes((*it).nodeB);
		(*it).fStyle.valueMask = 0xFFFFFFFF;
		(*it).nodeA->path->setFillStyle((*it).fStyle);
		(*it).sStyle.valueMask = 0xFFFFFFFF;
		(*it).nodeA->path->setStrokeStyle((*it).sStyle);
	}
}

void PathMingler::orderPaths(ObjectIntersection *ois, BezierNode **coverPathA, BezierNode **openPathA, BezierNode **coverPathB, BezierNode **openPathB) {
	if (ois->statusA == ois->statusB) {
		if (ois->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
			*coverPathA = ois->newNodeA;
			*openPathA = ois->addNodeA;
			*coverPathB = ois->newNodeB;
			*openPathB = ois->addNodeB;
		} else {
			*coverPathA = ois->addNodeA;
			*openPathA = ois->newNodeA;
			*coverPathB = ois->addNodeB;
			*openPathB = ois->newNodeB;
		}
	} else {
		if (ois->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
			*coverPathA = ois->newNodeA;
			*openPathA = ois->addNodeA;
			*coverPathB = ois->addNodeB;
			*openPathB = ois->newNodeB;
		} else {
			*coverPathA = ois->addNodeA;
			*openPathA = ois->newNodeA;
			*coverPathB = ois->newNodeB;
			*openPathB = ois->addNodeB;
		}
	}
}

void PathMingler::duplicateInnerPaths(ObjectIntersectionVector& oisVector, BezierPathMap& dupPathMap, GDDocument *doc) {
	for (ObjectIntersectionVector::iterator it = oisVector.begin(); it != oisVector.end(); it++) {
		DBG(g_printerr("PathMingler::duplicateInnerPaths objA=%lX nNA=%lX aNA=%lX\n", (long) (*it)->objectA, (long) (*it)->newNodeA, (long) (*it)->addNodeA);
		DBG(g_printerr("PathMingler::duplicateInnerPaths objA=%lX nNA.path=%lX aNA.path=%lX\n", (long) (*it)->objectA, (long) (*it)->newNodeA->path, (long) (*it)->addNodeA->path);)
		DBG(g_printerr("PathMingler::duplicateInnerPaths objB=%lX nNB=%lX aNB=%lX\n", (long) (*it)->objectB, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		DBG(g_printerr("PathMingler::duplicateInnerPaths objB=%lX nNB.path=%lX aNB.path=%lX\n", (long) (*it)->objectB, (long) (*it)->newNodeB->path, (long) (*it)->addNodeB->path);))
		// duplicate inner path of object A
		BezierPath *innerPathA;
		if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
			innerPathA = (*it)->newNodeA->path;
		} else {
			innerPathA = (*it)->addNodeA->path;
		}
		BezierPathMap::iterator bpmitA;
		if ((bpmitA = dupPathMap.find(innerPathA)) == dupPathMap.end()) {
			BezierPath *duplicatePathA = (BezierPath *)innerPathA->clone();
			doc->findObjectLayer(innerPathA)->addObject(duplicatePathA);
			dupPathMap[innerPathA] = duplicatePathA;
			DBG(g_printerr("PathMingler::duplicateInnerPaths duplicated PathA %lX to %lX\n", (long) innerPathA, (long) duplicatePathA);)
		} else {
			DBG(g_printerr("PathMingler::duplicateInnerPaths PathA %lX already dup: %lX -> %lX\n", (long) innerPathA, (long) (*bpmitA).first, (long) (*bpmitA).second);)
		}
		// duplicate inner path of object B
		BezierPath *innerPathB;
		if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
			innerPathB = (*it)->newNodeB->path;
		} else {
			innerPathB = (*it)->addNodeB->path;
		}
		BezierPathMap::iterator bpmitB;
		if ((bpmitB = dupPathMap.find(innerPathB)) == dupPathMap.end()) {
			BezierPath *duplicatePathB = (BezierPath *)innerPathB->clone();
			doc->findObjectLayer(innerPathB)->addObject(duplicatePathB);
			dupPathMap[innerPathB] = duplicatePathB;
			DBG(g_printerr("PathMingler::duplicateInnerPaths duplicated PathB %lX to %lX\n", (long) innerPathB, (long) duplicatePathB);)
		} else {
			DBG(g_printerr("PathMingler::duplicateInnerPaths PathB %lX already dup: %lX -> %lX\n", (long) innerPathB, (long) (*bpmitB).first, (long) (*bpmitB).second);)
		}
	}
}

void PathMingler::pathMingleUnite(GDDocument *doc) {
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleUnite CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleUnite SPLIT PATHS\n");)
	splitPaths(intersections);
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleUnite CONNECT PATHS\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleUnite: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		if ((*it)->coverage) {
			// cross point is covered, delete all paths
			removePaths.insert((*it)->newNodeA->path);
			removePaths.insert((*it)->newNodeB->path);
			removePaths.insert((*it)->addNodeA->path);
			removePaths.insert((*it)->addNodeB->path);
		} else {
			BezierNode *cPA, *oPA, *cPB, *oPB;
			orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
			NodePair np(oPA, oPB, (*it)->objectA, godMap[(*it)->objectA] );
			nodePairs.push_back(np);
			removePaths.insert(cPA->path);
			removePaths.insert(cPB->path);
		}
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleIntersect(GDDocument *doc) {
	// get intersections
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleIntersect CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleIntersect SPLIT PATHS\n");)
	splitPaths(intersections);
	// duplicate inner paths
	DBG(g_printerr("PathMingler::pathMingleIntersect DUPLICATE INNER PATHS\n");)
	BezierPathMap dupPathMap;
	duplicateInnerPaths(intersections, dupPathMap, doc);
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleExclude CONNECT PATHS\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleIntersect: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		BezierNode *iNodeA;
		BezierNode *iNodeB;
		BezierNode *cPA, *oPA, *cPB, *oPB;
		orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
		if ((*it)->coverage) {
			GObject *styleObject = (*it)->objectA;
			gint depth = 0;
			if ((*it)->coverObject && godMap[(*it)->coverObject] > godMap[(*it)->objectA]) {
				styleObject = (*it)->coverObject;
				depth = godMap[(*it)->coverObject];
			}
			if ((*it)->coverage % 2) {
				NodePair np1( getDupNode(oPA, dupPathMap), cPB, styleObject, depth );
				nodePairs.push_back(np1);
				NodePair np2( cPA, getDupNode(oPB, dupPathMap), styleObject, depth );
				nodePairs.push_back(np2);
				removePaths.insert(oPA->path);
				removePaths.insert(oPB->path);
				removePaths.insert(dupPathMap[cPA->path]);
				removePaths.insert(dupPathMap[cPB->path]);
			} else {
				NodePair np1( oPA, oPB, styleObject, depth );
				nodePairs.push_back(np1);
				styleObject = (godMap[(*it)->objectA] > godMap[(*it)->objectB]) ? (*it)->objectA : (*it)->objectB;
				NodePair np2( getDupNode(cPA, dupPathMap), getDupNode(cPB, dupPathMap), styleObject, godMap[styleObject] );
				nodePairs.push_back(np2);
				removePaths.insert(dupPathMap[oPA->path]);
				removePaths.insert(dupPathMap[oPB->path]);
				removePaths.insert(cPA->path);
				removePaths.insert(cPB->path);
			}
		} else {
			GObject *styleObject = (godMap[(*it)->objectA] > godMap[(*it)->objectB]) ? (*it)->objectA : (*it)->objectB;
			NodePair np(getDupNode(cPA, dupPathMap), getDupNode(cPB, dupPathMap), styleObject, godMap[styleObject] );
			nodePairs.push_back(np);
			removePaths.insert(cPA->path);
			removePaths.insert(cPB->path);
			removePaths.insert(oPA->path);
			removePaths.insert(oPB->path);
		}
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleExclude(GDDocument *doc) {
	// get intersections
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleExclude CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleExclude SPLIT PATHS\n");)
	splitPaths(intersections);
	// duplicate inner paths
	DBG(g_printerr("PathMingler::pathMingleExclude DUPLICATE INNER PATHS\n");)
	BezierPathMap dupPathMap;
	duplicateInnerPaths(intersections, dupPathMap, doc);
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleExclude CONNECT PATHS\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleExclude: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		BezierNode *iNodeA;
		BezierNode *iNodeB;
		BezierNode *cPA, *oPA, *cPB, *oPB;
		orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
		if ((*it)->coverage) {
			GObject *styleObject = (*it)->objectA;
			gint depth = 0;
			if ((*it)->coverObject && godMap[(*it)->coverObject] > godMap[(*it)->objectA]) {
				styleObject = (*it)->coverObject;
				depth = godMap[(*it)->coverObject];
			}
			if ((*it)->coverage % 2) {
				NodePair np1( oPA, oPB, styleObject, depth );
				nodePairs.push_back(np1);
				styleObject = (godMap[(*it)->objectA] > godMap[(*it)->objectB]) ? (*it)->objectA : (*it)->objectB;
				NodePair np2( getDupNode(cPA, dupPathMap), getDupNode(cPB, dupPathMap), styleObject, godMap[styleObject] );
				nodePairs.push_back(np2);
				removePaths.insert(dupPathMap[oPA->path]);
				removePaths.insert(dupPathMap[oPB->path]);
				removePaths.insert(cPA->path);
				removePaths.insert(cPB->path);
			} else {
				NodePair np1( getDupNode(oPA, dupPathMap), cPB, styleObject, depth );
				nodePairs.push_back(np1);
				NodePair np2( cPA, getDupNode(oPB, dupPathMap), styleObject, depth );
				nodePairs.push_back(np2);
				removePaths.insert(oPA->path);
				removePaths.insert(oPB->path);
				removePaths.insert(dupPathMap[cPA->path]);
				removePaths.insert(dupPathMap[cPB->path]);
			}
		} else {
			NodePair np1(oPA, cPB, (*it)->objectA, godMap[(*it)->objectA] );
			nodePairs.push_back(np1);
			NodePair np2( oPB, cPA, (*it)->objectB, godMap[(*it)->objectB] );
			nodePairs.push_back(np2);
			removePaths.insert(dupPathMap[cPA->path]);
			removePaths.insert(dupPathMap[cPB->path]);
		}
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleMinusFront(GDDocument *doc) {
	// get intersections
	GObjectDepthMap godMap;
	GObject *bottomObject = NULL;
	ObjectIntersectionVector intersections = getISVector(doc, godMap, NULL, &bottomObject);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleMinusFront CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleMinusFront SPLIT PATHS\n");)
	splitPaths(intersections);
	// duplicate inner paths
	/*g_printerr("PathMingler::pathMingleMinusFront DUPLICATE INNER PATHS\n");
	BezierPathMap dupPathMap;
	duplicateInnerPaths(intersections, dupPathMap, doc);*/
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleMinusFront CONNECT PATHS\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleMinusFront: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		BezierNode *iNodeA;
		BezierNode *iNodeB;
		BezierNode *cPA, *oPA, *cPB, *oPB;
		orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
		if ((*it)->coverage == 0 && bottomObject == (*it)->objectA) {
			NodePair np(oPA, cPB, (*it)->objectA, godMap[(*it)->objectA] );
			nodePairs.push_back(np);			
			removePaths.insert(cPA->path);
			removePaths.insert(oPB->path);
		} else if ((*it)->coverage == 0 && bottomObject == (*it)->objectB) {
			NodePair np(oPB, cPA, (*it)->objectB, godMap[(*it)->objectB] );
			nodePairs.push_back(np);			
			removePaths.insert(cPB->path);
			removePaths.insert(oPA->path);
		} else if ((*it)->coverage == 1 && bottomObject == (*it)->coverObject) {
			NodePair np(oPA, oPB, (*it)->coverObject, godMap[(*it)->coverObject] );
			nodePairs.push_back(np);			
			removePaths.insert(cPA->path);
			removePaths.insert(cPB->path);
		} else {
			removePaths.insert(cPA->path);
			removePaths.insert(oPA->path);
			removePaths.insert(cPB->path);
			removePaths.insert(oPB->path);
		}
	}
	//sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleMinusBack(GDDocument *doc) {
	// get intersections
	GObjectDepthMap godMap;
	GObject *topObject = NULL;
	ObjectIntersectionVector intersections = getISVector(doc, godMap, &topObject);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleMinusBack CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleMinusBack SPLIT PATHS\n");)
	splitPaths(intersections);
	// duplicate inner paths
	/*g_printerr("PathMingler::pathMingleMinusBack DUPLICATE INNER PATHS\n");
	BezierPathMap dupPathMap;
	duplicateInnerPaths(intersections, dupPathMap, doc);*/
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleMinusBack CONNECT PATHS\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleMinusBack: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		BezierNode *iNodeA;
		BezierNode *iNodeB;
		BezierNode *cPA, *oPA, *cPB, *oPB;
		orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
		if ((*it)->coverage == 0 && topObject == (*it)->objectA) {
			NodePair np(oPA, cPB, (*it)->objectA, godMap[(*it)->objectA] );
			nodePairs.push_back(np);			
			removePaths.insert(cPA->path);
			removePaths.insert(oPB->path);
		} else if ((*it)->coverage == 0 && topObject == (*it)->objectB) {
			NodePair np(oPB, cPA, (*it)->objectB, godMap[(*it)->objectB] );
			nodePairs.push_back(np);			
			removePaths.insert(cPB->path);
			removePaths.insert(oPA->path);
		} else if ((*it)->coverage == 1 && topObject == (*it)->coverObject) {
			NodePair np(oPA, oPB, (*it)->coverObject, godMap[(*it)->coverObject] );
			nodePairs.push_back(np);			
			removePaths.insert(cPA->path);
			removePaths.insert(cPB->path);
		} else {
			removePaths.insert(cPA->path);
			removePaths.insert(oPA->path);
			removePaths.insert(cPB->path);
			removePaths.insert(oPB->path);
		}
	}
	//sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleTrim(GDDocument *doc) {
	GObjectDepthMap godMap;
	ISObjectVector objectPaths = getIntersectionObjects(doc, godMap);
	// get all intersections
	ObjectIntersectionVector intersections = calcIntersections(objectPaths);
	// clean up
	for (ISObjectVector::iterator it = objectPaths.begin(); it != objectPaths.end(); it++) {
		delete (*it)->paths;
		delete *it;
	}
	// duplicate all paths
	BezierNodeMap dupNodeMap;
	BezierPathMap dupPathMap;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		(*it)->coverage = getCoverage(*it, godMap, &((*it)->coverObject));
		if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
			BezierPath *pathA = (*it)->intersection.nodeA->path;
			BezierPathMap::iterator bpmitA;
			if ((bpmitA = dupPathMap.find(pathA)) == dupPathMap.end()) {
				BezierPath *duplicatePathA = (BezierPath *)pathA->clone();
				doc->findObjectLayer(pathA)->addObject(duplicatePathA);
				dupPathMap[pathA] = duplicatePathA;
				BezierNode *oN = pathA->nodes;
				BezierNode *dN = duplicatePathA->nodes;
				while (oN) {
					dupNodeMap[oN] = dN;
					oN = oN->next;
					dN = dN->next;
					if (oN == pathA->nodes) {
						oN = NULL;
					}
				}
			}
		} else {
			BezierPath *pathB = (*it)->intersection.nodeB->path;
			BezierPathMap::iterator bpmitB;
			if ((bpmitB = dupPathMap.find(pathB)) == dupPathMap.end()) {
				BezierPath *duplicatePathB = (BezierPath *)pathB->clone();
				doc->findObjectLayer(pathB)->addObject(duplicatePathB);
				dupPathMap[pathB] = duplicatePathB;
				BezierNode *oN = pathB->nodes;
				BezierNode *dN = duplicatePathB->nodes;
				while (oN) {
					dupNodeMap[oN] = dN;
					oN = oN->next;
					dN = dN->next;
					if (oN == pathB->nodes) {
						oN = NULL;
					}
				}
			}
		}
	}
	// calculate all necessary new split nodes
	vector<SplitNode> sortVector;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				SplitNode sn1;
				sn1.baseNode = (*it)->intersection.nodeB;
				sn1.splitRatio = (*it)->intersection.splitRatioB;
				sn1.newNode = &((*it)->newNodeB);
				sortVector.push_back(sn1);
				SplitNode sn2;
				sn2.baseNode = dupNodeMap[(*it)->intersection.nodeA];
				sn2.splitRatio = (*it)->intersection.splitRatioA;
				sn2.newNode = &((*it)->dupNewNodeA);
				sortVector.push_back(sn2);
			} else {
				SplitNode sn1;
				sn1.baseNode = (*it)->intersection.nodeA;
				sn1.splitRatio = (*it)->intersection.splitRatioA;
				sn1.newNode = &((*it)->newNodeA);
				sortVector.push_back(sn1);
				SplitNode sn2;
				sn2.baseNode = dupNodeMap[(*it)->intersection.nodeB];
				sn2.splitRatio = (*it)->intersection.splitRatioB;
				sn2.newNode = &((*it)->dupNewNodeB);
				sortVector.push_back(sn2);
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					SplitNode sn1;
					sn1.baseNode = (*it)->intersection.nodeB;
					sn1.splitRatio = (*it)->intersection.splitRatioB;
					sn1.newNode = &((*it)->newNodeB);
					sortVector.push_back(sn1);
					SplitNode sn2;
					sn2.baseNode = dupNodeMap[(*it)->intersection.nodeA];
					sn2.splitRatio = (*it)->intersection.splitRatioA;
					sn2.newNode = &((*it)->dupNewNodeA);
					sortVector.push_back(sn2);
					SplitNode sn3;
					sn3.baseNode = dupNodeMap[(*it)->intersection.nodeB];
					sn3.splitRatio = (*it)->intersection.splitRatioB;
					sn3.newNode = &((*it)->dupNewNodeB);
					sortVector.push_back(sn3);
				} else {
					SplitNode sn1;
					sn1.baseNode = (*it)->intersection.nodeA;
					sn1.splitRatio = (*it)->intersection.splitRatioA;
					sn1.newNode = &((*it)->newNodeA);
					sortVector.push_back(sn1);
					SplitNode sn2;
					sn2.baseNode = dupNodeMap[(*it)->intersection.nodeB];
					sn2.splitRatio = (*it)->intersection.splitRatioB;
					sn2.newNode = &((*it)->dupNewNodeB);
					sortVector.push_back(sn2);
					SplitNode sn3;
					sn3.baseNode = dupNodeMap[(*it)->intersection.nodeA];
					sn3.splitRatio = (*it)->intersection.splitRatioA;
					sn3.newNode = &((*it)->dupNewNodeA);
					sortVector.push_back(sn3);
				}
			}
		}
			
	}
	// split at the calculated points in sorted order to have correct nodes to split
	sort(sortVector.begin(), sortVector.end());
	//g_printerr("Sorted splits are:\n");
	BezierNode *base = NULL;
	BezierNode *curve = NULL;
	gdouble lastSplit = 0;
	for (vector<SplitNode>::iterator it = sortVector.begin(); it != sortVector.end(); it++) {
		//g_printerr("n=%lX sr=%f\n", (long) (*it).baseNode, (*it).splitRatio);
		
		if (base != (*it).baseNode) {
			base = (*it).baseNode;
			curve = base;
			lastSplit = 0;
			//g_printerr("New base\n");
		}
		gdouble ratio = ((*it).splitRatio - lastSplit) / (1.0 - lastSplit);
		BezierNode *bnode;
		if (ratio > 0) {
			bnode = curve->subDivide(ratio);
		} else {
			bnode = curve;
		}
		*((*it).newNode) = bnode;
		curve = bnode;
		lastSplit = (*it).splitRatio;
		//g_printerr("Split curve: split node=%lX ratio=%f\n", (long) bnode, ratio);
		
	}
	// calculate cover status
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverage == 0 || godMap[(*it)->objectA] > godMap[(*it)->coverObject] && godMap[(*it)->objectB] > godMap[(*it)->coverObject]) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				(*it)->statusA = getStatus((*it)->dupNewNodeA, (*it)->objectB);
				(*it)->statusB = getStatus((*it)->newNodeB, (*it)->objectA);
			} else {
				(*it)->statusA = getStatus((*it)->newNodeA, (*it)->objectB);
				(*it)->statusB = getStatus((*it)->dupNewNodeB, (*it)->objectA);
			}
		}
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				(*it)->addNodeB = splitNode((*it)->newNodeB);
				(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
			} else {
				(*it)->addNodeA = splitNode((*it)->newNodeA);
				(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					(*it)->addNodeB = splitNode((*it)->newNodeB);
					(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
					(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
				} else {
					(*it)->addNodeA = splitNode((*it)->newNodeA);
					(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
					(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
				}
			}
		}
	}
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				BezierNode *odPA, *oPB, *cdPA, *cPB;
				if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					odPA = (*it)->dupAddNodeA;
					cdPA = (*it)->dupNewNodeA;
				} else {
					odPA = (*it)->dupNewNodeA;
					cdPA = (*it)->dupAddNodeA;
				}
				if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					oPB = (*it)->addNodeB;
					cPB = (*it)->newNodeB;
				} else {
					oPB = (*it)->newNodeB;
					cPB = (*it)->addNodeB;
				}
				NodePair np1(oPB, cdPA, (*it)->objectB, godMap[(*it)->objectB] );
				nodePairs.push_back(np1);
				removePaths.insert(odPA->path);
				removePaths.insert(cPB->path);
			} else {
				BezierNode *odPB, *oPA, *cdPB, *cPA;
				if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					odPB = (*it)->dupAddNodeB;
					cdPB = (*it)->dupNewNodeB;
				} else {
					odPB = (*it)->dupNewNodeB;
					cdPB = (*it)->dupAddNodeB;
				}
				if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					oPA = (*it)->addNodeA;
					cPA = (*it)->newNodeA;
				} else {
					oPA = (*it)->newNodeA;
					cPA = (*it)->addNodeA;
				}
				NodePair np1(oPA, cdPB, (*it)->objectA, godMap[(*it)->objectA] );
				nodePairs.push_back(np1);
				removePaths.insert(odPB->path);
				removePaths.insert(cPA->path);
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					BezierNode *odPA, *odPB, *oPB, *cdPA, *cdPB, *cPB;
					if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						odPA = (*it)->dupAddNodeA;
						cdPA = (*it)->dupNewNodeA;
					} else {
						odPA = (*it)->dupNewNodeA;
						cdPA = (*it)->dupAddNodeA;
					}
					if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						oPB = (*it)->addNodeB;
						cPB = (*it)->newNodeB;
						odPB = (*it)->dupAddNodeB;
						cdPB = (*it)->dupNewNodeB;
					} else {
						oPB = (*it)->newNodeB;
						cPB = (*it)->addNodeB;
						odPB = (*it)->dupNewNodeB;
						cdPB = (*it)->dupAddNodeB;
					}
					NodePair np1(odPA, odPB, (*it)->coverObject, coverGOD);
					nodePairs.push_back(np1);
					NodePair np2(oPB, cdPA, (*it)->objectB, godB);
					nodePairs.push_back(np2);
					removePaths.insert(cdPB->path);
					removePaths.insert(cPB->path);
				} else {
					BezierNode *odPA, *odPB, *oPA, *cdPA, *cdPB, *cPA;
					if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						odPB = (*it)->dupAddNodeB;
						cdPB = (*it)->dupNewNodeB;
					} else {
						odPB = (*it)->dupNewNodeB;
						cdPB = (*it)->dupAddNodeB;
					}
					if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						oPA = (*it)->addNodeA;
						cPA = (*it)->newNodeA;
						odPA = (*it)->dupAddNodeA;
						cdPA = (*it)->dupNewNodeA;
					} else {
						oPA = (*it)->newNodeA;
						cPA = (*it)->addNodeA;
						odPA = (*it)->dupNewNodeA;
						cdPA = (*it)->dupAddNodeA;
					}
					NodePair np1(odPA, odPB, (*it)->coverObject, coverGOD);
					nodePairs.push_back(np1);
					NodePair np2(oPA, cdPB, (*it)->objectA, godA);
					nodePairs.push_back(np2);
					removePaths.insert(cdPA->path);
					removePaths.insert(cPA->path);
				}
			}
		}
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleDivide(GDDocument *doc) {
	// get intersections
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	// calculate covering
	DBG(g_printerr("PathMingler::pathMingleDivide CALCULATE COVERING\n");)
	calcCoverage(intersections, godMap);
	// split paths
	DBG(g_printerr("PathMingler::pathMingleDivide SPLIT PATHS\n");)
	splitPaths(intersections);
	// duplicate inner paths
	DBG(g_printerr("PathMingler::pathMingleDivide DUPLICATE INNER PATHS\n");)
	BezierPathMap dupPathMap;
	duplicateInnerPaths(intersections, dupPathMap, doc);
	// connect paths
	DBG(g_printerr("PathMingler::pathMingleDivide CONNECT PATHS\n");)
	NodePairVector nodePairs;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("PathMingler::pathMingleDivide: Point: %f/%f statusA:%d statusB:%d cov:%d\n", (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y, (*it)->statusA, (*it)->statusB, (*it)->coverage);)
		DBG(g_printerr("objA=%lX objB=%lX nNA=%lX aNA=%lX nNB=%lX aNB=%lx\n", (long) (*it)->objectA, (long) (*it)->objectB, (long) (*it)->newNodeA, (long) (*it)->addNodeA, (long) (*it)->newNodeB, (long) (*it)->addNodeB);)
		BezierNode *iNodeA;
		BezierNode *iNodeB;
		BezierNode *cPA, *oPA, *cPB, *oPB;
		orderPaths(*it, &cPA, &oPA, &cPB, &oPB);
		if ((*it)->coverage) {
			GObject *styleObject = (*it)->objectA;
			gint depth = 0;
			if ((*it)->coverObject && godMap[(*it)->coverObject] > godMap[(*it)->objectA]) {
				styleObject = (*it)->coverObject;
				depth = godMap[(*it)->coverObject];
			}
			NodePair np1( oPA, oPB, styleObject, depth );
			nodePairs.push_back(np1);
			NodePair np2( getDupNode(oPA, dupPathMap), cPB, styleObject, depth );
			nodePairs.push_back(np2);
			NodePair np3( cPA, getDupNode(oPB, dupPathMap), styleObject, depth );
			nodePairs.push_back(np3);
		} else {
			NodePair np1(oPA, cPB, (*it)->objectA, godMap[(*it)->objectA] );
			nodePairs.push_back(np1);
			NodePair np2( oPB, cPA, (*it)->objectB, godMap[(*it)->objectB] );
			nodePairs.push_back(np2);
		}
		iNodeA = getDupNode(cPA, dupPathMap);
		iNodeB = getDupNode(cPB, dupPathMap);
		GObject *styleObject;
		if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
			styleObject = (*it)->objectA;
		} else {
			styleObject = (*it)->objectB;
		}
		NodePair np(iNodeA, iNodeB, styleObject, godMap[styleObject]);
		nodePairs.push_back(np);
		DBG(g_printerr("PathMingler::pathMingleDivide DP: nA=%lX nB=%lX\n", (long) iNodeA, (long) iNodeB);)
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleOutline(GDDocument *doc) {
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		// set outline only A part
		GDStyle transparent;
		transparent.drawType = GDSTYLE_DRAWTYPE_TRANSPARENT;
		transparent.color.colorType = GDCOLOR_TYPE_RGB;
		transparent.valueMask = GDSTYLE_DRAWTYPE_MASK;
		(*it)->newNodeA->getGObject()->setFillStyle(transparent);
		// and for B part
		(*it)->newNodeB->getGObject()->setFillStyle(transparent);
		// split up A part
		BezierNode *newNode = (BezierNode *) (*it)->newNodeA->clone();
		newNode->next = NULL;
		newNode->prev = NULL;
		(*it)->newNodeA->path->nodes->insert((*it)->newNodeA, newNode);
		(*it)->newNodeA->path->deleteCurve((*it)->newNodeA);
		// split up B part
		newNode = (BezierNode *) (*it)->newNodeB->clone();
		newNode->next = NULL;
		newNode->prev = NULL;
		(*it)->newNodeB->path->nodes->insert((*it)->newNodeB, newNode);
		(*it)->newNodeB->path->deleteCurve((*it)->newNodeB);
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleMerge(GDDocument *doc) {
	GObjectDepthMap godMap;
	ObjectIntersectionVector intersections = getISVector(doc, godMap);
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}

void PathMingler::pathMingleCrop(GDDocument *doc) {
	GObjectDepthMap godMap;
	ISObjectVector objectPaths = getIntersectionObjects(doc, godMap);
	GObject *topObject = objectPaths.back()->object;
	// get all intersections
	ObjectIntersectionVector iss = calcIntersections(objectPaths);
	GObjectDepthMap relevantObjects;
	// mark topObject as relevant
	relevantObjects[topObject] = 1;
	// check all intersections for topObject
	ObjectIntersectionVector intersections;
	for (ObjectIntersectionVector::iterator it = iss.begin(); it != iss.end(); it++) {
		(*it)->coverage = getCoverage(*it, godMap, &((*it)->coverObject));
		DBG(g_printerr("Coverage of %lX <-> %lX = %d (%f/%f)\n", (long) (*it)->objectA, (long) (*it)->objectB, (*it)->coverage, (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y);)
		if ((*it)->objectA == topObject && ((*it)->coverage == 0 || godMap[(*it)->coverObject] < godMap[(*it)->objectB])) {
			relevantObjects[(*it)->objectB] = 1;
			intersections.push_back(*it);
		} else if ((*it)->objectB == topObject && ((*it)->coverage == 0 || godMap[(*it)->coverObject] < godMap[(*it)->objectA])) {
			relevantObjects[(*it)->objectA] = 1;
			intersections.push_back(*it);
		} else if ((*it)->coverage == 1 && (*it)->coverObject == topObject) {
			relevantObjects[(*it)->objectA] = 1;
			relevantObjects[(*it)->objectB] = 1;
			intersections.push_back(*it);
		} else {
			delete (*it);
		}
	}
	// check all remaining objects for coverage by topObject
	for (ISObjectVector::iterator it = objectPaths.begin(); it != objectPaths.end(); it++) {
		if (relevantObjects.find((*it)->object) == relevantObjects.end()) {
			for (BezierPathVector::iterator bpit = (*it)->paths->begin(); bpit != (*it)->paths->end(); bpit++) {
				position p = (*bpit)->nodes->onCurve;
				if (topObject->checkArea(p.x, p.y)) {
					relevantObjects[(*it)->object] = 1;
					break;
				}
			}
		}
	}
	// remove topObject from coverage calculation
	godMap[topObject] = -1;
	// duplicate all paths
	BezierNodeMap dupNodeMap;
	BezierPathMap dupPathMap;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverObject == topObject) {
			DBG(g_printerr("coverage with topObject: %d\n", (*it)->coverage);)
			(*it)->coverage = getCoverage(*it, godMap, &((*it)->coverObject));
			(*it)->coverage--;
			DBG(g_printerr("coverObject after removal of topObject: %lX\n", (long) (*it)->coverObject);)
		}
		if ((*it)->objectA != topObject && (*it)->objectB != topObject) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				BezierPath *pathA = (*it)->intersection.nodeA->path;
				BezierPathMap::iterator bpmitA;
				if ((bpmitA = dupPathMap.find(pathA)) == dupPathMap.end()) {
					BezierPath *duplicatePathA = (BezierPath *)pathA->clone();
					doc->findObjectLayer(pathA)->addObject(duplicatePathA);
					dupPathMap[pathA] = duplicatePathA;
					BezierNode *oN = pathA->nodes;
					BezierNode *dN = duplicatePathA->nodes;
					while (oN) {
						dupNodeMap[oN] = dN;
						oN = oN->next;
						dN = dN->next;
						if (oN == pathA->nodes) {
							oN = NULL;
						}
					}
				}
			} else {
				BezierPath *pathB = (*it)->intersection.nodeB->path;
				BezierPathMap::iterator bpmitB;
				if ((bpmitB = dupPathMap.find(pathB)) == dupPathMap.end()) {
					BezierPath *duplicatePathB = (BezierPath *)pathB->clone();
					doc->findObjectLayer(pathB)->addObject(duplicatePathB);
					dupPathMap[pathB] = duplicatePathB;
					BezierNode *oN = pathB->nodes;
					BezierNode *dN = duplicatePathB->nodes;
					while (oN) {
						dupNodeMap[oN] = dN;
						oN = oN->next;
						dN = dN->next;
						if (oN == pathB->nodes) {
							oN = NULL;
						}
					}
				}
			}
		}
	}
	// clean up
	for (ISObjectVector::iterator it = objectPaths.begin(); it != objectPaths.end(); it++) {
		if (relevantObjects.find((*it)->object) == relevantObjects.end()) {
			// object is not used, delete it
			DBG(g_printerr("Remove Object: %lX\n", (long) (*it)->object);)
			(*it)->object->doc->removeObject((*it)->object);
			delete (*it)->object;
		}
		delete (*it)->paths;
		delete *it;
	}
	// calculate all necessary new split nodes
	DBG(g_printerr("PathMingler::pathMingleCrop CalcSplitNodes\n");)
	vector<SplitNode> sortVector;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("cov=%d A=%lX B=%lX tO=%lX srA=%f srB=%f nNA=%lX nNB=%lX\n", (*it)->coverage, (long) (*it)->objectA, (long) (*it)->objectB, (long) topObject, (*it)->intersection.splitRatioA, (*it)->intersection.splitRatioB, (long) (*it)->newNodeA, (long) (*it)->newNodeB);)
		DBG(g_printerr("godA=%d godB=%d godC=%d dupPA=%lX dupPB=%lX cObj=%lX (%f/%f)\n", godMap[(*it)->objectA], godMap[(*it)->objectB], godMap[(*it)->coverObject], (long) dupNodeMap[(*it)->intersection.nodeA], (long) dupNodeMap[(*it)->intersection.nodeB], (long) (*it)->coverObject, (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y);)
		DBG(g_printerr("nA=%lX nB=%lX dnA=%lX dnB=%lX\n", (long) (*it)->intersection.nodeA, (long) (*it)->intersection.nodeB, (long) dupNodeMap[(*it)->intersection.nodeA], (long) dupNodeMap[(*it)->intersection.nodeB]);)
		if ((*it)->coverage == 0 && ((*it)->objectA == topObject || (*it)->objectB == topObject)) {
			SplitNode sn1;
			sn1.baseNode = (*it)->intersection.nodeB;
			sn1.splitRatio = (*it)->intersection.splitRatioB;
			sn1.newNode = &((*it)->newNodeB);
			sortVector.push_back(sn1);
			SplitNode sn2;
			sn2.baseNode = (*it)->intersection.nodeA;
			sn2.splitRatio = (*it)->intersection.splitRatioA;
			sn2.newNode = &((*it)->newNodeA);
			sortVector.push_back(sn2);
			DBG(g_printerr("TOPOBJ involved, cov=0\n");)
		} else if ((*it)->objectA == topObject && godMap[(*it)->coverObject] < godMap[(*it)->objectB]) {
			SplitNode sn1;
			sn1.baseNode = (*it)->intersection.nodeB;
			sn1.splitRatio = (*it)->intersection.splitRatioB;
			sn1.newNode = &((*it)->newNodeB);
			sortVector.push_back(sn1);
			SplitNode sn2;
			sn2.baseNode = (*it)->intersection.nodeA;
			sn2.splitRatio = (*it)->intersection.splitRatioA;
			sn2.newNode = &((*it)->newNodeA);
			sortVector.push_back(sn2);
			SplitNode sn3;
			sn3.baseNode = dupNodeMap[(*it)->intersection.nodeB];
			sn3.splitRatio = (*it)->intersection.splitRatioB;
			sn3.newNode = &((*it)->dupNewNodeB);
			sortVector.push_back(sn3);
			DBG(g_printerr("TOPOBJ involved, objectB\n");)
		} else if ((*it)->objectB == topObject && godMap[(*it)->coverObject] < godMap[(*it)->objectA]) {
			SplitNode sn1;
			sn1.baseNode = (*it)->intersection.nodeB;
			sn1.splitRatio = (*it)->intersection.splitRatioB;
			sn1.newNode = &((*it)->newNodeB);
			sortVector.push_back(sn1);
			SplitNode sn2;
			sn2.baseNode = (*it)->intersection.nodeA;
			sn2.splitRatio = (*it)->intersection.splitRatioA;
			sn2.newNode = &((*it)->newNodeA);
			sortVector.push_back(sn2);
			SplitNode sn3;
			sn3.baseNode = dupNodeMap[(*it)->intersection.nodeA];
			sn3.splitRatio = (*it)->intersection.splitRatioA;
			sn3.newNode = &((*it)->dupNewNodeA);
			sortVector.push_back(sn3);
			DBG(g_printerr("TOPOBJ involved, objectA\n");)
		} else if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				SplitNode sn1;
				sn1.baseNode = (*it)->intersection.nodeB;
				sn1.splitRatio = (*it)->intersection.splitRatioB;
				sn1.newNode = &((*it)->newNodeB);
				sortVector.push_back(sn1);
				SplitNode sn2;
				sn2.baseNode = dupNodeMap[(*it)->intersection.nodeA];
				sn2.splitRatio = (*it)->intersection.splitRatioA;
				sn2.newNode = &((*it)->dupNewNodeA);
				sortVector.push_back(sn2);
				DBG(g_printerr("NodeB, dupNodeA\n");)
			} else {
				SplitNode sn1;
				sn1.baseNode = (*it)->intersection.nodeA;
				sn1.splitRatio = (*it)->intersection.splitRatioA;
				sn1.newNode = &((*it)->newNodeA);
				sortVector.push_back(sn1);
				SplitNode sn2;
				sn2.baseNode = dupNodeMap[(*it)->intersection.nodeB];
				sn2.splitRatio = (*it)->intersection.splitRatioB;
				sn2.newNode = &((*it)->dupNewNodeB);
				sortVector.push_back(sn2);
				DBG(g_printerr("NodeA, dupNodeB\n");)
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					SplitNode sn1;
					sn1.baseNode = (*it)->intersection.nodeB;
					sn1.splitRatio = (*it)->intersection.splitRatioB;
					sn1.newNode = &((*it)->newNodeB);
					sortVector.push_back(sn1);
					SplitNode sn2;
					sn2.baseNode = dupNodeMap[(*it)->intersection.nodeA];
					sn2.splitRatio = (*it)->intersection.splitRatioA;
					sn2.newNode = &((*it)->dupNewNodeA);
					sortVector.push_back(sn2);
					SplitNode sn3;
					sn3.baseNode = dupNodeMap[(*it)->intersection.nodeB];
					sn3.splitRatio = (*it)->intersection.splitRatioB;
					sn3.newNode = &((*it)->dupNewNodeB);
					sortVector.push_back(sn3);
					DBG(g_printerr("NodeB, dupNodeA, dupNodeB\n");)
				} else {
					SplitNode sn1;
					sn1.baseNode = (*it)->intersection.nodeA;
					sn1.splitRatio = (*it)->intersection.splitRatioA;
					sn1.newNode = &((*it)->newNodeA);
					sortVector.push_back(sn1);
					SplitNode sn2;
					sn2.baseNode = dupNodeMap[(*it)->intersection.nodeB];
					sn2.splitRatio = (*it)->intersection.splitRatioB;
					sn2.newNode = &((*it)->dupNewNodeB);
					sortVector.push_back(sn2);
					SplitNode sn3;
					sn3.baseNode = dupNodeMap[(*it)->intersection.nodeA];
					sn3.splitRatio = (*it)->intersection.splitRatioA;
					sn3.newNode = &((*it)->dupNewNodeA);
					sortVector.push_back(sn3);
					DBG(g_printerr("NodeA, dupNodeA, dupNodeB\n");)
				}
			}
		}
			
	}
	// split at the calculated points in sorted order to have correct nodes to split
	sort(sortVector.begin(), sortVector.end());
	DBG(g_printerr("Sorted splits are:\n");)
	BezierNode *base = NULL;
	BezierNode *curve = NULL;
	gdouble lastSplit = 0;
	for (vector<SplitNode>::iterator it = sortVector.begin(); it != sortVector.end(); it++) {
		DBG(g_printerr("n=%lX sr=%f\n", (long) (*it).baseNode, (*it).splitRatio);)
		
		if (base != (*it).baseNode) {
			base = (*it).baseNode;
			curve = base;
			lastSplit = 0;
			//g_printerr("New base\n");
		}
		gdouble ratio = ((*it).splitRatio - lastSplit) / (1.0 - lastSplit);
		BezierNode *bnode;
		if (ratio > 0) {
			bnode = curve->subDivide(ratio);
		} else {
			bnode = curve;
		}
		*((*it).newNode) = bnode;
		curve = bnode;
		lastSplit = (*it).splitRatio;
		DBG(g_printerr("Split curve: split node=%lX ratio=%f\n", (long) bnode, ratio);)
		
	}
	// calculate cover status
	DBG(g_printerr("PathMingler::pathMingleCrop get cover status:\n");)
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		if ((*it)->coverage == 0 || ((*it)->objectA == topObject || godMap[(*it)->objectA] > godMap[(*it)->coverObject]) && 
									((*it)->objectB == topObject ||godMap[(*it)->objectB] > godMap[(*it)->coverObject])) {
			DBG(g_printerr("oA=%lX oB=%lX\n", (long) (*it)->objectA, (long) (*it)->objectB);)
			if ((*it)->objectA == topObject || (*it)->objectB == topObject) {
				(*it)->statusA = getStatus((*it)->newNodeA, (*it)->objectB);
				(*it)->statusB = getStatus((*it)->newNodeB, (*it)->objectA);
			} else if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				(*it)->statusA = getStatus((*it)->dupNewNodeA, (*it)->objectB);
				(*it)->statusB = getStatus((*it)->newNodeB, (*it)->objectA);
			} else {
				(*it)->statusA = getStatus((*it)->newNodeA, (*it)->objectB);
				(*it)->statusB = getStatus((*it)->dupNewNodeB, (*it)->objectA);
			}
		}
	}
	// split up where necessary
	DBG(g_printerr("PathMingler::pathMingleCrop split paths:\n");)
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("oA=%lX oB=%lX cov=%d (%f/%f)\n", (long) (*it)->objectA, (long) (*it)->objectB, (*it)->coverage, (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y);)
		if ((*it)->coverage == 0 && ((*it)->objectA == topObject || (*it)->objectB == topObject)) {
			(*it)->addNodeA = splitNode((*it)->newNodeA);
			(*it)->addNodeB = splitNode((*it)->newNodeB);
		} else if ((*it)->objectA == topObject && godMap[(*it)->coverObject] < godMap[(*it)->objectB]) {
			(*it)->addNodeA = splitNode((*it)->newNodeA);
			(*it)->addNodeB = splitNode((*it)->newNodeB);
			(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
		} else if ((*it)->objectB == topObject && godMap[(*it)->coverObject] < godMap[(*it)->objectA]) {
			(*it)->addNodeA = splitNode((*it)->newNodeA);
			(*it)->addNodeB = splitNode((*it)->newNodeB);
			(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
		} else if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				(*it)->addNodeB = splitNode((*it)->newNodeB);
				(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
			} else {
				(*it)->addNodeA = splitNode((*it)->newNodeA);
				(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					(*it)->addNodeB = splitNode((*it)->newNodeB);
					(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
					(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
				} else {
					(*it)->addNodeA = splitNode((*it)->newNodeA);
					(*it)->dupAddNodeB = splitNode((*it)->dupNewNodeB);
					(*it)->dupAddNodeA = splitNode((*it)->dupNewNodeA);
				}
			}
		}
	}
	// connect nodes and remove unused paths
	DBG(g_printerr("PathMingler::pathMingleCrop connect nodes:\n");)
	NodePairVector nodePairs;
	BezierPathSet removePaths;
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		DBG(g_printerr("oA=%lX oB=%lX cov=%d (%f/%f)\n", (long) (*it)->objectA, (long) (*it)->objectB, (*it)->coverage, (*it)->intersection.intersectPos.x, (*it)->intersection.intersectPos.y);)
		DBG(g_printerr("gA=%d gB=%d, gCO=%d\n", godMap[(*it)->objectA], godMap[(*it)->objectB], godMap[(*it)->coverObject]);)
		DBG(g_printerr("PnNA=%lX PaNA=%lX PnNB=%lX PaNB=%lX\n", (*it)->newNodeA ? (long) (*it)->newNodeA->path : 0, (*it)->addNodeA ? (long) (*it)->addNodeA->path : 0, (*it)->newNodeB ? (long) (*it)->newNodeB->path : 0, (*it)->addNodeB ? (long) (*it)->addNodeB->path : 0);)
		DBG(g_printerr("PdnNA=%lX PdaNA=%lX PdnNB=%lX PdaNB=%lX\n", (*it)->dupNewNodeA ? (long) (*it)->dupNewNodeA->path : 0, (*it)->dupAddNodeA ? (long) (*it)->dupAddNodeA->path : 0, (*it)->dupNewNodeB ? (long) (*it)->dupNewNodeB->path : 0, (*it)->dupAddNodeB ? (long) (*it)->dupAddNodeB->path : 0);)
		if (((*it)->objectA == topObject && ((*it)->coverage == 0 || godMap[(*it)->coverObject] < godMap[(*it)->objectB])) ||
			((*it)->objectB == topObject && ((*it)->coverage == 0 || godMap[(*it)->coverObject] < godMap[(*it)->objectA]))) {
			BezierNode *oPA, *oPB, *cPA, *cPB, *doPA, *doPB, *dcPA, *dcPB;
			if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
				oPA = (*it)->addNodeA;
				cPA = (*it)->newNodeA;
				doPA = (*it)->dupAddNodeA;
				dcPA = (*it)->dupNewNodeA;
			} else {
				oPA = (*it)->newNodeA;
				cPA = (*it)->addNodeA;
				doPA = (*it)->dupNewNodeA;
				dcPA = (*it)->dupAddNodeA;
			}
			if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
				oPB = (*it)->addNodeB;
				cPB = (*it)->newNodeB;
				doPB = (*it)->dupAddNodeB;
				dcPB = (*it)->dupNewNodeB;
			} else {
				oPB = (*it)->newNodeB;
				cPB = (*it)->addNodeB;
				doPB = (*it)->dupNewNodeB;
				dcPB = (*it)->dupAddNodeB;
			}
			GObject *styleObject = ((*it)->objectA == topObject) ? (*it)->objectB : (*it)->objectA;
			NodePair np1(cPB, cPA, styleObject, godMap[styleObject] );
			nodePairs.push_back(np1);
			if ((*it)->coverage == 0) {
				removePaths.insert(oPA->path);
				removePaths.insert(oPB->path);
			} else if ((*it)->objectA == topObject) {
				NodePair np2(oPA, dcPB, (*it)->coverObject, godMap[(*it)->coverObject]);
				nodePairs.push_back(np2);
				removePaths.insert(doPB->path);
				removePaths.insert(oPB->path);
			} else if ((*it)->objectB == topObject) {
				NodePair np2(oPB, dcPA, (*it)->coverObject, godMap[(*it)->coverObject]);
				nodePairs.push_back(np2);
				removePaths.insert(doPA->path);
				removePaths.insert(oPA->path);
			}
		} else if ((*it)->coverage == 0) {
			if (godMap[(*it)->objectA] > godMap[(*it)->objectB]) {
				BezierNode *odPA, *oPB, *cdPA, *cPB;
				if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					odPA = (*it)->dupAddNodeA;
					cdPA = (*it)->dupNewNodeA;
				} else {
					odPA = (*it)->dupNewNodeA;
					cdPA = (*it)->dupAddNodeA;
				}
				if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					oPB = (*it)->addNodeB;
					cPB = (*it)->newNodeB;
				} else {
					oPB = (*it)->newNodeB;
					cPB = (*it)->addNodeB;
				}
				NodePair np1(oPB, cdPA, (*it)->objectB, godMap[(*it)->objectB] );
				nodePairs.push_back(np1);
				removePaths.insert(odPA->path);
				removePaths.insert(cPB->path);
			} else {
				BezierNode *odPB, *oPA, *cdPB, *cPA;
				if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					odPB = (*it)->dupAddNodeB;
					cdPB = (*it)->dupNewNodeB;
				} else {
					odPB = (*it)->dupNewNodeB;
					cdPB = (*it)->dupAddNodeB;
				}
				if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
					oPA = (*it)->addNodeA;
					cPA = (*it)->newNodeA;
				} else {
					oPA = (*it)->newNodeA;
					cPA = (*it)->addNodeA;
				}
				NodePair np1(oPA, cdPB, (*it)->objectA, godMap[(*it)->objectA] );
				nodePairs.push_back(np1);
				removePaths.insert(odPB->path);
				removePaths.insert(cPA->path);
			}
		} else {
			int coverGOD = godMap[(*it)->coverObject];
			int godA = godMap[(*it)->objectA];
			int godB = godMap[(*it)->objectB];
			if (godA > coverGOD && godB > coverGOD) {
				if (godA > godB) {
					BezierNode *odPA, *odPB, *oPB, *cdPA, *cdPB, *cPB;
					if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						odPA = (*it)->dupAddNodeA;
						cdPA = (*it)->dupNewNodeA;
					} else {
						odPA = (*it)->dupNewNodeA;
						cdPA = (*it)->dupAddNodeA;
					}
					if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						oPB = (*it)->addNodeB;
						cPB = (*it)->newNodeB;
						odPB = (*it)->dupAddNodeB;
						cdPB = (*it)->dupNewNodeB;
					} else {
						oPB = (*it)->newNodeB;
						cPB = (*it)->addNodeB;
						odPB = (*it)->dupNewNodeB;
						cdPB = (*it)->dupAddNodeB;
					}
					NodePair np1(odPA, odPB, (*it)->coverObject, coverGOD);
					nodePairs.push_back(np1);
					NodePair np2(oPB, cdPA, (*it)->objectB, godB);
					nodePairs.push_back(np2);
					removePaths.insert(cdPB->path);
					removePaths.insert(cPB->path);
				} else {
					BezierNode *odPA, *odPB, *oPA, *cdPA, *cdPB, *cPA;
					if ((*it)->statusB == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						odPB = (*it)->dupAddNodeB;
						cdPB = (*it)->dupNewNodeB;
					} else {
						odPB = (*it)->dupNewNodeB;
						cdPB = (*it)->dupAddNodeB;
					}
					if ((*it)->statusA == PATHMINGLER_FOREIGN_COVER_PREVIOUS) {
						oPA = (*it)->addNodeA;
						cPA = (*it)->newNodeA;
						odPA = (*it)->dupAddNodeA;
						cdPA = (*it)->dupNewNodeA;
					} else {
						oPA = (*it)->newNodeA;
						cPA = (*it)->addNodeA;
						odPA = (*it)->dupNewNodeA;
						cdPA = (*it)->dupAddNodeA;
					}
					NodePair np1(odPA, odPB, (*it)->coverObject, coverGOD);
					nodePairs.push_back(np1);
					NodePair np2(oPA, cdPB, (*it)->objectA, godA);
					nodePairs.push_back(np2);
					removePaths.insert(cdPA->path);
					removePaths.insert(cPA->path);
				}
			}
		}
	}
	sort(nodePairs.begin(), nodePairs.end());
	connectNodePairs(nodePairs);
	// remove unused paths
	DBG(g_printerr("PathMingler::pathMingleCrop remove unused paths\n");)
	for (BezierPathSet::iterator it = removePaths.begin(); it != removePaths.end(); it++) {
		DBG(g_printerr("Path: %lX\n", (long) *it);)
		(*it)->doc->removeObject(*it);
		delete *it;
	}
	for (ObjectIntersectionVector::iterator it = intersections.begin(); it != intersections.end(); it++) {
		delete *it;
	}
	doc->redraw();
}
