/* $Id: Preferences.cc,v 1.13 2004/03/24 19:31:43 pklaus Exp $
 *
 * Preferences.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "Preferences.hh"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>

#include <libgnome/libgnome.h>  /* for _() macro */

class PreferencesXMLDirectory : public XMLOBJECT {

	string path;
	Preferences *prefs;
	
public:
	PreferencesXMLDirectory(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, Preferences *prefs) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->prefs = prefs;
	}
	void handleText(const char *text, int len) {
		path.append(text, len);
	}
	void handleEndTag(const char *name) {
		if (path.size()) {
			prefs->fontDir.insert(prefs->fontDir.end(), path);
		}
		XMLOBJECT::handleEndTag(name);
	}
};

class PreferencesXMLDirectories : public XMLOBJECT {
	PreferencesXMLDirectory *dir;
	Preferences *prefs;
public:
	PreferencesXMLDirectories(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, Preferences *prefs) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->prefs = prefs;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "directory")) {
			dir = new PreferencesXMLDirectory(this, name, parser, attrs, prefs);
			dir->init();
		}
	}
};

class PreferencesXMLFloatValue : public XMLOBJECT {

	gdouble *value;
	string valString;
public:
	PreferencesXMLFloatValue(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, gdouble *value) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->value = value;
	}
	void handleText(const char *text, int len) {
		valString.append(text, len);
	}
	void handleEndTag(const char *name) {
		if (valString.size()) {
			*value = g_strtod(valString.c_str(), NULL);
		}
		XMLOBJECT::handleEndTag(name);
	}
};

class PreferencesXMLIntegerValue : public XMLOBJECT {

	guint *value;
	string valString;
public:
	PreferencesXMLIntegerValue(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, guint *value) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->value = value;
	}
	void handleText(const char *text, int len) {
		valString.append(text, len);
	}
	void handleEndTag(const char *name) {
		if (valString.size()) {
			*value = atol(valString.c_str());
		}
		XMLOBJECT::handleEndTag(name);
	}
};

class PreferencesXMLStringValue : public XMLOBJECT {

	string *value;
	string valString;
public:
	PreferencesXMLStringValue(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, string *value) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->value = value;
	}
	void handleText(const char *text, int len) {
		valString.append(text, len);
	}
	void handleEndTag(const char *name) {
		if (valString.size()) {
			*value = valString;
		}
		XMLOBJECT::handleEndTag(name);
	}
};

class PreferencesXMLDocument : public XMLOBJECT {
	PreferencesXMLIntegerValue *intVal;
	PreferencesXMLFloatValue *floatVal;
	PreferencesXMLDirectories *fDirs;
	PreferencesXMLStringValue *strVal;
	Preferences *prefs;
public:
	PreferencesXMLDocument(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, Preferences *prefs) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->prefs = prefs;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "catching-range")) {
			floatVal = new PreferencesXMLFloatValue(this, name, parser, attrs, &(prefs->catchingRange));
			floatVal->init();
		}
		if (! strcmp(name, "ruler-unit-factor")) {
			floatVal = new PreferencesXMLFloatValue(this, name, parser, attrs, &(prefs->rulerUnitFactor));
			floatVal->init();
		}
		if (! strcmp(name, "textcursor-blinktime")) {
			intVal = new PreferencesXMLIntegerValue(this, name, parser, attrs, &(prefs->textCursorBlinkTime));
			intVal->init();
		}
		if (! strcmp(name, "undo-stack-size")) {
			intVal = new PreferencesXMLIntegerValue(this, name, parser, attrs, &(prefs->undoStackSize));
			intVal->init();
		}
		if (! strcmp(name, "page-format")) {
			strVal = new PreferencesXMLStringValue(this, name, parser, attrs, &(prefs->pageFormat));
			strVal->init();
		}
		if (! strcmp(name, "font-directories")) {
			fDirs = new PreferencesXMLDirectories(this, name, parser, attrs, prefs);
			fDirs->init();
		}
	}
};


class PreferencesXMLRoot : public XMLOBJECT {
	PreferencesXMLDocument *document;
	Preferences *prefs;
	
public:
	PreferencesXMLRoot(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, Preferences *prefs) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->prefs = prefs;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "gdprefs")) {
			document = new PreferencesXMLDocument(this, name, parser, attrs, prefs);
			document->init();
		}
	}
};

Preferences::Preferences() {
	catchingRange = 6;
	pageWidth = 595.27559;
	pageHeight = 841.88976;
	pageFormat = "a4";
	rulerUnitFactor = 1.0;
	fontDir.insert(fontDir.begin(), "/usr/X11R6/lib/X11/fonts/Type1");
	//fontDir.insert(fontDir.begin(), "/usr/share/fonts/default/Type1");
	textCursorBlinkTime = 500;
	string homeDir = getenv("HOME");
	prefDir = homeDir + "/.gestalter";
	undoStackSize = 10;
	units.push_back(_("point"));
	factors.push_back(1.0);
	units.push_back(_("inch"));
	factors.push_back(1.0 / 72);
	units.push_back(_("cm"));
	factors.push_back(2.54 / 72);
	units.push_back(_("mm"));
	factors.push_back(25.4 / 72);
	units.push_back(_("pica"));
	factors.push_back(1.0 / 6);
	unitToFactorMap["point"] =  1.0;
	unitToFactorMap["pt"] =  1.0;
	unitToFactorMap["pica"] =  1.0 / 6;
	unitToFactorMap["inch"] =  1.0 / 72;
	unitToFactorMap["zoll"] =  1.0 / 72;
	unitToFactorMap["\""] =  1.0 / 72;
	unitToFactorMap["feet"] =  1.0 / 72 / 12;
	unitToFactorMap["ft"] =  1.0 / 72 / 12;
	unitToFactorMap["meter"] =  0.0254 / 72;
	unitToFactorMap["m"] =  0.0254 / 72;
	unitToFactorMap["dm"] =  0.254 / 72;
	unitToFactorMap["cm"] =  2.54 / 72;
	unitToFactorMap["mm"] =  25.4 / 72;
	
	// create the popup menu
	int selected = 0;
	int i = 0;
	unitPopupMenu = new Gtk::Menu();
	Gtk::RadioMenuItem::Group group;
	for (vector<string>::iterator it = units.begin(); it != units.end(); it++, i++) {
		//g_printerr("PreferencesDialog::PreferencesDialog inserting unit %s (%d) -> %f\n", (*it).first.c_str(), i, (*it).second);
		Gtk::RadioMenuItem *unitItem = manage(new Gtk::RadioMenuItem(group, *it));
		unitItem->set_user_data((gpointer) i);
		unitPopupMenu->append(*unitItem);
		unitItem->show();
	}
	unitPopupMenu->show();
}


gint Preferences::createPrefDirectory() {
	struct stat statbuf;
	int ret = stat(prefDir.c_str(), &statbuf);
	if (ret) {
		ret = mkdir(prefDir.c_str(), 0700);
	}
	return ret;
}

void Preferences::save() {
	createPrefDirectory();
	string prefFile = prefDir;
	prefFile.append("/preferences");
	ofstream prefOut(prefFile.c_str());
	prefOut.setf(ios::fixed, ios::floatfield);
	prefOut.precision(12);
	prefOut << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
	prefOut << "<gdprefs>\n";
	prefOut << "<catching-range>" << catchingRange << "</catching-range>\n";
	prefOut << "<textcursor-blinktime>" << textCursorBlinkTime << "</textcursor-blinktime>\n";
	prefOut << "<undo-stack-size>" << undoStackSize << "</undo-stack-size>\n";
	prefOut << "<page-format>" << pageFormat << "</page-format>\n";
	prefOut << "<ruler-unit-factor>" << rulerUnitFactor << "</ruler-unit-factor>\n";
	prefOut << "<font-directories>\n";
	for (vector<string>::iterator it = fontDir.begin(); it != fontDir.end(); it++) {
		prefOut << "<directory>" << *it << "</directory>\n";
	}
	prefOut << "</font-directories>\n";
	prefOut << "</gdprefs>\n";
}

void Preferences::load() {
	createPrefDirectory();
	string prefFile = prefDir;
	prefFile.append("/preferences");
	struct stat statbuf;
	int ret = stat(prefFile.c_str(), &statbuf);
	if (ret) {
		//g_printerr("No preferences file, generating one\n");
		save();
	}
	fontDir.clear();
	XMLPARSER *xmlParser = new XMLPARSER();
	PreferencesXMLRoot *pxmlr = new PreferencesXMLRoot(NULL, "prefs", NULL, NULL, this);
	xmlParser->parse(pxmlr, prefFile);
	//g_printerr("Preferences::load prefs parsed, setting page format\n");
	setPageSizeFromFormat();
}

string Preferences::getPrefDirectory() {
	createPrefDirectory();
	return prefDir;
}

void Preferences::setPageSizeFromFormat() {
	//g_printerr("Preferences::setPageSizeFromFormat pageFormat=%s\n", pageFormat.c_str());
	if (pageFormat.compare("a4") == 0 || pageFormat.compare("A4") == 0) {
		pageWidth = 595.27559;
		pageHeight = 841.88976;
	} else if (pageFormat.compare("a5") == 0 || pageFormat.compare("A5") == 0) {
		pageWidth = 420.94488;
		pageHeight = 595.27559;
	} else if (pageFormat.compare("a6") == 0 || pageFormat.compare("A6") == 0) {
		pageWidth = 297.637795;
		pageHeight = 420.94488;
	} else if (pageFormat.compare("a3") == 0 || pageFormat.compare("A3") == 0) {
		pageWidth = 841.88976;
		pageHeight = 1190.55118;
	} else if (pageFormat.compare("a2") == 0 || pageFormat.compare("A2") == 0) {
		pageWidth = 1190.55118;
		pageHeight = 1683.77952;
	} else if (pageFormat.compare("a1") == 0 || pageFormat.compare("A1") == 0) {
		pageWidth = 1683.77952;
		pageHeight = 2381.10236;
	} else if (pageFormat.compare("a0") == 0 || pageFormat.compare("A0") == 0) {
		pageWidth = 2381.10236;
		pageHeight = 3367.55904;
	} else if (pageFormat.compare("letter") == 0 || pageFormat.compare("Letter") == 0) {
		pageWidth = 612;
		pageHeight = 792;
	} else if (pageFormat.compare("legal") == 0 || pageFormat.compare("Legal") == 0) {
		pageWidth = 612;
		pageHeight = 1008;
	}
	//g_printerr("Preferences::setPageSizeFromFormat pageSize after decoding: w=%f h=%f\n", pageWidth, pageHeight);
}

gdouble Preferences::unit2Factor(string unit) {
	for (uint i = 0; i < units.size(); i++) {
		if (units[i] == unit) {
			return factors[i];
		}
	}
	return 1.0;
}

gdouble Preferences::unitStrToDouble(string str, gboolean fallbackDefault) {
	char *endNum;
	gdouble numValue = strtod(str.c_str(), &endNum);
	string ustr = endNum;
	//g_printerr("Preferences::unitStrToDouble restStr=%s\n", ustr.c_str());
	if (ustr.size()) {
		string::size_type unitEnd = ustr.find_first_of(" \t");
		//g_printerr("Preferences::unitStrToDouble unitEnd=%d\n", unitEnd);
		if (unitEnd == string::npos) {
			// no white space at the end
			unitEnd = ustr.size();
		}
		string unit = ustr.substr(0, unitEnd);
		// simple makeLowerCase
		for (uint i = 0; i < unit.size(); i++) {
			if (unit[i] >= 'A' && unit[i] <= 'Z') {
				unit[i] += 32; // convert to lowercase character
			}
		}
		//g_printerr("Preferences::unitStrToDouble unit=%s\n", unit.c_str());
		gdouble factor = 1.0;
		if (fallbackDefault) {
			factor = rulerUnitFactor;
		}
		for (map<string, gdouble>::iterator it = unitToFactorMap.begin(); it != unitToFactorMap.end(); it++) {
			//g_printerr("Preferences::unitStrToDouble unit=[%s] check=[%s]\n", unit.c_str(), (*it).first.c_str());
			if ((*it).first == unit) {
				factor = (*it).second;
				break;
			}
		}
		numValue /= factor;
	} else {
		// no unit specified
		if (fallbackDefault) {
			numValue /= rulerUnitFactor;
		}
	}
	return numValue;
}

string Preferences::factor2Unit(gdouble factor){
	for (uint i = 0; i < factors.size(); i++) {
		if (factors[i] == factor) {
			return units[i];
		}
	}
	return string("unknown");
}
