/* $Id: SelectionList.cc,v 1.13 2004/07/21 06:31:20 pklaus Exp $
 *
 * SelectionList.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "SelectionList.hh"
#include "DrawObject.hh"
#include "Properties.hh"
#include "GDStyle.hh"
#include "TextStyle.hh"
#include "TextObject.hh"
#include "TextProperties.hh"
#include "GDFont.hh"

SelectionList::SelectionList(GDDocument *document) {
	objects = new G_List<DrawObject *>;
	doc = document;
}

void SelectionList::reset() {
	delete objects;
	objects = new G_List<DrawObject *>;
}

void SelectionList::insertObject(DrawObject *object) {
	//g_printerr("SelectionList::insertObject object:%lX\n", object);
	SelectionIterator it;
	for (it = objects->begin(); *it; it++) {
		if (*it == object) {
			return;
		}
	}
	objects->insert(objects->end(), object);
	changed(doc);
	if (object && strstr(object->identify(), "Text")) {
		textChanged(doc);
	}
}

void SelectionList::removeObject(DrawObject *object) {
	//g_printerr("SelectionList::removeObject object:%lX\n", object);
	objects->remove(object);
	changed(doc);
	if (object && strstr(object->identify(), "Text")) {
		textChanged(doc);
	}
}

SelectionIterator SelectionList::getIterator() {
	return objects->begin();
}

void SelectionList::draw(DocWindow *docwin, RectangleArea& drawArea) {
	SelectionIterator it;
	for (it = objects->begin(); *it; it++) {
		//g_printerr("SelectionList::draw item:%lX (%s)\n", *it, (*it)->identify());
		(*it)->drawWithSelection(docwin->osBuffer, drawArea);
	}
}

void SelectionList::drawWireFrame(DocWindow *docwin, RectangleArea& drawArea) {
	SelectionIterator it;
	for (it = objects->begin(); *it; it++) {
		(*it)->drawWireFrameSelection(docwin, drawArea);
	}
}

void SelectionList::setFillStyle(GDStyle *fromStyle) {
	SelectionIterator it = objects->begin();
	for (; *it; it++) {
		//g_printerr("SelectionList::setFillStyle fromStyle %lX oldStyle %lX obj %lX\n", fromStyle->valueMask, (*it)->getFillStyle()->valueMask, *it);
		(*it)->setFillStyle(*fromStyle);
	}
}
		
void SelectionList::setStrokeStyle(GDStyle *fromStyle) {
	SelectionIterator it = objects->begin();
	for (; *it; it++) {
		(*it)->setStrokeStyle(*fromStyle);
	}
}

void SelectionList::setTextStyle(TextStyle& style) {
	SelectionIterator it = objects->begin();
	for (; *it; it++) {
		(*it)->setTextStyle(style);
	}
}

void SelectionList::fillProperties(Properties *properties) {
	SelectionIterator it = objects->begin();
	guint strokeColor = 0xFF000000;
	guint fillColor = 0xFF000000;
	gint strokeDrawType = 0;
	gint fillDrawType = 0;
	gint strokeColorType = 0;
	gint fillColorType = 0;
	gdouble strokeWidth = -1;
	guint r,g,b;
	GDStyle *strokeStyle = NULL;
	GDStyle *fillStyle = NULL;
	int i = 0;
	while (*it) {
		if (strokeStyle) {
			GDStyle *oStyle = (*it)->getStrokeStyle();
			if (oStyle) {
				*strokeStyle &= *oStyle;
			}
		} else {
			GDStyle *oStyle = (*it)->getStrokeStyle();
			if (oStyle) strokeStyle = new GDStyle(*oStyle);
		}
		if (fillStyle) {
			GDStyle *oStyle = (*it)->getFillStyle();
			if (oStyle) {
				*fillStyle &= *oStyle;
			}
		} else {
			GDStyle *oStyle = (*it)->getFillStyle();
			if (oStyle) fillStyle = new GDStyle(*oStyle);
		}
		it++;
	}
	if (fillStyle) {
		//g_printerr("SelectionList::fillProperties fillStyle: %lX\n", fillStyle->valueMask);
		properties->setFillStyle(fillStyle);
		delete fillStyle;
	}
	if (strokeStyle) {
		properties->setStrokeStyle(strokeStyle);
		delete strokeStyle;
	}
}

gboolean SelectionList::selectionNotEmpty() {
	return objects->size() > 0;
}

void SelectionList::fillTextProperties(TextProperties *textProperties) {
	SelectionIterator it = objects->begin();
	gboolean initialized = false;
	TextStyle tStyle;
	for (; *it; it++) {
		//g_printerr("SelectionList::fillTextProperties %lX\n", *it);
		if (strstr((*it)->identify(), "Text")) {
			//g_printerr("SelectionList::fillTextProperties is TextObject\n");
			if (initialized) {
				tStyle &= ((TextObject *) *it)->getTextStyle();
			} else {
				tStyle = ((TextObject *) *it)->getTextStyle();
				initialized = true;
			}
		}
	}
	if (! initialized) {
		tStyle.setValueMask(0);
	}
	//g_printerr("SelectionList::fillTextProperties: valueMask: %lX, font %s, size %f, lspace %f\n", tStyle.getValueMask(), tStyle.getFont() ? tStyle.getFont()->getName() : "NULL", tStyle.getPointSize(), tStyle.getLineSpace());
	textProperties->fillFont(tStyle.getFont(), tStyle.getValueMask() & TEXT_STYLE_FONT_MASK);
	textProperties->fillPointSize(tStyle.getPointSize(), tStyle.getValueMask() & TEXT_STYLE_POINT_SIZE_MASK);
	textProperties->fillLineSpace(tStyle.getLineSpace(), tStyle.getValueMask() & TEXT_STYLE_LINE_SPACE_MASK);
	textProperties->fillAutoLineSpace(tStyle.getAutoLineSpace(), tStyle.getValueMask() & TEXT_STYLE_AUTO_LINE_SPACE_MASK);
	textProperties->fillBaseLineShift(tStyle.getBaseLineShift(), tStyle.getValueMask() & TEXT_STYLE_BASE_LINE_SHIFT_MASK);
	textProperties->fillTracking(tStyle.getTracking(), tStyle.getValueMask() & TEXT_STYLE_TRACKING_MASK);
	textProperties->fillHorizontalScale(tStyle.getHorizontalScale(), tStyle.getValueMask() & TEXT_STYLE_HORIZONTAL_SCALE_MASK);
	//g_printerr("SelectionList::fillTextProperties: values set valueMask: %lX, font %s, size %f, lspace %f\n", tStyle.getValueMask(), tStyle.getFont() ? tStyle.getFont()->getName() : "NULL", tStyle.getPointSize(), tStyle.getLineSpace());
}

SelectionList::~SelectionList() {
	delete objects;
}
