/* $Id: StandardTextObject.cc,v 1.56 2004/10/15 07:26:09 pklaus Exp $
 *
 * StandardTextObject.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "StandardTextObject.hh"
#include "TransformationMatrix.hh"
#include "DocWindow.hh"
#include "GDDocument.hh"
#include "BoundingBox.hh"
#include "GDCharacter.hh"
#include "TextRow.hh"
#include "FontPool.hh"
#include "GDFont.hh"
#include "BezierNode.hh"
#include "Layer.hh"
#include "GDApplication.hh"
#include "ObjectGroup.hh"
#include "GObject.hh"
#include <libart_lgpl/art_rgb_rgba_affine.h>
#include <libart_lgpl/art_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_rgb_svp.h>
#include <libart_lgpl/art_svp_vpath_stroke.h>
#include <libart_lgpl/art_svp_render_aa.h>
#include <math.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdk.h>
#include FT_GLYPH_H


extern FT_Library freetype_library;



void invertRect(OutputDescription& odes, position tl, position tr, position bl, position br) {
	GdkPoint gdkpts[4];
	
	odes.docwin->doc->context.set_function(GDK_INVERT);
	gdkpts[0].x = (gint) tl.x;
	gdkpts[0].y = (gint) tl.y;
	gdkpts[1].x = (gint) tr.x;
	gdkpts[1].y = (gint) tr.y;
	gdkpts[2].x = (gint) br.x;
	gdkpts[2].y = (gint) br.y;
	gdkpts[3].x = (gint) bl.x;
	gdkpts[3].y = (gint) bl.y;
	odes.docwin->DocWinDrawingArea->get_window().draw_polygon(
			odes.docwin->doc->context, 1, gdkpts, 4);
}

typedef struct invertPrivate {
	guchar *buffer;
	size_t	rowstride;
	guint width;
} invertPrivate;

void invertCallback(void *data, int y, int start, ArtSVPRenderAAStep *steps, gint n_steps) {
	invertPrivate *priv_data = (invertPrivate *) data;
	guchar *base = priv_data->buffer + y * priv_data->rowstride;
	guint x = 0;
	int val = start;
	for (x = 0; x < priv_data->width; x++) {
		if (n_steps && x >= (guint) steps->x) {
			val -= steps->delta;
			steps++;
			n_steps--;
		}
		if (val >> 16 > 128) {
			base[0] = 255 - base[0];
			base[1] = 255 - base[1];
			base[2] = 255 - base[2];
		}
		base += 3;
	}
}

void bufferInvertRect(OutputDescription& odes, position tl, position tr, position bl, position br) {
	
	ArtVpath apath[6];
	ArtSVP *asvp;
	invertPrivate priv_data;
	
	priv_data.buffer = odes.osBuffer->pixbuffer;
	priv_data.rowstride = odes.osBuffer->pxbRowstride;
	priv_data.width = odes.osBuffer->pxbWidth;
	
	apath[0].code = ART_MOVETO;
	apath[0].x = tl.x;
	apath[0].y = tl.y;
	apath[1].code = ART_LINETO;
	apath[1].x = tr.x;
	apath[1].y = tr.y;
	apath[2].code = ART_LINETO;
	apath[2].x = br.x;
	apath[2].y = br.y;
	apath[3].code = ART_LINETO;
	apath[3].x = bl.x;
	apath[3].y = bl.y;
	apath[4].code = ART_LINETO;
	apath[4].x = tl.x;
	apath[4].y = tl.y;
	apath[5].code = ART_END;
	apath[5].x = 0;
	apath[5].y = 0;
	asvp = art_svp_from_vpath(apath);
	art_svp_render_aa(asvp, 0, 0,
				odes.osBuffer->pxbWidth,
				odes.osBuffer->pxbHeight,
				&invertCallback,
				&priv_data);
	art_svp_free(asvp);	
}


StandardTextObject::StandardTextObject(GDDocument *document) : TextObject(document) {
	//g_printerr("StandardTextObject::StandardTextObject initialising\n");
	fillStyle = NULL;
	strokeStyle = NULL;
	TextRow *textrow = new TextRow();
	textrow->defaultStyle = document->currentTextStyle;
	text.insert(text.end(), textrow);
	selStartRow = selEndRow = selStartColumn = selEndColumn = 0;
	//selectAll(); // commented out for code cleanup
}

StandardTextObject::StandardTextObject(GDDocument *document, gint dummy) : TextObject(document) {
	//g_printerr("StandardTextObject::StandardTextObject initialising\n");
	fillStyle = NULL;
	strokeStyle = NULL;
	selStartRow = selEndRow = selStartColumn = selEndColumn = 0;
	//selectAll(); // commented out for code cleanup
}

void StandardTextObject::bufferDrawPoint(OffscreenBuffer& osBuffer, position point, guint color) {
	//g_printerr("StandardTextObject::bufferDrawPoint\n");
	ArtVpath apath[6];
	ArtSVP *asvp;
	apath[0].code = ART_MOVETO;
	apath[0].x = ROUND(point.x) - BEZIER_HALFSIZE;
	apath[0].y = ROUND(point.y) - BEZIER_HALFSIZE;
	apath[1].code = ART_LINETO;
	apath[1].x = apath[0].x + BEZIER_SIZE;
	apath[1].y = apath[0].y;
	apath[2].code = ART_LINETO;
	apath[2].x = apath[1].x;
	apath[2].y = apath[1].y + BEZIER_SIZE;
	apath[3].code = ART_LINETO;
	apath[3].x = apath[0].x;
	apath[3].y = apath[2].y;
	apath[4].code = ART_LINETO;
	apath[4].x = apath[0].x;
	apath[4].y = apath[0].y;
	apath[5].code = ART_END;
	apath[5].x = 0;
	apath[5].y = 0;
	asvp = art_svp_from_vpath(apath);
	art_rgb_svp_alpha(asvp, 0, 0,
				osBuffer.pxbWidth,
				osBuffer.pxbHeight,
				color,
				osBuffer.pixbuffer,
				osBuffer.pxbRowstride,
				NULL);
	art_svp_free(asvp);	
}


// GObject methods
void StandardTextObject::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//doDraw(docwin, &context, bufferDraw); // old version without draw/fill style
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		ObjectGroup *textOutline = convert2Outline();
		textOutline->draw(osBuffer, drawArea);
		delete textOutline;
	}
}

void StandardTextObject::drawBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		ObjectGroup *textOutline = convert2Outline();
		textOutline->drawBlack(osBuffer, drawArea);
		delete textOutline;
	}
	//doDraw(docwin, NULL, bufferDraw);
}

void StandardTextObject::fillBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		ObjectGroup *textOutline = convert2Outline();
		textOutline->fillBlack(osBuffer, drawArea);
		delete textOutline;
	}
	//doDraw(docwin, NULL, bufferDraw);
}

void StandardTextObject::doDraw(OffscreenBuffer& osBuffer, DocWindow *docwin, Gdk_GC *context, drawing_routine what, RectangleArea& drawArea) {
	//g_printerr("StandardTextObject::doDraw %lX\n", this);
	TransformationMatrix tmp = tMatrix;
	TransformationMatrix t;
	t.setTranslation(- osBuffer.origin.x, - osBuffer.origin.y);
	tmp *= t;
	t.reset();
	t.setScale(osBuffer.scaleFactor, osBuffer.scaleFactor);
	tmp *= t;
	//tmp.printMatrix(); // DEBUGGING
	FT_Matrix matrix;
	matrix.xx = (FT_Fixed) (tmp.m[0][0] * 0x10000);
	matrix.yy = (FT_Fixed) (tmp.m[1][1] * 0x10000);
	matrix.xy = (FT_Fixed) (-tmp.m[0][1] * 0x10000);
	matrix.yx = (FT_Fixed) (-tmp.m[1][0] * 0x10000);
	FT_Vector start;
	gdouble rowY = 0;
	for (guint r = 0; r < text.size(); r++) {
		//g_printerr("StandardTextObject::doDraw drawing line %d at Y: %f\n", r, rowY);
		TextRow *textrow = text[r];
		GDFont *oldfont = NULL;
		double oldsize = -1;
		double oldhs = -1;
		FT_Face face = NULL;
		gdouble x = 0;
		int dispHeight = osBuffer.pxbHeight;
		if (textrow->size() > 0) {
			TextStyle *primeStyle = textrow->getChar(0)->getTextStyle();
			if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				x = -(textrow->getWidth());
			} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
				x = -(textrow->getWidth() / 2.0);
			}
			int gindex = -1;
			for (guint i = 0; i < textrow->size(); i++) {
				FT_Glyph glyph;
				GDCharacter *ch = textrow->getChar(i);
				//g_printerr("StandardTextObject::doDraw character %d: %c (%d/%d)\n", i, ch->unicode, r, i);
				TextStyle chStyle = *(ch->getTextStyle());
				if (chStyle.getFont() != oldfont || chStyle.getPointSize() != oldsize || chStyle.getHorizontalScale() != oldhs) {
					oldfont = chStyle.getFont();
					oldsize = chStyle.getPointSize();
					oldhs = chStyle.getHorizontalScale();
					face = oldfont->getFace();
					int pixWidth = 0;
					if (oldhs != 1) {
						pixWidth = (int) (oldsize * oldhs);
					}
					//g_printerr("Font: name: %s psName: %s\n", oldfont->getName(), oldfont->getPSName());
					//g_printerr("Changed Font: face: %lX, size: %f\n", face, oldsize);
					FT_Set_Pixel_Sizes(face, pixWidth, (int) (oldsize));
				}
				if (chStyle.getTracking()) {
					x += (chStyle.getTracking() * oldsize / 1000.0 * osBuffer.scaleFactor);
				}
				if (gindex == -1) {
					gindex = FT_Get_Char_Index(face, ch->getUnicode());
				}
				int error = FT_Load_Glyph(face, gindex, (what != bufferDraw) ? FT_LOAD_NO_HINTING : FT_LOAD_DEFAULT);
				if (error) {
					g_printerr("StandardTextObject::doDraw ERROR cannot load glyph %d\n", gindex);
				}
				position st;
				st.x = x;
				st.y = rowY - chStyle.getBaseLineShift();
				//g_printerr("StandardTextObject::doDraw x=%f y=%f\n", st.x, st.y);
				//g_printerr("StandardTextObject::doDraw 1 st.x: %f st.y: %f, dh: %d\n", st.x, st.y, dispHeight);
				//g_printerr("StandardTextObject::doDraw rowY: %f, baseLS: %f\n", rowY, ch->style.getBaseLineShift());
				tmp.transform(st);
				//g_printerr("StandardTextObject::doDraw x=%f y=%f\n", st.x, st.y);
				start.x = (FT_Fixed) (st.x * 64);
				start.y = (FT_Fixed) ((dispHeight - st.y) * 64);
				//g_printerr("StandardTextObject::doDraw 2 st.x: %f st.y: %f, dh: %d\n", st.x, st.y, dispHeight);
				//g_printerr("StandardTextObject::doDraw rowY: %d, baseLS: %f\n", rowY, ch->style.getBaseLineShift());
				error = FT_Get_Glyph(face->glyph, &glyph);
				if (error) {
					g_printerr("StandardTextObject::doDraw cannot get glyph %d\n", error);
				}
				error = FT_Glyph_Transform(glyph, &matrix, &start);
				if (error) {
					g_printerr("StandardTextObject::doDraw cannot transform glyph %d\n", error);
				}
				// calc bounding rect
				BoundingBox bbox;
				position cor[4];
				cor[0].x = x;
				cor[0].y = rowY + textrow->getDescender();
				cor[1].x = x;
				cor[1].y = rowY - textrow->getAscender();
				cor[2].x = x + ((gdouble) face->glyph->advance.x) / 64.0;
				cor[2].y = cor[1].y;
				cor[3].x = cor[2].x;
				cor[3].y = cor[2].y;
				for (int t = 0; t < 4; t++) {
					tMatrix.transform(cor[t]);
					bbox.include(cor[t]);
				}
				//g_printerr("bbox: %f/%f,%f/%f drawArea: %f/%f,%f/%f res: %s\n", bbox.left, bbox.top, bbox.right, bbox.bottom, drawArea.left, drawArea.top, drawArea.right, drawArea.bottom, (drawArea.intersect(bbox) ? "yes" : "no"));
				if (drawArea.intersect(bbox)) {
					if (what == bufferDraw) {
						error = bufferCopyCharBitmap(osBuffer.pixbuffer, osBuffer.pxbSize,
									osBuffer.pxbRowstride, 
									&glyph, 0, dispHeight);
					} else if (what == drawDrawable) {
						error = copyCharBitmap(docwin->DocWinDrawingArea->get_window(), 
								*context, &glyph, 0, dispHeight);
					} else if (what == xorDraw) {
						error = selCopyCharBitmap(docwin->DocWinDrawingArea->get_window(), *context, &glyph, 0, dispHeight);
					}
					if (error > 1) {
						g_printerr("StandardTextObject::doDraw cannot render glyph %d\n", gindex);
					}
				}
				x += ((gdouble) face->glyph->advance.x) / 64.0; // calculate the x advance before transforming because transforming alters the advance vector
				//g_printerr("after advance: x=%f (advance = %d)\n",x, face->glyph->advance.x);
				if (i < textrow->size() - 1) {
					int oldindex = gindex;
					gindex = FT_Get_Char_Index(face, textrow->getChar(i+1)->getUnicode());
					FT_Vector kerning;
					error = FT_Get_Kerning(face, oldindex, gindex, ft_kerning_default, &kerning);
					//g_printerr("Kerning: %d %d\n", kerning.x, kerning.y);
					x += ((gdouble) kerning.x) / 64.0;
					//g_printerr("after kerning: x=%f\n", x);
				}
				FT_Done_Glyph(glyph);
			}
		}
		rowY += textrow->getLineSpace();
	}
}

void StandardTextObject::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		GdkGCValues values;
		gdk_gc_get_values(context.gdkobj(), &values);
		context.set_function(GDK_AND);
		//context.set_function(GDK_EQUIV);
		doDraw(docwin->osBuffer, docwin, &context, drawDrawable, drawArea);
		context.set_function(values.function);
	}
}


void StandardTextObject::drawPoints(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
}

BoundingBox StandardTextObject::getBoundingBox() {
	BoundingBox bbox;
	position pointTL, pointTR, pointBL, pointBR;
	gdouble baseline = 0;
	for (guint i = 0; i < text.size(); i++) {
		pointTL.y = pointTR.y = baseline - text[i]->getAscender();
		pointTL.x = pointBL.x = 0;
		pointBL.y = pointBR.y = baseline + text[i]->getHeight() - text[i]->getAscender();
		pointTR.x = pointBR.x = text[i]->getWidth();
		tMatrix.transform(pointTL);
		tMatrix.transform(pointTR);
		tMatrix.transform(pointBL);
		tMatrix.transform(pointBR);
		bbox.include(pointTL);
		bbox.include(pointTR);
		bbox.include(pointBL);
		bbox.include(pointBR);
	}
		
	return bbox;
}

gdouble StandardTextObject::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	position origin;
	origin.x = 0;
	origin.y = 0;
	tMatrix.transform(origin);
	gdouble dx = fabs(origin.x - x);
	gdouble dy = fabs(origin.y - y);
	if (dx < tolerance && dy < tolerance) {
		ax = origin.x;
		ay = origin.y;
		return dx * dx + dy * dy;
	}
	return G_MAXDOUBLE;
}

SelectableItem *StandardTextObject::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	//g_printerr("StandardTextObject::checkPoints x: %f y: %f tol: %f\n", x, y, tolerance);
	
	position origin;
	origin.x = 0;
	origin.y = 0;
	tMatrix.transform(origin);
	if (fabs(origin.x - x) < tolerance && fabs(origin.y - y) < tolerance) {
		return this;
	}
	return NULL;
}

GObject *StandardTextObject::checkArea(gdouble x, gdouble y) {
	//g_printerr("StandardTextObject::checkArea x: %f  y							: %f\n", x, y);
	position check;
	check.x = x; check.y = y;
	invTMatrix.transform(check);
	//g_printerr("StandardTextObject::checkArea: check: %f - %f\n", check.x, check.y);
	gdouble baseline = 0;
	for (guint i = 0; i < text.size(); i++) {
		gdouble justCorrect = 0;
		if (text[i]->size() > 0) {
			TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
			if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				justCorrect = -(text[i]->getWidth());
			} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
				justCorrect = -(text[i]->getWidth() / 2.0);
			}
		}
		//g_printerr("StandardTextObject::checkArea: line %d: b: %f, w: %f, h: %f, a: %f\n", i, baseline, text[i]->getWidth(), text[i]->getHeight(), text[i]->getAscender());
		if (check.x > justCorrect && check.x < text[i]->getWidth() + justCorrect &&
				check.y > baseline - text[i]->getAscender() &&
				check.y < baseline + text[i]->getHeight() - text[i]->getAscender()) {
			return this;
		}
		baseline += text[i]->getLineSpace();
	}
	return NULL;
}

SelectableItem *StandardTextObject::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	position check;
	check.x = x; check.y = y;
	invTMatrix.transform(check);
	gdouble baseline = 0;
	for (guint i = 0; i < text.size(); i++) {
		gdouble justCorrect = 0;
		if (text[i]->size() > 0) {
			TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
			if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				justCorrect = -(text[i]->getWidth());
			} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
				justCorrect = -(text[i]->getWidth() / 2.0);
			}
		}
		if (fabs(check.y - baseline) < tolerance && check.x > justCorrect - tolerance && 
						check.x < text[i]->getWidth() + justCorrect + tolerance) {
			return this;
		}
		baseline += text[i]->getLineSpace();
	}
	return NULL;
}

void StandardTextObject::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	position origin = { 0, 0 };
	position end = { 0, 0 };
	end.x = text[0]->getWidth();
	tMatrix.transform(origin);
	tMatrix.transform(end);
	if (checkRectangleLineIntersection(origin, end, left, top, right, bottom)) {
		if ((what & 3) == 2) {
			selectObject((what & 128) != 0);
		} else {
			select((what & 128) != 0);
		}
	}
}

void StandardTextObject::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	position origin = { 0, 0 };
	position end = { 0, 0 };
	end.x = text[0]->getWidth();
	tMatrix.transform(origin);
	tMatrix.transform(end);
	if (checkRectangleLineIntersection(origin, end, left, top, right, bottom)) {
		if ((what & 3) == 2) {
			deSelectObject();
		} else {
			deSelect();
		}
	}
}

gint StandardTextObject::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
	}
	return 0;
}

void StandardTextObject::deSelectAll() {
	editing = false;
	TextObject::deSelectAll();
}

gint StandardTextObject::writeXML(ostream& xmlout) {
	TransformationMatrix tmp = tMatrix;
	TransformationMatrix t;
	t.setTranslation(-doc->nullPoint.x, -doc->nullPoint.y);
	tmp *= t;
	xmlout << "<text ";
	xmlout << " transform=\"";
	tmp.writeXML(xmlout);
	xmlout << "\">\n";
	gdouble y = 0;
	for (guint r = 0; r < text.size(); r++) {
		TextRow *textrow = text[r];
		textrow->writeXML(xmlout, 0, y, doc);
		y += textrow->getLineSpace();
	}
	
	xmlout << "</text>\n";
	return 0;
}

gint StandardTextObject::writePS(ostream& psout) {
	// calculate the matrix
	TransformationMatrix tmp = tMatrix;
	TransformationMatrix t;
	// translate to doc origin
	t.setTranslation(-doc->nullPoint.x, -doc->nullPoint.y);
	tmp *= t;
	// change the coordinate system to postscript
	// flip on the x-axis
	t.reset();
	t.m[1][1] = -1;
	tmp *= t;
	// move by page height
	t.reset();
	t.setTranslation(0, doc->pageHeight);
	tmp *= t;
	tmp.m[0][1] *= -1;
	tmp.m[1][1] *= -1;
	psout << "gs\n";
	tmp.writePS(psout);
	psout << " concat\n";
	gdouble y = 0;
	TextStyle *lastStyle = NULL;
	for (guint r = 0; r < text.size(); r++) {
		TextRow *textrow = text[r];
		lastStyle = textrow->writePS(psout, 0, y, lastStyle, doc);
		y -= textrow->getLineSpace();
	}
	
	psout << "gr\n";
	return 0;
}


// DrawObject methods
void StandardTextObject::drawWithSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("StandardTextObject::drawWithSelection\n");
	if (editing) {
		if (selStartRow != selEndRow || selStartColumn != selEndColumn) {
			bufferDrawTextSelection(osBuffer, drawArea);
		}
	} else {
		Layer *myLayer = doc->findObjectLayer(this);
		guint color = 0xFFFFFFFF;
		if (myLayer) {
			color = myLayer->getSelectionColor();
		} else {
			g_printerr("PixelImage::drawWithSelection: no layer found for text %lX\n", (long) this);
		}
		TransformationMatrix tmp = tMatrix;
		TransformationMatrix t;
		t.setTranslation(- osBuffer.origin.x, - osBuffer.origin.y);
		tmp *= t;
		t.reset();
		t.setScale(osBuffer.scaleFactor, osBuffer.scaleFactor);
		tmp *= t;
		position origin = { 0, 0 };
		tmp.transform(origin);
		if (text.size() > 0) {
			//g_printerr("StandardTextObject::drawWithSelection drawn line of 0\n");
			gdouble y = 0;
			for (guint i = 0; i < text.size(); i++) {
				gdouble justCorrect = 0;
				if (text[i]->size() > 0) {
					TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
					if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
						justCorrect = -(text[i]->getWidth());
					} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
						justCorrect = -(text[i]->getWidth() / 2.0);
					}
				}
				position s = { justCorrect, y};
				position e = { text[i]->getWidth() + justCorrect, y };
				tmp.transform(s);
				tmp.transform(e);
				RectangleArea lrect(	s.x / osBuffer.scaleFactor + osBuffer.origin.x, 
										s.y / osBuffer.scaleFactor + osBuffer.origin.y, 
										e.x / osBuffer.scaleFactor + osBuffer.origin.x, 
										e.y / osBuffer.scaleFactor + osBuffer.origin.y);
				if (drawArea.intersect(lrect)) {
					ArtVpath apath[3];
					apath[0].code = ART_MOVETO;
					apath[0].x = s.x;
					apath[0].y = s.y;
					apath[1].code = ART_LINETO;
					apath[1].x = e.x;
					apath[1].y = e.y;
					apath[2].code = ART_END;
					apath[2].x = 0;
					apath[2].y = 0;
					ArtSVP *asvp = art_svp_vpath_stroke(apath, 
												ART_PATH_STROKE_JOIN_MITER,
												ART_PATH_STROKE_CAP_BUTT,
												1, 4, 0.5);
					art_rgb_svp_alpha(asvp, 0, 0,
									osBuffer.pxbWidth,
									osBuffer.pxbHeight,
									color,
									osBuffer.pixbuffer,
									osBuffer.pxbRowstride,
									NULL);
					art_svp_free(asvp);
				}
				//g_printerr("StandardTextObject::drawWithSelection drawn line of %d (%f,%f/%f,%f)\n", i, s.x, s.y, e.x, e.y);
				y += text[i]->getLineSpace();
			}
		}
		bufferDrawPoint(osBuffer, origin, color);
	}
}

void StandardTextObject::drawWireFrameSelection(DocWindow *docwin, RectangleArea& drawArea) {
	if (editing) {
		if (selStartRow != selEndRow || selStartColumn != selEndColumn) {
			drawTextSelection(docwin, drawArea);
		}
	} else {
		Layer *myLayer = docwin->doc->findObjectLayer(this);
		guint color = 0xFFFFFFFF;
		if (myLayer) {
			color = myLayer->getSelectionColor();
		} else {
			g_printerr("PixelImage::drawWireFrameSelection: no layer found for text %lX\n", (long) this);
		}
		//g_printerr("PixelImage::drawWireFrameSelection got color\n");
		Gdk_Colormap cmap = Gdk_Colormap::get_system();
		Gdk_Color gcolor;
		gcolor.set_rgb(((color >> 24) & 0xFF) << 8, 
						((color >> 16) & 0xFF) << 8,
						((color >> 8) & 0xFF) << 8);
		cmap.alloc(gcolor);
		Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
		Gdk_GC context = Gdk_GC(drawable);
		context.set_foreground(gcolor);
		context.set_line_width(1);
		context.set_function(GDK_COPY);
		TransformationMatrix tmp = tMatrix;
		TransformationMatrix t;
		t.setTranslation(- docwin->osBuffer.origin.x, - docwin->osBuffer.origin.y);
		tmp *= t;
		t.reset();
		t.setScale(docwin->osBuffer.scaleFactor, docwin->osBuffer.scaleFactor);
		tmp *= t;
		position origin = { 0, 0 };
		tmp.transform(origin);
		drawPoint(docwin, context, origin);
		if (text.size() > 0) {
			gdouble y = 0;
			for (guint i = 0; i < text.size(); i++) {
				gdouble justCorrect = 0;
				if (text[i]->size() > 0) {
					TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
					if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
						justCorrect = -(text[i]->getWidth());
					} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
						justCorrect = -(text[i]->getWidth() / 2.0);
					}
				}
				position s = { justCorrect, y};
				position e = { text[i]->getWidth() + justCorrect, y };
				tmp.transform(s);
				tmp.transform(e);
				RectangleArea lrect(	s.x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x, 
										s.y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y, 
										e.x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x, 
										e.y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y);
				if (drawArea.intersect(lrect)) {
					docwin->DocWinDrawingArea->get_window().draw_line(context,
								(int) s.x, (int) s.y, (int) e.x, (int) e.y);
				}
				y += text[i]->getLineSpace();
			}
		}
	}
}

void StandardTextObject::gatherResource(GDResourceList& collection, gint type) {
	vector<TextRow *>::iterator it = text.begin();
	for (; it != text.end(); it++) {
		TextRow *tr = *it;
		for (guint i = 0; i < tr->size(); i++) {
			if (type & GDRESOURCE_TYPE_FONT) {
				collection.insert((GDResource *)(tr->getChar(i)->getTextStyle()->getFont()));
			}
			if (type & (GDRESOURCE_TYPE_GRADIENT | GDRESOURCE_TYPE_COLOR | GDRESOURCE_TYPE_PATTERN)) {
				if (fillStyle) {
					fillStyle->gatherResource(collection, type);
				}
				if (strokeStyle) {
					strokeStyle->gatherResource(collection, type);
				}
			}
		}
	}
}


// SelectableItem methods
void StandardTextObject::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	//g_printerr("StandardTextObject::drawSelection %lX\n", (long) this);
	GdkGCValues values;
	gdk_gc_get_values(context.gdkobj(), &values);
	context.set_function(GDK_XOR);
	doDraw(docwin->osBuffer, docwin, &context, xorDraw, drawArea);
	context.set_function(values.function);
}

void StandardTextObject::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	//g_printerr("StandardTextObject::transformByMatrix %lX\n", (long) this);
	//tMatrix.printMatrix();
	//g_printerr("Matrix m: \n");
	//m.printMatrix();
	tMatrix *= m;
	//g_printerr("New Object matrix:\n");
	//tMatrix.printMatrix();
	invTMatrix = tMatrix.inverse();
	
	// transform character styles
	vector<TextRow *>::iterator it = text.begin();
	for (; it != text.end(); it++) {
		(*it)->transformByMatrix(m, incFillPattern);
	}	
}

void StandardTextObject::setFillMatrix(TransformationMatrix& m) {
	// set character styles
	vector<TextRow *>::iterator it = text.begin();
	for (; it != text.end(); it++) {
		(*it)->setFillMatrix(m);
	}	
}

SelectableItem *StandardTextObject::clone() {
	//g_printerr("StandardTextObject::clone %lX\n", this);
	StandardTextObject *sto = new StandardTextObject(doc);
	sto->tMatrix = tMatrix;
	sto->removeRow(0); // the constructor inserts the first row
	for (guint i = 0; i < text.size(); i++) {
		//g_printerr("StandardTextObject::clone cloning line %d\n", i);
		TextRow *ntr = new TextRow(*(text[i]));
		sto->text.insert(sto->text.end(), ntr);
	}
	return sto;
}

void StandardTextObject::setCoords(SelectableItem* from) {
	//g_printerr("StandardTextObject::setCoords %lX\n", this);
	StandardTextObject *stoFrom = (StandardTextObject *) from;
	tMatrix = stoFrom->tMatrix;
}

gboolean StandardTextObject::selected() {
	//g_printerr("StandardTextObject::selected %lX\n", this);
	return selection;
}

int StandardTextObject::getCoordinates(gdouble& x, gdouble& y) {
	//g_printerr("StandardTextObject::getCoordinates %lX\n", this);
	position origin = { 0, 0 };
	tMatrix.transform(origin);
	x = origin.x; y = origin.y;
	return 1;
}

void StandardTextObject::clear() {
	//g_printerr("StandardTextObject::clear %lX\n", this);
	doc->removeObject(this);
	deSelectAll(); // remove from document's selection
	delete this;
}

gchar *StandardTextObject::identify() {
	return "TextStandard";
}

gint StandardTextObject::insertKeyVal(int keyval, int state) {
	//g_printerr("StandardTextObject::insertKeyVal: val: %d - %c\n", keyval, keyval);
	if (keyval == GDK_Left) {
		if (selStartColumn == 0) {
			if (selStartRow > 0) {
				selStartRow--;
				selEndRow = selStartRow;
				selStartColumn = selEndColumn = text[selStartRow]->size();
			}
		} else {
			selStartColumn--;
			selEndColumn = selStartColumn;
		}
	} else if (keyval == GDK_Right) {
		if ((guint) selEndColumn >= text[selStartRow]->size()) {
			if ((guint) selEndRow < text.size() - 1) {
				selEndRow++;
				selStartRow = selEndRow;
				selStartColumn = selEndColumn = 0;
			}
		} else {
			selEndColumn++;
			selStartColumn = selEndColumn;
		}
	} else if (keyval == GDK_Up) {
		if (selStartRow > 0) {
			selStartRow--;
			selEndRow = selStartRow;
			if ((guint) selStartColumn > text[selStartRow]->size()) {
				selEndColumn = selStartColumn = text[selStartRow]->size();
			}
		}
	} else if (keyval == GDK_Down) {
		if ((guint) selEndRow < text.size() - 1) {
			selEndRow++;
			selStartRow = selEndRow;
			if ((guint) selEndColumn > text[selStartRow]->size()) {
				selEndColumn = selStartColumn = text[selStartRow]->size();
			}
		}
	}
	return 0;
}

gint StandardTextObject::insertCharacter(int character) {
	//g_printerr("StandardTextObject::insertCharacter: char: %d - %c\n", character, character);
	removeSelectedText();
	if (character == 8) { // backspace
		if (selStartColumn == 0) {
			if (selStartRow > 0) {
				//delete current text row
				selStartColumn = selEndColumn = text[selStartRow - 1]->size();
				for (; text[selStartRow]->size();) {
					text[selStartRow - 1]->insertChar(-1, text[selStartRow]->removeChar(0));
				}
				removeRow(selStartRow);
				selStartRow--;
				selEndRow--;
			}
		} else {
			selStartColumn--;
			selEndColumn--;
			delete text[selStartRow]->removeChar(selStartColumn);
		}
	} else if (character == 127) { // delete
		if ((guint) selStartColumn >= text[selStartRow]->size()) {
			if ((guint) selStartRow < text.size() - 1) {
				for (; text[selStartRow+1]->size();) {
					text[selStartRow]->insertChar(-1, text[selStartRow+1]->removeChar(0));
				}
				removeRow(selStartRow + 1);
			}
		} else {
			delete text[selStartRow]->removeChar(selStartColumn+1);
		}
	} else if (character == 13) { // carriage return
		TextRow *newRow = new TextRow();
		newRow->defaultStyle = doc->currentTextStyle;
		insertRow(selStartRow, newRow);
		for (; (guint) selStartColumn < text[selStartRow]->size();) {
			newRow->insertChar(-1, text[selStartRow]->removeChar(selStartColumn));
		}
		selStartRow++;
		selEndRow++;
		selStartColumn = 0;
		selEndColumn = 0;
	} else {
		GDCharacter *newChar = new GDCharacter(character);
		newChar->setTextStyle(doc->currentTextStyle);
		gint insPos = -1;
		if ((guint) selStartColumn < text[selStartRow]->size()) {
			insPos = selStartColumn;
		}
		text[selStartRow]->insertChar(insPos, newChar);
		selStartColumn++;
		selEndColumn++;
	}
	return 0;
}

gint StandardTextObject::startTextSelection(DocWindow *docwin, Gdk_GC& context, gdouble x, gdouble y) {
	//g_printerr("StandardTextObject::startTextSelection: x: %f, y: %f\n", x, y);
	//g_printerr("StandardTextObject::startTextSelection: sr: %d er: %d sc: %d ec: %d\n", selStartRow, selEndRow, selStartColumn, selEndColumn);
	if (selEndRow != selStartRow || selEndColumn != selStartColumn){
		// re-invert the old selection to remove it
		RectangleArea rect = docwin->getContentRectangle();
		drawTextSelection(docwin, rect);
		// inform the selection that the text selection has changed
		doc->selection->textChanged(doc);
	}
	calcTextPosition(x, y, selOriginColumn, selOriginRow);
	//g_printerr("StandardTextObject::startTextSelection: textpos: %d/%d\n", selStartColumn, selStartRow);
	selEndColumn = selStartColumn = selTargetColumn = selOriginColumn;
	selEndRow = selStartRow = selTargetRow = selOriginRow;
	return 0;
}

gint StandardTextObject::setTextSelection(DocWindow *docwin, gdouble x, gdouble y) {
	//g_printerr("StandardTextObject::setTextSelection: x: %f, y: %f\n", x, y);
	calcTextPosition(x, y, selTargetColumn, selTargetRow);
	//g_printerr("StandardTextObject::setTextSelection: textpos: %d/%d\n", selEndColumn, selEndRow);
	OutputDescription odes;
	odes.docwin = docwin;
	odes.osBuffer = &(docwin->osBuffer);
	if (selTargetRow < selOriginRow || 
			(selTargetRow == selOriginRow && selTargetColumn < selOriginColumn)) {
		if (selStartRow != selTargetRow || selStartColumn != selTargetColumn) {
			if (selStartRow < selTargetRow || 
					(selStartRow == selTargetRow && selStartColumn < selTargetColumn)) {
				drawTextSelRange(odes, &invertRect, selStartRow, selStartColumn, selTargetRow, selTargetColumn);
			} else {
				drawTextSelRange(odes, &invertRect, selTargetRow, selTargetColumn, selStartRow, selStartColumn);
			}
			// inform the selection that the text selection has changed
			doc->selection->textChanged(doc);
			doc->selection->changed(doc);
		}
		selStartRow = selTargetRow;
		selStartColumn = selTargetColumn;
		selEndRow = selOriginRow;
		selEndColumn = selOriginColumn;
	} else {
		if (selEndRow != selTargetRow || selEndColumn != selTargetColumn) {
			if (selEndRow < selTargetRow ||
					(selEndRow == selTargetRow && selEndColumn < selTargetColumn)) {
				drawTextSelRange(odes, &invertRect, selEndRow, selEndColumn, selTargetRow, selTargetColumn);
			} else {
				drawTextSelRange(odes, &invertRect, selTargetRow, selTargetColumn, selEndRow, selEndColumn);
			}
			// inform the selection that the text selection has changed
			doc->selection->textChanged(doc);
			doc->selection->changed(doc);
		}
		selStartRow = selOriginRow;
		selStartColumn = selOriginColumn;
		selEndRow = selTargetRow;
		selEndColumn = selTargetColumn;
	}
	if (selStartColumn == 0) {
		//doc->currentTextStyle = text[selStartRow]->defaultStyle;
		if (text[selStartRow]->size() > 0) {
			doc->currentTextStyle = *(text[selStartRow]->getChar(selStartColumn)->getTextStyle());
		} else {
			doc->currentTextStyle = text[selStartRow]->defaultStyle;
		}
	} else {
		doc->currentTextStyle = *(text[selStartRow]->getChar(selStartColumn-1)->getTextStyle());
	}
	return 0;
}

void StandardTextObject::calcTextPosition(gdouble px, gdouble py, gint &tcol, gint &trow) {
	position point;
	point.x = px; point.y = py;
	invTMatrix.transform(point);
	gdouble y = 0;
	guint i;
	for (i = 0; i < text.size(); i++) {
		gdouble lh = text[i]->getLineSpace();
		if (lh < text[i]->getHeight()) {
			lh = text[i]->getHeight();
		}
		gdouble asc = text[i]->getAscender();
		if (y - asc < point.y && y - asc + lh > point.y) {
			trow = i;
			break;
		}
		y += text[i]->getLineSpace();
	}
	if (i >= text.size()) {
		return;
	}
	gdouble justCorrect = 0;
	if (text[i]->size() > 0) {
		TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
		if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
			justCorrect = -(text[i]->getWidth());
		} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
			justCorrect = -(text[i]->getWidth() / 2.0);
		}
	}
	TextRow *txtr = text[i];
	guint k = 0;
	while (k < txtr->size() && txtr->getChar(k)->endX + justCorrect < point.x) {
		k++;
	}
	if (k <= txtr->size()) {
		tcol = k;
	}
}

void StandardTextObject::drawTextCursorMark(DocWindow *docwin, RectangleArea& drawArea) {
	//g_printerr("StandardTextObject::drawTextCursorMark sel: %d/%d - %d/%d\n", selStartRow, selStartColumn, selEndRow, selEndColumn);
	if (selStartColumn == selEndColumn && selStartRow == selEndRow) {
		// null selection, show blinking insertion cursor mark
		Gdk_Window dwaw = docwin->DocWinDrawingArea->get_window();
		Gdk_GC blackgc(dwaw);
		Gdk_GC bgc = docwin->DocWinDrawingArea->get_style()->get_black_gc();
		blackgc.copy(bgc);
		blackgc.set_function(GDK_INVERT);
		gdouble x,y;
		y = 0;
		for (guint i = 0; i < text.size() && i < (guint) selStartRow; i++) {
			y += text[i]->getLineSpace();
		}
		TransformationMatrix tmp = tMatrix;
		TransformationMatrix t;
		t.setTranslation(- docwin->osBuffer.origin.x, - docwin->osBuffer.origin.y);
		tmp *= t;
		t.reset();
		t.setScale(docwin->osBuffer.scaleFactor, docwin->osBuffer.scaleFactor);
		tmp *= t;
		gdouble justCorrect = 0;
		if (text[selStartRow]->size() > 0) {
			TextStyle *primeStyle = text[selStartRow]->getChar(0)->getTextStyle();
			if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				justCorrect = -(text[selStartRow]->getWidth());
			} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
				justCorrect = -(text[selStartRow]->getWidth() / 2.0);
			}
		}
		x = selStartColumn ? text[selStartRow]->getChar(selStartColumn - 1)->endX : 0;
		x += justCorrect;
		gdouble height = text[selStartRow]->getLineSpace();
		y += text[selStartRow]->getHeight() - text[selStartRow]->getAscender();
		position ptmp;
		GdkPoint gdkpts[4];
		ptmp.x = x;
		ptmp.y = y;
		tmp.transform(ptmp);
		gdkpts[0].x = (gint) ptmp.x;
		gdkpts[0].y = (gint) ptmp.y;
		ptmp.x = x + 2;
		ptmp.y = y;
		tmp.transform(ptmp);
		gdkpts[1].x = (gint) ptmp.x;
		gdkpts[1].y = (gint) ptmp.y;
		ptmp.x = x + 2;
		ptmp.y = y - height;
		tmp.transform(ptmp);
		gdkpts[2].x = (gint) ptmp.x;
		gdkpts[2].y = (gint) ptmp.y;
		ptmp.x = x;
		ptmp.y = y - height;
		tmp.transform(ptmp);
		gdkpts[3].x = (gint) ptmp.x;
		gdkpts[3].y = (gint) ptmp.y;
		RectangleArea rect(	gdkpts[0].x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x,
							gdkpts[0].y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y,
							gdkpts[2].x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x,
							gdkpts[2].y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y);
		if (drawArea.intersect(rect)) {
		//g_printerr("StandardTextObject::drawTextCursorMark: h: %f  %d/%d %d/%d\n", height, gdkpts[0].x, gdkpts[0].y, gdkpts[2].x, gdkpts[2].y);
			docwin->DocWinDrawingArea->get_window().draw_polygon(blackgc, 1, gdkpts, 4);
		}

	}
}

void StandardTextObject::drawTextSelection(DocWindow *docwin, RectangleArea& drawArea) {
	OutputDescription odes;
	odes.docwin = docwin;
	odes.osBuffer = &(docwin->osBuffer);
	drawTextSelRange(odes, &invertRect, selStartRow, selStartColumn,
					selEndRow, selEndColumn);
}

void StandardTextObject::bufferDrawTextSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	OutputDescription odes;
	odes.osBuffer = &osBuffer;
	drawTextSelRange(odes, &bufferInvertRect, selStartRow, selStartColumn,
					selEndRow, selEndColumn);
}

void StandardTextObject::drawTextSelRange(OutputDescription& odes, 
				invertDrawingFunction drawInvert,
				guint sRow, guint sColumn, guint eRow, guint eColumn) {
	//g_printerr("StandardTextObject::drawTextSelRange: Start (%d/%d) End (%d/%d)\n", sRow, sColumn, eRow, eColumn);
	gdouble y = 0;
	if (text.size() > 0) {
		y = - text[0]->getAscender();
	}
	TransformationMatrix tmp = tMatrix;
	TransformationMatrix t;
	t.setTranslation( - odes.osBuffer->origin.x, - odes.osBuffer->origin.y);
	tmp *= t;
	t.reset();
	t.setScale(odes.osBuffer->scaleFactor, odes.osBuffer->scaleFactor);
	tmp *= t;
	guint i = 0;
	for (i = 0; i < text.size() && i <= eRow; i++) {
		if (i >= sRow && i <= eRow) {
			gdouble justCorrect = 0;
			if (text[i]->size() > 0) {
				TextStyle *primeStyle = text[i]->getChar(0)->getTextStyle();
				if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
					justCorrect = -(text[i]->getWidth());
				} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
					justCorrect = -(text[i]->getWidth() / 2.0);
				}
			}
			gdouble height = text[i]->getHeight();
			gdouble xs = 0;
			if (i == sRow) {
				guint sCol = (sColumn < text[i]->size()) ? sColumn : text[i]->size() - 1;
				xs = text[i]->getChar(sCol)->startX;
			}
			xs += justCorrect;
			gdouble xe = text[i]->getChar(text[i]->size() - 1)->endX;
			if (i == eRow && eColumn < text[i]->size()) {
				xe = text[i]->getChar(eColumn)->startX;
			}
			xe += justCorrect;
			position tl = { xs, y };
			position tr = { xe, y };
			position bl = { xs, y + height };
			position br = { xe, y + height };
			tmp.transform(tl);
			tmp.transform(tr);
			tmp.transform(bl);
			tmp.transform(br);
			(*drawInvert)(odes, tl, tr, bl, br);
		}
		y += text[i]->getLineSpace();
	}
	//g_printerr("StandardTextObject::drawTextSelRange: finished\n");
}

void StandardTextObject::insertRow(gint after, TextRow *row) {
	vector<TextRow *>::iterator it = text.begin();
	//g_printerr("StandardTextObject::insertRow: after: %d, size: %d\n", after, text.size());
	if ((guint) after >= text.size() - 1) {
		it = text.end();
	} else {
		it += after;
	}
	text.insert(it, row);
}

void StandardTextObject::removeRow(gint pos) {
	vector<TextRow *>::iterator it = text.begin();
	if (pos == -1) {
		it = text.end();
	} else {
		it += pos;
	}
	delete *it;
	text.erase(it);
}

void StandardTextObject::removeSelectedText() {
	if (selStartRow == selEndRow && selStartColumn == selEndColumn) return;
	if (selStartRow == selEndRow) {
		text[selStartRow]->removeChars(selStartColumn, selEndColumn - 1);
	} else {
		text[selStartRow]->removeChars(selStartColumn, -1);
		if (selEndColumn > 0) {
			text[selEndRow]->removeChars(-1, selEndColumn - 1);
		}
		for (; text[selEndRow]->size();) {
			text[selStartRow]->insertChar(-1, text[selEndRow]->removeChar(0));
		}
		while (selEndRow > selStartRow) {
			removeRow(selEndRow--);
		}
	}
	selEndColumn = selStartColumn;
}

GDStyle *StandardTextObject::getStrokeStyle() {
	if (strokeStyle) delete strokeStyle;
	strokeStyle = NULL;
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			for (int i = selStartColumn; i < selEndColumn; i++) {
				if (strokeStyle) {
					*strokeStyle &= *(tr->getChar(i)->getStrokeStyle());
				} else {
					strokeStyle = new GDStyle(*(tr->getChar(i)->getStrokeStyle()));
				}
			}
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				if (strokeStyle) {
					*strokeStyle &= *(tr->getChar(i)->getStrokeStyle());
				} else {
					strokeStyle = new GDStyle(*(tr->getChar(i)->getStrokeStyle()));
				}
			}
			for (int j = selStartRow + 1; j < selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					if (strokeStyle) {
						*strokeStyle &= *(tr->getChar(i)->getStrokeStyle());
					} else {
						strokeStyle = new GDStyle(*(tr->getChar(i)->getStrokeStyle()));
					}
				}
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				if (strokeStyle) {
					*strokeStyle &= *(tr->getChar(i)->getStrokeStyle());
				} else {
					strokeStyle = new GDStyle(*(tr->getChar(i)->getStrokeStyle()));
				}
			}
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				if (strokeStyle) {
					*strokeStyle &= *(tr->getChar(i)->getStrokeStyle());
				} else {
					strokeStyle = new GDStyle(*(tr->getChar(i)->getStrokeStyle()));
				}
			}
		}
	}
	return strokeStyle;
}

GDStyle *StandardTextObject::getFillStyle() {
	if (fillStyle) delete fillStyle;
	fillStyle = NULL;
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			for (int i = selStartColumn; i < selEndColumn; i++) {
				if (fillStyle) {
					*fillStyle &= *(tr->getChar(i)->getFillStyle());
				} else {
					fillStyle = new GDStyle(*(tr->getChar(i)->getFillStyle()));
				}
			}
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				if (fillStyle) {
					*fillStyle &= *(tr->getChar(i)->getFillStyle());
				} else {
					fillStyle = new GDStyle(*(tr->getChar(i)->getFillStyle()));
				}
			}
			for (int j = selStartRow + 1; j < selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					if (fillStyle) {
						*fillStyle &= *(tr->getChar(i)->getFillStyle());
					} else {
						fillStyle = new GDStyle(*(tr->getChar(i)->getFillStyle()));
					}
				}
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				if (fillStyle) {
					*fillStyle &= *(tr->getChar(i)->getFillStyle());
				} else {
					fillStyle = new GDStyle(*(tr->getChar(i)->getFillStyle()));
				}
			}
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				if (fillStyle) {
					*fillStyle &= *(tr->getChar(i)->getFillStyle());
				} else {
					fillStyle = new GDStyle(*(tr->getChar(i)->getFillStyle()));
				}
			}
		}
	}
	return fillStyle;
}

void StandardTextObject::setStrokeStyle(const GDStyle& from) {
	if (from.valueMask & GDSTYLE_GUIDE_MASK) return; // text objects cannot be guides
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			for (int i = selStartColumn; i < selEndColumn; i++) {
				*(tr->getChar(i)->getStrokeStyle()) |= from;
			}
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				*(tr->getChar(i)->getStrokeStyle()) |= from;
			}
			for (guint j = selStartRow + 1; j < (guint) selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					*(tr->getChar(i)->getStrokeStyle()) |= from;
				}
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				*(tr->getChar(i)->getStrokeStyle()) |= from;
			}
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				*(tr->getChar(i)->getStrokeStyle()) |= from;
			}
		}
	}
}

void StandardTextObject::setFillStyle(const GDStyle& from) {
	if (from.valueMask & GDSTYLE_GUIDE_MASK) return; // text objects cannot be guides
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			for (int i = selStartColumn; i < selEndColumn; i++) {
				*(tr->getChar(i)->getFillStyle()) |= from;
			}
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				*(tr->getChar(i)->getFillStyle()) |= from;
			}
			for (int j = selStartRow + 1; j < selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					*(tr->getChar(i)->getFillStyle()) |= from;
				}
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				*(tr->getChar(i)->getFillStyle()) |= from;
			}
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				*(tr->getChar(i)->getFillStyle()) |= from;
			}
		}
	}
}

void StandardTextObject::setTextStyle(TextStyle& style) {
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			for (int i = selStartColumn; i < selEndColumn; i++) {
				*(tr->getChar(i)->getTextStyle()) |= style;
			}
			if (tr->size()) {
				tr->defaultStyle = *(tr->getChar(0)->getTextStyle());
			} else {
				tr->defaultStyle |= style;
			}
			tr->calcLength();
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				*(tr->getChar(i)->getTextStyle()) |= style;
			}
			if (tr->size()) {
				tr->defaultStyle = *(tr->getChar(0)->getTextStyle());
			} else {
				tr->defaultStyle |= style;
			}
			tr->calcLength();
			for (int j = selStartRow + 1; j < selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					*(tr->getChar(i)->getTextStyle()) |= style;
				}
				if (tr->size()) {
					tr->defaultStyle = *(tr->getChar(0)->getTextStyle());
				} else {
					tr->defaultStyle |= style;
				}
				tr->calcLength();
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				*(tr->getChar(i)->getTextStyle()) |= style;
			}
			if (tr->size()) {
				tr->defaultStyle = *(tr->getChar(0)->getTextStyle());
			} else {
				tr->defaultStyle |= style;
			}
			tr->calcLength();
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				*(tr->getChar(i)->getTextStyle()) |= style;
			}
			if (tr->size()) {
				tr->defaultStyle = *(tr->getChar(0)->getTextStyle());
			} else {
				tr->defaultStyle |= style;
			}
			tr->calcLength();
		}
	}
}

TextStyle StandardTextObject::getTextStyle() {
	gboolean initialized = false;
	TextStyle style;
	if (editing) {
		if (selStartRow == selEndRow) {
			TextRow *tr = text[selStartRow];
			if (selStartColumn == selEndColumn) {
				if (tr->size()) {
					style = selEndColumn ? *(tr->getChar(selEndColumn - 1)->getTextStyle()) : *(tr->getChar(0)->getTextStyle());
				} else {
					style = doc->currentTextStyle;
				}
				initialized = true;
			} else {
				for (int i = selStartColumn; i < selEndColumn; i++) {
					if (initialized) {
						style &= *(tr->getChar(i)->getTextStyle());
					} else {
						style = *(tr->getChar(i)->getTextStyle());
						initialized = true;
					}
				}
			}
		} else {
			TextRow *tr = text[selStartRow];
			for (guint i = selStartColumn; i < tr->size(); i++) {
				if (initialized) {
					style &= *(tr->getChar(i)->getTextStyle());
				} else {
					style = *(tr->getChar(i)->getTextStyle());
					initialized = true;
				}
			}
			for (int j = selStartRow + 1; j < selEndRow; j++) {
				TextRow *tr = text[j];
				for (guint i = 0; i < tr->size(); i++) {
					if (initialized) {
						style &= *(tr->getChar(i)->getTextStyle());
					} else {
						style = *(tr->getChar(i)->getTextStyle());
						initialized = true;
					}
				}
			}
			tr = text[selEndRow];
			for (int i = 0; i < selEndColumn; i++) {
				if (initialized) {
					style &= *(tr->getChar(i)->getTextStyle());
				} else {
					style = *(tr->getChar(i)->getTextStyle());
					initialized = true;
				}
			}
		}
	} else {
		vector<TextRow *>::iterator it = text.begin();
		for (; it != text.end(); it++) {
			TextRow *tr = *it;
			for (guint i = 0; i < tr->size(); i++) {
				//g_printerr("StandardTextObject::getTextStyle: Char %d (%c)\n", i, tr->getChar(i)->unicode);
				if (initialized) {
					style &= *(tr->getChar(i)->getTextStyle());
				} else {
					style = *(tr->getChar(i)->getTextStyle());
					initialized = true;
				}
			}
		}
	}
	return style;
}

vector<GDCharacter> StandardTextObject::getTextSelection() {
	vector<GDCharacter> tSel;
	if (editing) {
		for (int i = selStartRow; i <= selEndRow; i++) {
			TextRow *tr = text[i];
			int start = (i == selStartRow) ? selStartColumn : 0;
			int end = (i == selEndRow) ? selEndColumn : tr->size();
			for (int x = start; x < end; x++) {
				tSel.insert(tSel.end(), *(tr->getChar(x)));
			}
			if ((guint) end == tr->size() && i < selEndRow) {
				tSel.insert(tSel.end(), GDCharacter(CHARCODE_LINEFEED));
			}
		}
	}
	//g_printerr("returned Text Selection size %d\n", tSel.size());
	return tSel;
}

void StandardTextObject::insertTextSelection(vector<GDCharacter> selection) {
	removeSelectedText();
	vector<GDCharacter>::iterator it = selection.begin();
	for(; it != selection.end(); it++) {
		if ((*it).getUnicode() == CHARCODE_LINEFEED) {
			TextRow *newRow = new TextRow();
			newRow->defaultStyle = doc->currentTextStyle;
			insertRow(selStartRow, newRow);
			for (; (guint) selStartColumn < text[selStartRow]->size();) {
				newRow->insertChar(-1, text[selStartRow]->removeChar(selStartColumn));
			}
			selStartRow++;
			selEndRow++;
			selStartColumn = 0;
			selEndColumn = 0;
		} else {
			GDCharacter *newChar = new GDCharacter(*it);
			gint insPos = -1;
			if ((guint) selStartColumn < text[selStartRow]->size()) {
				insPos = selStartColumn;
			}
			text[selStartRow]->insertChar(insPos, newChar);
			selStartColumn++;
			selEndColumn++;
		}
	}
}

void StandardTextObject::setDocument(GDDocument *document) {
	if (doc != document) {
		for (vector<TextRow *>::iterator it = text.begin(); it != text.end(); it++) {
			(*it)->changeDocument(document);
		}
	}
	GObject::setDocument(document);
}

ObjectGroup *StandardTextObject::convert2Outline() {
	//g_printerr("StandardTextObject::convert2Outline %lX\n", this);
	ObjectGroup *outlines = new ObjectGroup(doc);
	gdouble rowY = 0;
	for (guint r = 0; r < text.size(); r++) {
		//g_printerr("StandardTextObject::convert2Outline drawing line %d at Y: %f\n", r, rowY);
		TextRow *textrow = text[r];
		GDFont *oldfont = NULL;
		double oldsize = -1;
		double oldhs = -1;
		FT_Face face = NULL;
		gdouble x = 0;
		if (textrow->size() > 0) {
			TextStyle *primeStyle = textrow->getChar(0)->getTextStyle();
			if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				x = -(textrow->getWidth());
			} else if (primeStyle->getJustification() == TEXT_STYLE_JUSTIFICATION_CENTER) {
				x = -(textrow->getWidth() / 2.0);
			}
			int gindex = -1;
			for (guint i = 0; i < textrow->size(); i++) {
				GDCharacter *ch = textrow->getChar(i);
				//g_printerr("StandardTextObject::convert2Outline character %d: %c (%d/%d)\n", i, ch->unicode, r, i);
				TextStyle chStyle = *(ch->getTextStyle());
				if (chStyle.getFont() != oldfont || 
						chStyle.getPointSize() != oldsize || 
						chStyle.getHorizontalScale() != oldhs) {
					oldfont = chStyle.getFont();
					oldsize = chStyle.getPointSize();
					oldhs = chStyle.getHorizontalScale();
					face = oldfont->getFace();
					int pixWidth = 0;
					if (oldhs != 1) {
						pixWidth = (int) (oldsize * oldhs);
					}
					//g_printerr("Font: name: %s psName: %s\n", oldfont->getName(), oldfont->getPSName());
					//g_printerr("Changed Font: face: %lX, size: %f\n", face, oldsize);
					FT_Set_Pixel_Sizes(face, pixWidth, (int) (oldsize));
				}
				if (chStyle.getTracking()) {
					x += (chStyle.getTracking() * oldsize / 1000);
				}
				if (gindex == -1) {
					gindex = FT_Get_Char_Index(face, ch->getUnicode());
				}
				int error = FT_Load_Glyph(face, gindex, FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING);
				if (error) {
					g_printerr("StandardTextObject::convert2Outline ERROR cannot load glyph %d\n", gindex);
				}
				position st;
				st.x = x;
				st.y = rowY - chStyle.getBaseLineShift();
				//g_printerr("StandardTextObject::convert2Outline x=%f y=%f\n", st.x, st.y);
				x += ((gdouble) face->glyph->advance.x) / 64.0; // calculate the x advance before transforming because transforming alters the advance vector
				//g_printerr("after advance: x=%f (advance = %d)\n",x, face->glyph->advance.x);
				FT_Glyph glyph;
				error = FT_Get_Glyph(face->glyph, &glyph);
				if (error) {
					g_printerr("StandardTextObject::convert2Outline cannot get glyph %d\n", error);
				}
				GObject *charOutline = ch->convert2Outline(doc);
				if (charOutline) {
					// transformation and stuff is only necessary if an outline exists for this character
					TransformationMatrix trans;
					trans.setTranslation(st.x, st.y);
					trans *= tMatrix;
					charOutline->transformByMatrix(trans, FALSE);
					outlines->addPart(charOutline);
				}
				FT_Done_Glyph(glyph);
				if (i < textrow->size() - 1) {
					int oldindex = gindex;
					gindex = FT_Get_Char_Index(face, textrow->getChar(i+1)->getUnicode());
					FT_Vector kerning;
					error = FT_Get_Kerning(face, oldindex, gindex, ft_kerning_default, &kerning);
					//g_printerr("Kerning: %d %d\n", kerning.x, kerning.y);
					x += ((gdouble) kerning.x) / 64.0;
				}
			}
		}
		rowY += textrow->getLineSpace();
	}
	return outlines;
}



StandardTextObject::~StandardTextObject() {
	// deSelectAll(); // commented out for code cleanup
	for (vector<TextRow *>::iterator it = text.begin(); it != text.end(); it++) {
		delete *it;
	}
}
