/* $Id: TransformationMatrix.cc,v 1.8 2004/03/08 10:24:00 pklaus Exp $
 *
 * TransformationMatrix.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "TransformationMatrix.hh"
#include <math.h>
#include <iostream>

TransformationMatrix::TransformationMatrix() {
	reset();
}

void TransformationMatrix::reset() {
	m[0][0] = 1;
	m[0][1] = 0;
	m[0][2] = 0;
	m[1][0] = 0;
	m[1][1] = 1;
	m[1][2] = 0;
	m[2][0] = 0;
	m[2][1] = 0;
	m[2][2] = 1;
}

gboolean TransformationMatrix::isIdentityMatrix() {
	if (	m[0][0] == 1	&&
			m[0][1] == 0	&&
			m[0][2] == 0	&&
			m[1][0] == 0	&&
			m[1][1] == 1	&&
			m[1][2] == 0	&&
			m[2][0] == 0	&&
			m[2][1] == 0	&&
			m[2][2] == 1	) {
		return true;
	} else {
		return false;
	}
}

TransformationMatrix::TransformationMatrix(const TransformationMatrix& n) {
	m[0][0] = n.m[0][0];
	m[0][1] = n.m[0][1];
	m[0][2] = n.m[0][2];
	m[1][0] = n.m[1][0];
	m[1][1] = n.m[1][1];
	m[1][2] = n.m[1][2];
	m[2][0] = n.m[2][0];
	m[2][1] = n.m[2][1];
	m[2][2] = n.m[2][2];
}

void TransformationMatrix::setTranslation(gdouble dx, gdouble dy) {
	m[0][2] = dx;
	m[1][2] = dy;
}

void TransformationMatrix::setRotation(gdouble angle) {
	//g_printerr("TransformationMatrix::setRotation angle=%f \n", angle * 180 / M_PI);
	gdouble ca = cos(angle);
	gdouble sa = sin(angle);
	m[0][0] = ca;
	m[0][1] = -sa;
	m[1][0] = sa;
	m[1][1] = ca;
}

void TransformationMatrix::setScale(gdouble hf, gdouble vf) {
	m[0][0] = hf;
	m[1][1] = vf;
}

void TransformationMatrix::setShear(gdouble ax, gdouble ay) {
	m[0][1] = ax;
	m[1][0] = ay;
}


void TransformationMatrix::printMatrix() {
	for (int i = 0; i < 3; i++) {
		g_printerr("   %f   %f   %f\n", m[i][0], m[i][1], m[i][2]);
	}
	g_printerr("\n");
}

void TransformationMatrix::setCenter(gdouble x, gdouble y) {
	TransformationMatrix tmp = *this;
	//g_printerr("TransformationMatrix::setCenter x:%f, y:%f\n", x, y);
	//this->printMatrix();
	reset();
	setTranslation(-x, -y);
	//this->printMatrix();
	*this *= tmp;
	//this->printMatrix();
	tmp.reset();
	tmp.setTranslation(x, y);
	//tmp.printMatrix();
	*this *= tmp;
	//this->printMatrix();
}

void TransformationMatrix::transform(position& p) {
	gdouble xo = p.x;
	gdouble yo = p.y;
	p.x = m[0][0] * xo + m[0][1] * yo + m[0][2];
	p.y = m[1][0] * xo + m[1][1] * yo + m[1][2];
}

gdouble TransformationMatrix::det22(int i, int k) {
	int i0, i1, k0, k1;
	if (i == 0) {
		i0 = 1; i1 = 2;
	} else if (i == 1) {
		i0 = 0; i1 = 2;
	} else {
		i0 = 0; i1 = 1;
	}
	if (k == 0) {
		k0 = 1; k1 = 2;
	} else if (k == 1) {
		k0 = 0; k1 = 2;
	} else {
		k0 = 0; k1 = 1;
	}
	return m[k0][i0] * m[k1][i1] - m[k0][i1] * m[k1][i0];
}

TransformationMatrix TransformationMatrix::inverse() {
	TransformationMatrix inv;
	gdouble revDet = 1 / (
					m[0][0] * ( m[1][1] * m[2][2] - m[1][2] * m[2][1] ) -
					m[0][1] * ( m[1][0] * m[2][2] - m[1][2] * m[2][0] ) +
					m[0][2] * ( m[1][0] * m[2][1] - m[1][1] * m[2][0] ) );
	//g_printerr("TransformationMatrix::inverse revDet: %f\n", revDet);
	for (int i = 0; i < 3; i++) {
		for (int k = 0; k < 3; k++) {
			inv.m[i][k] = (((i+k)%2) ? -revDet : revDet) * det22(i, k);
			//g_printerr("TransformationMatrix::inverse: %d %d  o: %f i: %f\n", i,k, m[i][k], inv.m[i][k]);
		}
	}
	return inv;
}

TransformationMatrix& TransformationMatrix::operator*=(const TransformationMatrix &a) {
	gdouble x[3][3];
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			/*x[i][j] = 	m[i][0] * a.m[0][j] + 
						m[i][1] * a.m[1][j] + 
						m[i][2] * a.m[2][j];*/
			x[i][j] = 	m[0][j] * a.m[i][0] + 
						m[1][j] * a.m[i][1] + 
						m[2][j] * a.m[i][2];
		}
	}
	/* was m = x; */
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			m[i][j] = x[i][j];
		}
	}
	return *this;
}

TransformationMatrix TransformationMatrix::operator* (const TransformationMatrix& b) {
	TransformationMatrix prod = *this;
	prod *= b;
	return prod;
}

gboolean TransformationMatrix::operator==(const TransformationMatrix& b) {
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			if (m[i][j] != b.m[i][j]) {
				return false;
			}
		}
	}
	return true;
}

gdouble TransformationMatrix::getAngle(gdouble x, gdouble y) {
	gdouble ax = fabs(x);
	gdouble ay = fabs(y);
	gdouble gamma;
	if (ax > ay) {
		gamma = atan(ay/ax);
	} else {
		gamma = M_PI_2 - atan(ax/ay);
	}
	//g_printerr("TransformationMatrix::getAngle: x: %f, y: %f, gamma: %f\n", x, y, gamma * 180 / M_PI);
	if (x < 0 && y >= 0) {
		gamma = M_PI - gamma;
	} else if (x < 0 && y < 0) {
		gamma += M_PI;
	} else if (x >= 0 && y < 0) {
		gamma = -gamma;
	}
	//g_printerr("TransformationMatrix::getAngle: x: %f, y: %f, corr_gamma: %f\n", x, y, gamma * 180 / M_PI);
	return gamma;
}

void TransformationMatrix::writePS(ostream& psout) {
	psout << "[" << m[0][0] << " " << m[1][0];
	psout << " " << m[0][1] << " " << m[1][1];
	psout << " " << m[0][2] << " " << m[1][2];
	psout << "]";
}

void TransformationMatrix::writeXML(ostream& xmlout) {
	xmlout << "matrix(";
	xmlout << m[0][0] << ",";
	xmlout << m[1][0] << ",";
	xmlout << m[0][1] << ",";
	xmlout << m[1][1] << ",";
	xmlout << m[0][2] << ",";
	xmlout << m[1][2] << ")";
}
